


/*		MODULE FOR CRYSTAL IMAGE 0.85 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.

		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */




#include <gtk/gtk.h>
#include <gdk/gdk.h>

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <fftw3.h>


#include "includeall.h"





void generator_file_ok (char* title)
{

	cur_slice=0;
	gtk_adjustment_init_val ((GtkAdjustment*)slice_spinbutton_adj, 0, 0, mainimg.zmax-1);
	prepare_display_window (title);

	/* Finally, create an imlib image for display */

	make_minmax (&mainimg);
	make_imlib_buffer_from_main_image (0);
	file_ok = 1;
	redraw (0);

}



/********************************************************************

	Some very simple functions

*********************************************************************/


int make_spec_function (image_type* img, int imgsize, int func)
{
int x,y, xm, ym;
float *p1;
float d;

	if (allocate_image (img, B_FLOAT, imgsize,imgsize,1)==-1)
		fatal_memerr ();
	img->minval.imin=0; img->maxval.imax=255;
	p1 = img->data;
	xm = ym = imgsize/2;


	/* One pass over the image  */

	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
		{
			d = sqrt(ARITHM ( (x-xm), (y-ym)));
			if (F_ISZERO(d))
				p1[x + img->xmax*y] = xm;
			else
				p1[x + img->xmax*y] = xm/d;

		}

	make_minmax (img);

	return 0;

}




/********************************************************************

	Create a 2D frequency-domain mask from one of the fft filters

	In rev. 1.0.0, major changes were made. To keep the FFT filter
	GUI in the main menu and this menu consistent, the FFT GUI was
	expanded with callback functions so that it can be called from
	here. This cut out over 230 lines of code. Moreover, in this
	fashion, generating the windowing functions is now possible
	has has been added.

	Unfortunately, this level of integration violates the spirit of
	the DSOs and brings up the question if the generator module
	shouldn't become part of the main codebase, perhaps as
	File -> New

	The function make_fftmask is used by the macro hookup. The
	macro function is probably not up-to-date.

*********************************************************************/

ffilt_mat (*mkfft_filter_func) (int x, int y, float a, float b, float c, float d) = NULL;


void make_fftmask (image_type* img, int size, float a, float b, float c, float d)
{
int x,y,idx,xm,ym;
float* p;
ffilt_mat H;

	/* Allocate 2 FLOAT slices (real and imag) */

	if (allocate_image (img, B_FLOAT, size,size, 2)==-1)
		fatal_memerr ();
	p = img->data;

	idx = SQR(size); xm=ym=size/2;

	for (y=0; y<size; y++)
		for (x=0; x<size; x++)
		{
			H = mkfft_filter_func (x-xm,y-ym, a,b,c,d);
			p[x + size*y      ] = H.A[0];			/* Real component */
			p[x + size*y + idx] = H.A[3];			/* Imag component */
			dp (3, "(%d %d)  -> %f\n",x,y, H.A[0]);
		}

	make_minmax (img);


}

/* Callback function, called from the GUI in fft.c if the GUI window was built
	from here. There are two ways this callback function is called:

	* Filter mask (MTF). Pointer to the MTF is passed in func.
	* Window function. Pointer func is NULL, a (integer) is the window function.
*/



void mkfftmask_go  (gpointer func, int size, float a, float b, float c, float d)
{
int x,y,idx,xm,ym;
float* p;
ffilt_mat H;
WINDOW_TYPE fftwnd;
gchar wndtitle[256];


	idx = SQR(size); xm=ym=size/2;

	if (func)					/* MTF / filter mask */
	{
		mkfft_filter_func = func;

		/* Allocate 2 FLOAT slices (real and imag) */

		if (allocate_image (&mainimg, B_FLOAT, size,size, 2)==-1)
			fatal_memerr ();
		p = mainimg.data;

		for (y=0; y<size; y++)
			for (x=0; x<size; x++)
			{
				H = mkfft_filter_func (x-xm,y-ym, a,b,c,d);
				p[x + size*y      ] = H.A[0];			/* Real component */
				p[x + size*y + idx] = H.A[3];			/* Imag component */
				dp (3, "(%d %d)  -> %f\n",x,y, H.A[0]);
			}

		strcpy (wndtitle, "FFT Filter Mask");
	}
	else					/* Window function */
	{
		/* Allocate one FLOAT slice */

		if (allocate_image (&mainimg, B_FLOAT, size,size, 1)==-1)
			fatal_memerr ();
		p = mainimg.data;

		fftwnd = IROUND(a);
		for (y=0; y<size; y++)
			for (x=0; x<size; x++)
				p[x + size*y] = 100.0*fft_windowfn ( fftwnd, x, y, size, size);

		switch (fftwnd)
		{
			case W_WELCH:
				strcpy (wndtitle, "FFT Window (Welch)");
				break;
			case W_HAMMING:
				strcpy (wndtitle, "FFT Window (Hamming)");
				break;
			case W_HANN:
				strcpy (wndtitle, "FFT Window (Hann)");
				break;
			case W_GAUSS:
				strcpy (wndtitle, "FFT Window (Gaussian)");
				break;
			default:
				strcpy (wndtitle, "FFT Window Function");
		}
	}

	generator_file_ok (wndtitle);

}


void mkfftmask_gui ()
{
	create_fftfilt_wnd (NULL, &mkfftmask_go);

}




/**************************************************************

	Psychedelic rings: Something similar to the test image in
	Bankmann, page 412

***************************************************************/


void make_psy_rings (image_type* img, int imgsize, int frequency)
{
int x,y,xm,ym;
float r,buf;

	if (allocate_image (img, B_CHAR, imgsize,imgsize,1)==-1)
		fatal_memerr ();
	img->minval.imin=0; img->maxval.imax=255;
	xm=ym=imgsize/2;

	for (y=0; y<imgsize; y++)
		for (x=0; x<imgsize; x++)
		{
			r = sqrt( SQR(x-xm)+SQR(y-ym) );
			buf = 1.0+sin(frequency*sqrt(r));
			((unsigned char*)img->data)[x + img->xmax*y] = (int)(0.4999+127*buf);

		}


}


GtkAdjustment *psyrings_imgsize_adj;
GtkAdjustment *psyrings_freq_adj;




void on_psyrings_cancelbutton_clicked  (GtkButton *button, gpointer user_data)
{

	gtk_widget_destroy (GTK_WIDGET (user_data));
}


void on_psyrings_okbutton_clicked  (GtkButton *button, gpointer user_data)
{
int imgsize, freq;


	imgsize  = psyrings_imgsize_adj->value;
	freq     = psyrings_freq_adj->value;

	make_psy_rings (&mainimg, imgsize, freq);
	generator_file_ok ("Psychedelic rings");
	statusbar_display ("Image generation finished");

	gtk_widget_destroy (GTK_WIDGET (user_data));
}


void make_psyrings_gui ()
{
GtkWidget* wnd;
GtkWidget* vbox;
GtkWidget* table;
GtkWidget *label1;
GtkWidget *okbutton, *cancelbutton;
GtkWidget *psyrings_imgsize, *psyrings_freq;


	wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (wnd), "wnd001", wnd);
	gtk_container_set_border_width (GTK_CONTAINER (wnd), 3);
	gtk_window_set_title (GTK_WINDOW (wnd), "Create psychedelic rings");
	gtk_window_set_modal (GTK_WINDOW (wnd), TRUE);
	vbox = vbox_in_window (wnd,"vbox001");
	table = MakeTable (wnd,vbox,"table001",2,4);
	label1 = MakeLabel (wnd,table,"Image size","label1",0,0);
	label1 = MakeLabel (wnd,table,"Frequency","label1",0,1);

	psyrings_imgsize_adj = GTK_ADJUSTMENT (gtk_adjustment_new (512, 128, 2048, 1, 128, 0));
	psyrings_imgsize = MakeHScale (psyrings_imgsize_adj,wnd,table,"psyrings_imgsize",1,0);
	gtk_scale_set_digits (GTK_SCALE (psyrings_imgsize), 0);

	psyrings_freq_adj = GTK_ADJUSTMENT (gtk_adjustment_new (5, 1, 20, 1, 1, 0));
	psyrings_freq = MakeHScale (psyrings_freq_adj,wnd,table,"psyrings_imgsize",1,1);
	gtk_scale_set_digits (GTK_SCALE (psyrings_freq), 0);

	okbutton = Button_in_table (wnd,table,"OK","psyrings_okbutton",0,2,NULL);
	cancelbutton = Button_in_table (wnd,table,"Cancel","psyrings_cancelbutton",1,2,NULL);

	gtk_signal_connect (GTK_OBJECT (wnd), "destroy",
                      GTK_SIGNAL_FUNC (on_psyrings_cancelbutton_clicked), wnd);
	gtk_signal_connect (GTK_OBJECT (cancelbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_psyrings_cancelbutton_clicked), wnd);
	gtk_signal_connect (GTK_OBJECT (okbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_psyrings_okbutton_clicked), wnd);

	gtk_window_set_default_size (GTK_WINDOW(wnd), 240, -1);
	gtk_widget_show (wnd);

}







/**************************************************************

	Rectangular grid. Just that. A bunch of horizontal
	and vertical lines. Or (fill set to TRUE), a square
	checkerboard pattern.

***************************************************************/


void make_grid (image_type* img, int imgsize, int xdist, int ydist, int fill)
{
int x,y,xm,ym,i,j;
unsigned char flag;

	if (allocate_image (img, B_CHAR, imgsize,imgsize,1)==-1)
		fatal_memerr ();
	img->minval.imin=0; img->maxval.imax=255;
	xm=ym=imgsize/2;
	flag = 0;

	if (fill)						/* Checkerboard pattern */
	{
		for (y=0; y<imgsize; y++)
		{
			j = (y / ydist);			/* integer division (!) */
			for (x=0; x<imgsize; x++)
			{
				i = (x / xdist);
				flag = ( (i & 1) == (j & 1) );		/* one-bit exnor */
				((unsigned char*)img->data)[x + img->xmax*y] = flag*255;
			}
		}
	}
	else			/* One-pixel wide lines. Ensure that one line pair meets in the image center */
	{
		for (y=0; y<imgsize/2; y+=ydist)
			for (x=0; x<imgsize; x++)
			{
				((unsigned char*)img->data)[x + img->xmax*(ym+y)] = 255;
				((unsigned char*)img->data)[x + img->xmax*(ym-y)] = 255;
			}

		for (x=0; x<imgsize/2; x+=xdist)
			for (y=0; y<imgsize; y++)
			{
				((unsigned char*)img->data)[xm+x + img->xmax*y] = 255;
				((unsigned char*)img->data)[xm-x + img->xmax*y] = 255;
			}
	}



}


GtkAdjustment *ggrid_imgsize_adj;
GtkAdjustment *ggrid_xstep_adj;
GtkAdjustment *ggrid_ystep_adj;
GtkWidget *checkerboard_chkbtn;



void on_ggrid_cancelbutton_clicked  (GtkButton *button, gpointer user_data)
{

	gtk_widget_destroy (GTK_WIDGET (user_data));
}


void on_ggrid_okbutton_clicked  (GtkButton *button, gpointer user_data)
{
int imgsize, xstep, ystep, fill;


	imgsize  = ggrid_imgsize_adj->value;
	xstep    = ggrid_xstep_adj->value;
	ystep    = ggrid_ystep_adj->value;
	fill = gtk_toggle_button_get_active ((GtkToggleButton*)checkerboard_chkbtn);

	make_grid (&mainimg, imgsize, xstep, ystep, fill);
	generator_file_ok ("Rectangular grid");
	statusbar_display ("Image generation finished");

	gtk_widget_destroy (GTK_WIDGET (user_data));
}


void make_grid_gui ()
{
GtkWidget* wnd;
GtkWidget* vbox;
GtkWidget* table;
GtkWidget *label1;
GtkWidget *okbutton, *cancelbutton;
GtkWidget *ggrid_imgsize, *ggrid_xstep, *ggrid_ystep;


	wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (wnd), "wnd001", wnd);
	gtk_container_set_border_width (GTK_CONTAINER (wnd), 3);
	gtk_window_set_title (GTK_WINDOW (wnd), "Create rectangular grid");
	gtk_window_set_modal (GTK_WINDOW (wnd), TRUE);
	vbox = vbox_in_window (wnd,"vbox001");
	table = MakeTable (wnd,vbox,"table001",2,6);
	label1 = MakeLabel (wnd,table,"Image size","label1",0,0);
	label1 = MakeLabel (wnd,table,"X distance","label1",0,1);
	label1 = MakeLabel (wnd,table,"Y distance","label1",0,2);
	label1 = MakeLabel (wnd,table,"Fill","label1",0,3);

	ggrid_imgsize_adj = GTK_ADJUSTMENT (gtk_adjustment_new (512, 128, 2048, 1, 128, 0));
	ggrid_imgsize = MakeHScale (ggrid_imgsize_adj,wnd,table,"ggrid_imgsize",1,0);
	gtk_scale_set_digits (GTK_SCALE (ggrid_imgsize), 0);

	ggrid_xstep_adj = GTK_ADJUSTMENT (gtk_adjustment_new (32, 2, 256, 1, 8, 0));
	ggrid_xstep = MakeHScale (ggrid_xstep_adj,wnd,table,"ggrid_xstep",1,1);
	gtk_scale_set_digits (GTK_SCALE (ggrid_xstep), 0);
	ggrid_ystep_adj = GTK_ADJUSTMENT (gtk_adjustment_new (32, 2, 256, 1, 8, 0));
	ggrid_ystep = MakeHScale (ggrid_ystep_adj,wnd,table,"ggrid_ystep",1,2);
	gtk_scale_set_digits (GTK_SCALE (ggrid_ystep), 0);

	checkerboard_chkbtn= Checkbutton_in_table (wnd, table, " ", "checkercheck1", 1,3, NULL);
	okbutton = Button_in_table (wnd,table,"OK","ggrid_okbutton",0,4,NULL);
	cancelbutton = Button_in_table (wnd,table,"Cancel","ggrid_cancelbutton",1,4,NULL);

	gtk_signal_connect (GTK_OBJECT (wnd), "destroy",
                      GTK_SIGNAL_FUNC (on_ggrid_cancelbutton_clicked), wnd);
	gtk_signal_connect (GTK_OBJECT (cancelbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_ggrid_cancelbutton_clicked), wnd);
	gtk_signal_connect (GTK_OBJECT (okbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_ggrid_okbutton_clicked), wnd);

	gtk_window_set_default_size (GTK_WINDOW(wnd), 240, -1);
	gtk_widget_show (wnd);

}




/**************************************************************

	Sinusoidal waves in X and Y direction


***************************************************************/



void make_sinexy (image_type *img, int imgsize, int freqx, int ampx,
					int freqy, int ampy, int rot)
{
float xarg, yarg, r;
float st, ct;
int x,y,x1,y1,imgh;


	if (allocate_image (img, B_FLOAT, imgsize,imgsize,1)==-1)
		fatal_memerr ();

	imgh = imgsize/2;
	st = sin(PI*rot/180.0);
	ct = cos(PI*rot/180.0);

	for (y=0; y<imgsize; y++)
		for (x=0; x<imgsize; x++)
		{
			x1 = (x-imgh)*ct + (y-imgh)*st;
			y1 = -(x-imgh)*st + (y-imgh)*ct;
			xarg = x1*freqx*2*PI/imgsize;
			yarg = y1*freqy*2*PI/imgsize;
			r = ampx*cos(xarg)+ampy*cos(yarg);
			((float*)img->data)[x + img->xmax*y] = r;
		}


}

/* Interesting side note for the sinusoidals used in FFT demos.
	When the cosine is used, the FFT is real-valued whereas if we
	use
			r = ampx*sin(xarg)+ampy*sin(yarg);

	instead, the FFT is imaginary. This, of course, has to do with
	the FFT's symmetry properties that when the TD image is real and even
	(cosine!), then the FT is also real and even. Conversely, when the TD
	is real and odd (sine!), then the FT is imaginary and odd.
	Maybe we should make this a user-selectable option? */



GtkAdjustment *sinexy_imgsize_adj;
GtkAdjustment *sinexy_freqx_adj;
GtkAdjustment *sinexy_freqy_adj;
GtkAdjustment *sinexy_ampx_adj;
GtkAdjustment *sinexy_ampy_adj;
GtkAdjustment *sinexy_rot_adj;



void on_sinexy_cancelbutton_clicked  (GtkButton *button, gpointer user_data)
{

	gtk_widget_destroy (GTK_WIDGET (user_data));
}


void on_sinexy_okbutton_clicked  (GtkButton *button, gpointer user_data)
{
int imgsize;
int freqx, ampx, freqy, ampy, rot;

	imgsize  = sinexy_imgsize_adj->value;
	freqx = sinexy_freqx_adj->value;
	ampx = sinexy_ampx_adj->value;
	freqy = sinexy_freqy_adj->value;
	ampy = sinexy_ampy_adj->value;
	rot = sinexy_rot_adj->value;

	make_sinexy (&mainimg, imgsize, freqx, ampx, freqy, ampy, rot);
	generator_file_ok ("X-Y sinusoids");
	statusbar_display ("Image generation finished");

	gtk_widget_destroy (GTK_WIDGET (user_data));
}



void make_sinexy_gui ()
{
GtkWidget* wnd;
GtkWidget* vbox;
GtkWidget* table;
GtkWidget *label1, *label;
GtkWidget *okbutton, *cancelbutton;
GtkWidget *sinexy_imgsize;
GtkWidget *sinexy_freqx;
GtkWidget *sinexy_freqy;
GtkWidget *sinexy_ampx;
GtkWidget *sinexy_ampy;
GtkWidget *sinexy_rot;



	wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (wnd), "wnd001", wnd);
	gtk_container_set_border_width (GTK_CONTAINER (wnd), 3);
	gtk_window_set_title (GTK_WINDOW (wnd), "Create Sinusoids");
	gtk_window_set_modal (GTK_WINDOW (wnd), TRUE);
	vbox = vbox_in_window (wnd,"vbox001");
	table = MakeTable (wnd,vbox,"table001",2,7);
	label1 = MakeLabel (wnd,table,"Image size","label1",0,0);
	label = MakeLabel (wnd,table,"Frequency X","label2",0,1);
	label = MakeLabel (wnd,table,"Amplitude X","label3",0,2);
	label = MakeLabel (wnd,table,"Frequency Y","label4",0,3);
	label = MakeLabel (wnd,table,"Amplitude Y","label5",0,4);
	label = MakeLabel (wnd,table,"Rotation","label6",0,5);

	sinexy_imgsize_adj = GTK_ADJUSTMENT (gtk_adjustment_new (512, 64, 2048, 1, 128, 0));
	sinexy_imgsize = MakeHScale (sinexy_imgsize_adj,wnd,table,"sinexy_imgsize",1,0);
	gtk_scale_set_digits (GTK_SCALE (sinexy_imgsize), 0);

	sinexy_freqx_adj = GTK_ADJUSTMENT (gtk_adjustment_new (10, 2, 100, 1, 10, 0));
	sinexy_freqx = MakeHScale (sinexy_freqx_adj,wnd,table,"sinexy_freqx",1,1);
	gtk_scale_set_digits (GTK_SCALE (sinexy_freqx), 0);

	sinexy_ampx_adj = GTK_ADJUSTMENT (gtk_adjustment_new (255, 0, 255, 1, 10, 0));
	sinexy_ampx = MakeHScale (sinexy_ampx_adj,wnd,table,"sinexy_ampx",1,2);
	gtk_scale_set_digits (GTK_SCALE (sinexy_ampx), 0);

	sinexy_freqy_adj = GTK_ADJUSTMENT (gtk_adjustment_new (10, 2, 100, 1, 10, 0));
	sinexy_freqy = MakeHScale (sinexy_freqy_adj,wnd,table,"sinexy_freqy",1,3);
	gtk_scale_set_digits (GTK_SCALE (sinexy_freqy), 0);

	sinexy_ampy_adj = GTK_ADJUSTMENT (gtk_adjustment_new (255, 0, 255, 1, 10, 0));
	sinexy_ampy = MakeHScale (sinexy_ampy_adj,wnd,table,"sinexy_ampy",1,4);
	gtk_scale_set_digits (GTK_SCALE (sinexy_ampy), 0);

	sinexy_rot_adj = GTK_ADJUSTMENT (gtk_adjustment_new (0, 0, 359, 1, 10, 0));
	sinexy_rot = MakeHScale (sinexy_rot_adj,wnd,table,"sinexy_rot",1,5);
	gtk_scale_set_digits (GTK_SCALE (sinexy_rot), 0);

	okbutton = Button_in_table (wnd,table,"OK","sinexy_okbutton",0,6,NULL);
	cancelbutton = Button_in_table (wnd,table,"Cancel","sinexy_cancelbutton",1,6,NULL);

	gtk_signal_connect (GTK_OBJECT (wnd), "destroy",
                      GTK_SIGNAL_FUNC (on_sinexy_cancelbutton_clicked), wnd);
	gtk_signal_connect (GTK_OBJECT (cancelbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_sinexy_cancelbutton_clicked), wnd);
	gtk_signal_connect (GTK_OBJECT (okbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_sinexy_okbutton_clicked), wnd);

	gtk_window_set_default_size (GTK_WINDOW(wnd), 280, -1);
	gtk_widget_show (wnd);

}



/**************************************************************

	The Shepp-Logan Head Phantom


***************************************************************/


void make_headphantom (image_type* img, int imgsize, int contrast, float noise)
{

	/* Define 10 ellipses as x0,y0, major, minor, rotation, value */

float ellipsedef[10][6] = {	{0,0,		0.92,0.69,90,		2},
							{0,-0.0184,	0.874,0.6624,90,	-0.98},
							{0.22,0,	0.31,0.11,72,		-0.02},
							{-0.22,0,	0.41,0.16,108,		-0.02},
							{0,0.35,	0.25,0.21,90,		0.01},
							{0,0.1,		0.046,0.046,0,		0.01},
							{0,-0.1,	0.046,0.046,0,		0.01},
							{-0.08,-0.605,0.0406,0.023,0,	0.01},
							{0,-0.605,	0.023,0.023,0,		0.01},
							{0.06,-0.605,0.046,0.023,90,	0.01}	};
float Ellipsedef[10][6] = {	{0,0,		0.92,0.69,90,		2},
							{0,-0.0184,	0.874,0.6624,90,	-0.8},
							{0.22,0,	0.31,0.11,72,		-0.5},
							{-0.22,0,	0.41,0.16,108,		-0.5},
							{0,0.35,	0.25,0.21,90,		0.3},
							{0,0.1,		0.046,0.046,0,		0.1},
							{0,-0.1,	0.046,0.046,0,		0.1},
							{-0.08,-0.605,0.0406,0.023,0,	0.1},
							{0,-0.605,	0.023,0.023,0,		0.1},
							{0.06,-0.605,0.046,0.023,90,	0.1}	};

int x,y,xm,ym,i,j,cnt;
float xr,yr,xt,yt,xs,ys;
float buf;
float ellips[10][6];

	if (allocate_image (img, B_FLOAT, imgsize,imgsize,1)==-1)
		fatal_memerr ();
	img->minval.fmin=0; img->maxval.fmax=2.0;
	xm=ym=imgsize/2;

	if (contrast)
	{
		for (i=0; i<10; i++)
			for (j=0; j<6; j++)
				ellips[i][j] = Ellipsedef[i][j];
	}
	else
	{
		for (i=0; i<10; i++)
			for (j=0; j<6; j++)
				ellips[i][j] = ellipsedef[i][j];
	}

	for (y=0; y<imgsize; y++)
	{
		yr = (float)(ym-y)/(float)ym;
		for (x=0; x<imgsize; x++)
		{
			xr = (float)(x-xm)/(float)xm;

			cnt=0; buf=0.0;
			for (i=0; i<10; i++)			/* Run over each ellipse */
			{
				xt = xr-ellips[i][0];
				yt = yr-ellips[i][1];	/* Translate ellipse origin */

				xs = xt*COS(ellips[i][4]) + yt*SIN(ellipsedef[i][4]);
				ys = -xt*SIN(ellips[i][4]) + yt*COS(ellipsedef[i][4]);

				if ( (SQR(xs)/SQR(ellips[i][2]) + SQR(ys)/SQR(ellips[i][3])) <=1.0)
				{
					cnt++; buf+=ellips[i][5];
				}
			}
			((float*)img->data)[x + img->xmax*y] = buf;

		}
	}

	/* Add noise if requested */

	if (!F_ISZERO (noise))
	{
		for (y=0; y<imgsize; y++)
		{
			for (x=0; x<imgsize; x++)
			{
				buf = gauss_rnd()*noise + ((float*)img->data)[x + img->xmax*y];
				if (buf<0.0) buf = 0.0;		/* Clamp so we don't get negative values */
				((float*)img->data)[x + img->xmax*y] = buf;
			}
		}
	}


}



GtkAdjustment *headphantom_imgsize_adj;
GtkAdjustment *headphantom_noise_adj;
int sheppcontrast;

void on_headphantom_cancelbutton_clicked  (GtkButton *button, gpointer user_data)
{

	gtk_widget_destroy (GTK_WIDGET (user_data));
}


void on_headphantom_okbutton_clicked  (GtkButton *button, gpointer user_data)
{
int imgsize;
float noise;


	imgsize  = headphantom_imgsize_adj->value;
	noise  = headphantom_noise_adj->value;

	make_headphantom (&mainimg, imgsize, sheppcontrast, noise/100.0);
	generator_file_ok ("Shepp-Logan head phantom");
	statusbar_display ("Image generation finished");

	gtk_widget_destroy (GTK_WIDGET (user_data));
}


void on_shepp_contr_clicked (GtkButton *button, gpointer user_data)
{
	sheppcontrast = !sheppcontrast;
}


void make_headphantom_gui ()
{
GtkWidget* wnd;
GtkWidget* vbox;
GtkWidget* table;
GtkWidget *label1, *label2;
GtkWidget *okbutton, *cancelbutton;
GtkWidget *headphantom_imgsize;
GtkWidget *headphantom_noise;
GtkWidget *contrast_checkbutton;


	sheppcontrast = 0;
	wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (wnd), "wnd001", wnd);
	gtk_container_set_border_width (GTK_CONTAINER (wnd), 3);
	gtk_window_set_title (GTK_WINDOW (wnd), "Create Shepp-Logan head phantom");
	gtk_window_set_modal (GTK_WINDOW (wnd), TRUE);
	vbox = vbox_in_window (wnd,"vbox001");
	table = MakeTable (wnd,vbox,"table001",2,4);
	label1 = MakeLabel (wnd,table,"Image size","label1",0,0);
	label2 = MakeLabel (wnd,table,"Additive noise","label2",0,1);
	label2 = MakeLabel (wnd,table,"Increase contrast","label2",0,2);

	headphantom_imgsize_adj = GTK_ADJUSTMENT (gtk_adjustment_new (512, 64, 2048, 1, 128, 0));
	headphantom_imgsize = MakeHScale (headphantom_imgsize_adj,wnd,table,"headphantom_imgsize",1,0);
	gtk_scale_set_digits (GTK_SCALE (headphantom_imgsize), 0);

	headphantom_noise_adj = GTK_ADJUSTMENT (gtk_adjustment_new (0, 0, 100, 1, 1, 0));
	headphantom_noise = MakeHScale (headphantom_noise_adj,wnd,table,"headphantom_noise",1,1);
	gtk_scale_set_digits (GTK_SCALE (headphantom_noise), 0);

	contrast_checkbutton = Checkbutton_in_table (wnd,table," ","shepp_chkbutton",1,2,on_shepp_contr_clicked);

	okbutton = Button_in_table (wnd,table,"OK","headphantom_okbutton",0,3,NULL);
	cancelbutton = Button_in_table (wnd,table,"Cancel","headphantom_cancelbutton",1,3,NULL);

	gtk_signal_connect (GTK_OBJECT (wnd), "destroy",
                      GTK_SIGNAL_FUNC (on_headphantom_cancelbutton_clicked), wnd);
	gtk_signal_connect (GTK_OBJECT (cancelbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_headphantom_cancelbutton_clicked), wnd);
	gtk_signal_connect (GTK_OBJECT (okbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_headphantom_okbutton_clicked), wnd);

	gtk_window_set_default_size (GTK_WINDOW(wnd), 260, -1);
	gtk_widget_show (wnd);

}






/**************************************************************

	Create a 3D stack of noise data
	Various types from uniform to FBM are possible
	sd is the standard deviation, H is the Hurst exponent

***************************************************************/


void make_noise (image_type* img, int imgsize, int zsize, int noisetype, float sd, float H)
{
int x,y,z,k,n;
int rndnum;
float *p, *p1;
image_type slice;
float a,b,seed1,seed2;



	if ((noisetype==4) || (noisetype==5) || (noisetype==6) || (noisetype==7))
	{
		if (allocate_image (img, B_FLOAT, imgsize,imgsize,zsize)==-1)
			fatal_memerr ();
	}
	else
	{
		if (allocate_image (img, B_CHAR, imgsize,imgsize,zsize)==-1)
			fatal_memerr ();
	}

	p = (float*)calloc(imgsize*imgsize,sizeof(float));
	if (!p) fatal_memerr();

	img->minval.imin=0; img->maxval.imax=255;
	rndnum=0;
	dp (1, "Noise type %d requested, SD=%f, H=%f. \n",noisetype, sd, H);


	for (z=0; z<zsize; z++)
	{
		if (noisetype<=3)
		{
			for (y=0; y<imgsize; y++)
			{
				for (x=0; x<imgsize; x++)
				{
						if (noisetype==1)					/* Box-distributed from 0 to 255 */
							rndnum = ((int)(nr_rand2()*256)) & 0xff;
						else if (noisetype==2)				/* Exponential distribution with unity mean */
							rndnum =  ((int)(eran1())) & 0xff;
						else if (noisetype==3)				/* Gaussian distribution with 127 mean and sd=32 */
							rndnum =  ((int)(gauss_rnd()*sd + 127.0)) & 0xff;

					((unsigned char*)img->data)[x + img->xmax*(y+img->ymax*z)] = rndnum;
				}
			}
		}
		else if ((noisetype==4))			/* Fractional Brownian Noise */
		{
			create_fbm (p, sd, H, imgsize*imgsize, (noisetype==5));
			for (x=0; x<imgsize*imgsize; x++)
				((float*)img->data)[x +(img->xmax*img->ymax*z)] = p[x];
		}
		else if ((noisetype==5))			/* Fractional Brownian Motion, 2D spatial-domain synthesis */
		{

			n = create_fbm_2d (&slice, sd, H, imgsize, (noisetype==5) );
			p1 = slice.data;
			for (y=0; y<imgsize; y++)
				for (x=0; x<imgsize; x++)
			{
				((float*)img->data)[x +(img->xmax*(y+img->ymax*z))] = p1[x+n*y];
			}
			freebuf (&slice);


		}
		else if (noisetype==6)			/* 2D spectral-synthesis FBM */
		{
			create_2d_fbm_spectral (&slice, H, imgsize);
			p1 = slice.data;
			for (x=0; x<imgsize*imgsize; x++)
			{
				((float*)img->data)[x +(img->xmax*img->ymax*z)] = p1[x];
			}
			freebuf (&slice);
		}
		else if (noisetype == 7)		/* 2D Perlin noise with persistence H */
		{
			k = (int) (log (imgsize)/log(2));
			seed1 = 2.0*nr_rand2(); seed2 = 2.0*nr_rand2();
			dp (2, "Creating Perlin noise, H=%f, %d octaves\n",H,k);

			/* We fill the 2D map pixel by pixel. The two seed values are used to bias the
			   random generator (otherwise we'll get the same map over and over). k is the
			   number of octaves for 1/f scaling and goes up to pixel resolution. */

			for (y=0; y<imgsize; y++)
			{
				for (x=0; x<imgsize; x++)
				{
					a = (float)x/(float)imgsize; b = (float)y/(float)imgsize;
					((float*)img->data)[x +img->xmax*(y+img->ymax*z)] = perlin_noise (a+seed1,b+seed2,H,k);
				}
			}
		}
	}
	free (p);
}


GtkAdjustment *mknoise_imgsize_adj;
GtkAdjustment *mknoise_zsize_adj;
GtkAdjustment *mknoise_H_adj;
GtkAdjustment *mknoise_SD_adj;
GtkWidget *mknoise_H;
GtkWidget *mknoise_SD;
GtkWidget *mknoise_l4, *mknoise_l5;
int mknoise_function;


void on_mknoise_cancelbutton_clicked  (GtkButton *button, gpointer user_data)
{

	gtk_widget_destroy (GTK_WIDGET (user_data));
}


void on_mknoise_okbutton_clicked  (GtkButton *button, gpointer user_data)
{
int imgsize, zsize;
float sd, H;
char s[256];


	imgsize  = 	mknoise_imgsize_adj->value;
	zsize  = 	mknoise_zsize_adj->value;
	sd =  		mknoise_SD_adj->value;
	H = 		mknoise_H_adj->value;

	make_noise (&mainimg, imgsize, zsize, mknoise_function+1, sd, H);

	switch (mknoise_function)
	{
		case 0:
			strcpy (s, "White Noise");
			break;
		case 1:
			strcpy (s, "Exponential 1/P Noise");
			break;
		case 2:
			strcpy (s, "Gaussian Noise");
			break;
		case 3:
			strcpy (s, "Fractional Brownian Noise (1D)");
			break;
		case 4:
			strcpy (s, "Fractional Brownian Motion (2D)");
			break;
		case 5:
			strcpy (s, "Spectral-synth Fractional Brownian Noise (2D)");
			break;
		case 6:
			strcpy (s, "Perlin Noise (2D)");
			break;
	}
	generator_file_ok (s);
	statusbar_display ("Noise generation finished");

	gtk_widget_destroy (GTK_WIDGET (user_data));
}


void mknoise_set_menu (GtkWidget *widget, gpointer user_data)
{

	mknoise_function = *(int*) user_data;

	switch (mknoise_function)
	{
		case 0:
		case 1:			/* Hide both SD and H adjustments */
			gtk_widget_hide (mknoise_H);
			gtk_widget_hide (mknoise_SD);
			gtk_widget_hide (mknoise_l4);
			gtk_widget_hide (mknoise_l5);
			break;
		case 2:			/* Hide H, but show SD */
			gtk_widget_hide (mknoise_H);
			gtk_widget_show (mknoise_SD);
			gtk_widget_show (mknoise_l4);
			gtk_widget_hide (mknoise_l5);
			break;
		case 3:
		case 4:			/* Show both SD and H */
			gtk_widget_show (mknoise_H);
			gtk_widget_show (mknoise_SD);
			gtk_widget_show (mknoise_l4);
			gtk_widget_show (mknoise_l5);
			break;
		case 5:
		case 6:			/* Show H, hide SD */
			gtk_widget_show (mknoise_H);
			gtk_widget_hide (mknoise_SD);
			gtk_widget_hide (mknoise_l4);
			gtk_widget_show (mknoise_l5);
			break;
	}
}


void make_noise_gui ()
{
GtkWidget* wnd;
GtkWidget* vbox;
GtkWidget* table;
GtkWidget *label1, *label2, *label3, *hsep1, *hsep2;
GtkWidget *okbutton, *cancelbutton;
GtkWidget* mknoise_imgsize;
GtkWidget* mknoise_zsize;
GtkWidget *mknoise_optionmenu1;
GtkWidget *mknoise_optionmenu1_menu;
GtkWidget *glade_menuitem;


	wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (wnd), "wnd001", wnd);
	gtk_container_set_border_width (GTK_CONTAINER (wnd), 3);
	gtk_window_set_title (GTK_WINDOW (wnd), "Create 3D noise stack");
	gtk_window_set_modal (GTK_WINDOW (wnd), TRUE);
	vbox = vbox_in_window (wnd,"vbox001");
	table = MakeTable (wnd,vbox,"table001",2,7);
	label1 = MakeLabel (wnd,table,"Image size","label1",0,0);
	label2 = MakeLabel (wnd,table,"Stack (z) size","label2",0,1);
	label3 = MakeLabel (wnd,table,"Noise type","label3",0,2);
	mknoise_l4 = MakeLabel (wnd,table,"Std Dev","label4",0,3);
	mknoise_l5 = MakeLabel (wnd,table,"Hurst exp","label5",0,4);

	mknoise_imgsize_adj = GTK_ADJUSTMENT (gtk_adjustment_new (512, 128, 2048, 1, 128, 0));
	mknoise_imgsize = MakeHScale (mknoise_imgsize_adj,wnd,table,"mknoise_imgsize",1,0);
	gtk_scale_set_digits (GTK_SCALE (mknoise_imgsize), 0);

	mknoise_zsize_adj = GTK_ADJUSTMENT (gtk_adjustment_new (16, 1, 512, 1, 16, 0));
	mknoise_zsize = MakeHScale (mknoise_zsize_adj,wnd,table,"mknoise_zsize",1,1);
	gtk_scale_set_digits (GTK_SCALE (mknoise_zsize), 0);

	mknoise_optionmenu1 = Make_Optionmenu (wnd, table,"mknoise_menu",1,2);
	gtk_container_set_border_width (GTK_CONTAINER (mknoise_optionmenu1), 1);
	mknoise_optionmenu1_menu = gtk_menu_new ();
	glade_menuitem = Option_menuitem (mknoise_optionmenu1_menu, GTK_SIGNAL_FUNC(mknoise_set_menu), "Box 0<x<256", (gpointer) &menuopt0);
	glade_menuitem = Option_menuitem (mknoise_optionmenu1_menu, GTK_SIGNAL_FUNC(mknoise_set_menu), "Exponential", (gpointer) &menuopt1);
	glade_menuitem = Option_menuitem (mknoise_optionmenu1_menu, GTK_SIGNAL_FUNC(mknoise_set_menu), "Gaussian", (gpointer) &menuopt2);
	glade_menuitem = Option_menuitem (mknoise_optionmenu1_menu, GTK_SIGNAL_FUNC(mknoise_set_menu), "Fractional Brownian Noise (1D)", (gpointer) &menuopt3);
	glade_menuitem = Option_menuitem (mknoise_optionmenu1_menu, GTK_SIGNAL_FUNC(mknoise_set_menu), "Fractional Brownian Motion (2D)", (gpointer) &menuopt4);
	glade_menuitem = Option_menuitem (mknoise_optionmenu1_menu, GTK_SIGNAL_FUNC(mknoise_set_menu), "2D FBM by spectral synthesis", (gpointer) &menuopt5);
	glade_menuitem = Option_menuitem (mknoise_optionmenu1_menu, GTK_SIGNAL_FUNC(mknoise_set_menu), "2D Perlin Noise", (gpointer) &menuopt6);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (mknoise_optionmenu1), mknoise_optionmenu1_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (mknoise_optionmenu1), 0);
	mknoise_function=0;

	mknoise_SD_adj = GTK_ADJUSTMENT (gtk_adjustment_new (32, 0, 127, 0.1, 1, 0));
	mknoise_SD = MakeHScale (mknoise_SD_adj,wnd,table,"mknoise_SD",1,3);
	gtk_scale_set_digits (GTK_SCALE (mknoise_SD), 1);

	mknoise_H_adj = GTK_ADJUSTMENT (gtk_adjustment_new (0.5, 0, 1, 0.05, 0.2, 0));
	mknoise_H = MakeHScale (mknoise_H_adj,wnd,table,"mknoise_H",1,4);
	gtk_scale_set_digits (GTK_SCALE (mknoise_H), 2);

	hsep1 = MakeHsep (wnd, table, "hsep1", 0,5);
	hsep2 = MakeHsep (wnd, table, "hsep2", 1,5);

	okbutton = Button_in_table (wnd,table,"OK","mknoise_okbutton",0,6,NULL);
	cancelbutton = Button_in_table (wnd,table,"Cancel","mknoise_cancelbutton",1,6,NULL);

	gtk_widget_hide (mknoise_H);
	gtk_widget_hide (mknoise_SD);
	gtk_widget_hide (mknoise_l4);
	gtk_widget_hide (mknoise_l5);

	gtk_signal_connect (GTK_OBJECT (wnd), "destroy",
                      GTK_SIGNAL_FUNC (on_mknoise_cancelbutton_clicked), wnd);
	gtk_signal_connect (GTK_OBJECT (cancelbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_mknoise_cancelbutton_clicked), wnd);
	gtk_signal_connect (GTK_OBJECT (okbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_mknoise_okbutton_clicked), wnd);

	gtk_widget_show (wnd);

}



/**************************************************************

	Create an empty image or a sierpinsky triangle

***************************************************************/


void make_sierp_2d (image_type* img, int imgsize)
{
int x,y;
unsigned char c1, c2;

	if (allocate_image (img, B_CHAR, imgsize,imgsize,1)==-1)
		fatal_memerr ();
	img->minval.imin=0; img->maxval.imax=255;

	/* This algorithm uses the cellular automaton approach where - line by line -
		a pixel is set if the two top neighbors are not equal (i.e., odd sum) */

	((unsigned char*)img->data)[0]=255;		/* top left corner is ON */
	for (y=1; y<imgsize; y++)
	{
		((unsigned char*)img->data)[img->ymax*y]=255;	/* at x==0 */
		for (x=1; x<=y; x++)
		{
			c1 = ((unsigned char*)img->data)[(x-1 + img->xmax*(y-1))];
			c2 = ((unsigned char*)img->data)[(x   + img->xmax*(y-1))];
			if (c1!=c2) ((unsigned char*)img->data)[x + img->xmax*y] = 255;
		}
	}

}


/* Same approach, but for 3D (LOTS of memory!!! */

void make_sierp_3d (image_type* img, int imgsize)
{
int x,y,z, zidx, zidx1;
unsigned char c1, c2, c3;
unsigned char *p;				/* Saves us some typeacsting */

	if (allocate_image (img, B_CHAR, imgsize,imgsize,imgsize)==-1)
		fatal_memerr ();
	img->minval.imin=0; img->maxval.imax=1;
	p = img->data;

	/* This algorithm also uses the cellular automaton approach. Assume each Z slice
		to be an evolution in time. Then each pixel value at (x,y,t) is calculated from
		the previous time step as
			(I(x,y,t-1) + I(x-1,y,t-1) + I(x,y-1,t-1)) mod 2
	*/

	((unsigned char*)img->data)[0]=1;		/* top left corner at (0,0,0) is ON */

	for (z=1; z<img->zmax; z++)
	{
		zidx = img->xmax*img->ymax*z; zidx1 = img->xmax*img->ymax*(z-1);
		for (y=0; y<imgsize; y++)
			for (x=0; x<imgsize; x++)
			{
				if (x==0) c1=0; else c1=p[x-1 + img->xmax*y     + zidx1];
				if (y==0) c2=0; else c2=p[x   + img->xmax*(y-1) + zidx1];
				c3 = p[x + img->xmax*y+zidx1];
				p[x + img->xmax*y + zidx] = (c1+c2+c3) & 0x01;
			}
	}

}


void make_empty (image_type* img, int imgsize, int value)
{
int x,y;

	if (allocate_image (img, B_CHAR, imgsize,imgsize,1)==-1)
		fatal_memerr ();
	img->minval.imin=0; img->maxval.imax=255;
	if (value)
	{
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
				((unsigned char*)img->data)[x + img->xmax*y] = value;
	}
}


GtkAdjustment *sierp_imgsize_adj, *sierp_value_adj;
int sierp_want3d;



void on_sierp_cancelbutton_clicked  (GtkButton *button, gpointer user_data)
{
	gtk_widget_destroy (GTK_WIDGET (user_data));
}


void on_sierp_want3d_clicked  (GtkButton *button, gpointer user_data)
{
	sierp_want3d = !sierp_want3d;
}


void on_sierp_okbutton_clicked  (GtkButton *button, gpointer user_data)
{
int imgsize;

	imgsize  = sierp_imgsize_adj->value;

	if (sierp_want3d)
		make_sierp_3d (&mainimg, imgsize);
	else
		make_sierp_2d (&mainimg, imgsize);
	generator_file_ok ("Sierpinski Triangle");
	statusbar_display ("Sierpinski triangle finished");

	gtk_widget_destroy (GTK_WIDGET (user_data));
}

void on_empty_okbutton_clicked  (GtkButton *button, gpointer user_data)
{
int imgsize, value;

	imgsize  = sierp_imgsize_adj->value;
	value    = sierp_value_adj->value;

	make_empty (&mainimg, imgsize, value);
	generator_file_ok ("Empty Image");
	gtk_widget_destroy (GTK_WIDGET (user_data));
}


/* GUI differentiates between empty image (fn==0) and Sierpinsky (fn!=0) */

void make_sierp_gui (int fn)
{
GtkWidget* wnd;
GtkWidget* vbox;
GtkWidget* table;
GtkWidget *label1;
GtkWidget *okbutton, *cancelbutton;
GtkWidget *chkbutton;
GtkWidget *sierp_imgsize, *sierp_value;


	sierp_want3d = 0;
	wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (wnd), "wnd001", wnd);
	gtk_container_set_border_width (GTK_CONTAINER (wnd), 3);
	if (fn)
		gtk_window_set_title (GTK_WINDOW (wnd), "Create Sierpinski gasket");
	else
		gtk_window_set_title (GTK_WINDOW (wnd), "Create empty image");
	gtk_window_set_modal (GTK_WINDOW (wnd), TRUE);
	vbox = vbox_in_window (wnd,"vbox001");
	table = MakeTable (wnd,vbox,"table001",2,5);

	label1 = MakeLabel (wnd,table,"Image size","label1",0,0);
	sierp_imgsize_adj = GTK_ADJUSTMENT (gtk_adjustment_new (512, 128, 2048, 1, 128, 0));
	sierp_imgsize = MakeHScale (sierp_imgsize_adj,wnd,table,"sierp_imgsize",1,0);
	gtk_scale_set_digits (GTK_SCALE (sierp_imgsize), 0);

	if (fn)				/* Sierpinski gets a 2D/3D button */
	{
		label1 = MakeLabel (wnd,table,"3D pyramid?","label2",0,1);
		chkbutton = Checkbutton_in_table (wnd,table," ", "sierp_chkbutton",1,1,on_sierp_want3d_clicked);
	}
	else				/* Empty image gets a value slider */
	{
		label1 = MakeLabel (wnd,table,"Fill value","label2",0,1);
		sierp_value_adj = GTK_ADJUSTMENT (gtk_adjustment_new (0, 0, 255, 1, 16, 0));
		sierp_value = MakeHScale (sierp_value_adj,wnd,table,"sierp_value",1,1);
		gtk_scale_set_digits (GTK_SCALE (sierp_value), 0);
	}


	okbutton = Button_in_table (wnd,table,"OK","sierp_okbutton",0,4,NULL);
	cancelbutton = Button_in_table (wnd,table,"Cancel","sierp_cancelbutton",1,4,NULL);

	gtk_signal_connect (GTK_OBJECT (wnd), "destroy",
                      GTK_SIGNAL_FUNC (on_sierp_cancelbutton_clicked), wnd);
	gtk_signal_connect (GTK_OBJECT (cancelbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_sierp_cancelbutton_clicked), wnd);
	if (fn)
		gtk_signal_connect (GTK_OBJECT (okbutton), "clicked",
        	              GTK_SIGNAL_FUNC (on_sierp_okbutton_clicked), wnd);
	else
		gtk_signal_connect (GTK_OBJECT (okbutton), "clicked",
        	              GTK_SIGNAL_FUNC (on_empty_okbutton_clicked), wnd);

	gtk_window_set_default_size (GTK_WINDOW(wnd), 240, -1);
	gtk_widget_show (wnd);

}



/**************************************************************

	Another variation on the fractal theme - several cellular
	automata (see e.g., Peitgen p.415ff)

***************************************************************/


/* Majority rule automaton */

int make_automaton_1 (image_type* img, int imgsize, int percent)
{
int x,y,itercnt,stable,n, changed, cy_cnt, cycle;
unsigned char *p1, *p2;
image_type temp;

	if (allocate_image (img, B_CHAR, imgsize,imgsize,1)==-1)
		fatal_memerr ();
	img->minval.imin=0; img->maxval.imax=255;
	p1 = img->data;
	if (clone_image (img, &temp) ==-1)
		fatal_memerr ();
	p2 = temp.data;

	/* Initialize the field with dense random noise. Note that we
		leave out the row of boundary pixels */

	for (y=1; y<img->ymax-1; y++)
		for (x=1; x<img->xmax-1; x++)
			if ((rand() % 1000)/10 < percent) p1[x + img->xmax*y] = 255;

	itercnt=0; stable=0; cy_cnt=0; cycle=0;
	while (!stable)
	{

		/* Run over the image, create next time step */

		for (y=1; y<img->ymax-1; y++)
			for (x=1; x<img->xmax-1; x++)
			{
				n = skel_numneigh (img,x,y);
				if (p1[x + img->xmax*y]) n++;
				if (n>=5) p2[x + img->xmax*y] = 255; else  p2[x + img->xmax*y] = 0;
			}

		/* Copy to original image, count # of changes */

		changed=0;
		for (y=1; y<img->ymax-1; y++)
			for (x=1; x<img->xmax-1; x++)
			{
				if (p1[x + img->xmax*y] != p2[x + img->xmax*y]) changed++;
				p1[x + img->xmax*y] = p2[x + img->xmax*y];
			}

		dp (1, "Automaton: Iteration %d, changes %d\n",itercnt,changed);
		if (changed==0) stable=1;
		itercnt++;

		if (cycle==changed) cy_cnt++; else cy_cnt=0;
		if (cy_cnt>100) stable=1;
		cycle=changed;

	}
	return cy_cnt;

}


/* One-in-eight rule automaton */

void make_automaton_2 (image_type* img, int imgsize)
{
int x,y,xm,ym,n, boundary_not_reached, itercnt;
unsigned char *p1, *p2;
image_type temp;


	if (allocate_image (img, B_CHAR, imgsize,imgsize,1)==-1)
		fatal_memerr ();
	img->minval.imin=0; img->maxval.imax=255;
	p1 = img->data;
	xm = img->xmax/2; ym = img->ymax/2;
	p1[xm + ym*img->xmax]=255;
	if (clone_image (img, &temp) ==-1)
		fatal_memerr ();
	p2 = temp.data;

	boundary_not_reached=1; itercnt=0;
	while ((boundary_not_reached) && (itercnt<xm+10))
	{
		/* Run over the image, create next time step.
		   Cells stay alive. New cells become alive if exactly
		   one neighbor is alive */

		for (y=1; y<img->ymax-1; y++)
			for (x=1; x<img->xmax-1; x++)
			{
				n = skel_numneigh (img,x,y);
				if (n==1)
				{
					p2[x + img->xmax*y] = 255;
					if ((x<=1) || (x>img->xmax-3)) boundary_not_reached=0;
					if ((y<=1) || (y>img->ymax-3)) boundary_not_reached=0;
				}
			}

		/* Copy to original image */

		for (y=1; y<img->ymax-1; y++)
			for (x=1; x<img->xmax-1; x++)
				p1[x + img->xmax*y] = p2[x + img->xmax*y];

		dp (1, "Automaton: Iteration %d done\n",++itercnt);
	}

}


/* Modulo-2 automaton (Sierpinski triangle automaton) */

int make_automaton_3 (image_type* img, int imgsize, int percent)
{
int x,y, xm,ym;
unsigned char *p1;
unsigned char a,b;
unsigned char *tmp;

	if (allocate_image (img, B_CHAR, imgsize,imgsize,1)==-1)
		fatal_memerr ();
	img->minval.imin=0; img->maxval.imax=255;
	p1 = img->data;

	/* We need to exceed the field to the right or the triangle will be crippled */

	xm = imgsize*2; ym=imgsize;
	tmp = calloc (xm*ym, sizeof (char));

	/* Initialize the top line with dense random noise.
		If noise probability is zero, we want a conventional triangle */

	if (percent <1)
		tmp[imgsize/2] = 255;		/* Top central pixel */
	else
	{
		for (x=0; x<img->xmax; x++)							/* Top row */
			if ((rand() % 1000)/10 < percent) tmp[x] = 255;
	}

	/* From top to bottom, set a pixel to the xor of the top diagonal neighbors */

	for (y=1; y<ym; y++)
		for (x=0; x<xm; x++)
		{
			if (x==0)
				a=0;
			else
				a=tmp[x-1 + xm*(y-1)];
			if (x==xm-1)
				b=0;
			else
				b=tmp[x+1 + xm*(y-1)];
			tmp[x + xm*y] = (a ^ b);
		}

	/* Now copy the relevant part of tmp into result image */

	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
			p1[x + img->xmax*y] = tmp[x + xm*y];

	free (tmp);

	return 0;

}





GtkAdjustment *automaton_percent_adj;
GtkAdjustment *automaton_imgsize_adj;
int automaton_function;



void on_automaton_cancelbutton_clicked  (GtkButton *button, gpointer user_data)
{

	gtk_widget_destroy (GTK_WIDGET (user_data));
}


void on_automaton_okbutton_clicked  (GtkButton *button, gpointer user_data)
{
int imgsize;
int percent;


	imgsize  = automaton_imgsize_adj->value;
	percent = automaton_percent_adj->value;

	if (automaton_function==0)
	{

		if (make_automaton_1 (&mainimg, imgsize, percent) > 100)
			statusbar_display ("Automaton: Periodic behavior suspected. Abort.");
		else
			statusbar_display ("Automaton finished");
	}
	else if (automaton_function==1)
	{
		make_automaton_2 (&mainimg, imgsize);
		statusbar_display ("Automaton finished");
	}
	else if (automaton_function==2)
	{
		make_automaton_3 (&mainimg, imgsize, percent);
		statusbar_display ("Automaton finished");
	}
	else if (automaton_function==3)
	{
		make_automaton_3 (&mainimg, imgsize, 0);
		statusbar_display ("Automaton finished");
	}
	generator_file_ok ("Cellular automaton");

	gtk_widget_destroy (GTK_WIDGET (user_data));
}


void automaton_set_menu (GtkWidget *widget, gpointer user_data)
{

	automaton_function = *(int*) user_data;

}


void make_automaton_gui ()
{
GtkWidget* wnd;
GtkWidget* vbox;
GtkWidget* table;
GtkWidget *label1, *label2, *label3;
GtkWidget *okbutton, *cancelbutton;
GtkWidget *automaton_imgsize, *automaton_percent;
GtkWidget *automaton_optionmenu1;
GtkWidget *automaton_optionmenu1_menu;
GtkWidget *glade_menuitem;


	wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (wnd), "wnd001", wnd);
	gtk_container_set_border_width (GTK_CONTAINER (wnd), 3);
	gtk_window_set_title (GTK_WINDOW (wnd), "Create cellular automaton patterns");
	gtk_window_set_modal (GTK_WINDOW (wnd), TRUE);
	vbox = vbox_in_window (wnd,"vbox001");
	table = MakeTable (wnd,vbox,"table001",2,4);
	label1 = MakeLabel (wnd,table,"Image size","label1",0,0);
	label2 = MakeLabel (wnd,table,"Automaton type","label2",0,1);
	label3 = MakeLabel (wnd,table,"Noise probability","label2",0,2);

	automaton_imgsize_adj = GTK_ADJUSTMENT (gtk_adjustment_new (512, 128, 2048, 1, 128, 0));
	automaton_imgsize = MakeHScale (automaton_imgsize_adj,wnd,table,"automaton_imgsize",1,0);
	gtk_scale_set_digits (GTK_SCALE (automaton_imgsize), 0);

	automaton_optionmenu1 = Make_Optionmenu (wnd, table,"automat_menu",1,1);
	gtk_container_set_border_width (GTK_CONTAINER (automaton_optionmenu1), 1);
	automaton_optionmenu1_menu = gtk_menu_new ();
	glade_menuitem = Option_menuitem (automaton_optionmenu1_menu, GTK_SIGNAL_FUNC(automaton_set_menu), "Majority", (gpointer) &menuopt0);
	glade_menuitem = Option_menuitem (automaton_optionmenu1_menu, GTK_SIGNAL_FUNC(automaton_set_menu), "One-in-eight", (gpointer) &menuopt1);
	glade_menuitem = Option_menuitem (automaton_optionmenu1_menu, GTK_SIGNAL_FUNC(automaton_set_menu), "Mod2 evolution", (gpointer) &menuopt2);
	glade_menuitem = Option_menuitem (automaton_optionmenu1_menu, GTK_SIGNAL_FUNC(automaton_set_menu), "Sierpinski triangle", (gpointer) &menuopt3);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (automaton_optionmenu1), automaton_optionmenu1_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (automaton_optionmenu1), 0);
	automaton_function=0;

	automaton_percent_adj = GTK_ADJUSTMENT (gtk_adjustment_new (50, 1, 99, 1, 10, 0));
	automaton_percent = MakeHScale (automaton_percent_adj,wnd,table,"automaton_percent",1,2);
	gtk_scale_set_digits (GTK_SCALE (automaton_percent), 0);


	okbutton = Button_in_table (wnd,table,"OK","automaton_okbutton",0,3,NULL);
	cancelbutton = Button_in_table (wnd,table,"Cancel","automaton_cancelbutton",1,3,NULL);

	gtk_signal_connect (GTK_OBJECT (wnd), "destroy",
                      GTK_SIGNAL_FUNC (on_automaton_cancelbutton_clicked), wnd);
	gtk_signal_connect (GTK_OBJECT (cancelbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_automaton_cancelbutton_clicked), wnd);
	gtk_signal_connect (GTK_OBJECT (okbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_automaton_okbutton_clicked), wnd);

	gtk_widget_show (wnd);

}




/**************************************************************

	Create a DLA cluster by simulation

***************************************************************/

#define BUF(X,Y)	    buf[(((X) + 2*imgsize*(Y)) > allocsize) ? 0 :(X) + 2*imgsize*(Y)]

void make_dla  (image_type* img, int imgsize, int numparts, int nb8, float omega)
{
int xm, ym;    		/* Mid of lattice */
int r, maxrad;  	/* SQR of max DLA aggregate radius */
int n, x, y, x1,y1;
int direction, walkcnt, nlost, updatecnt;
double phi,rad,sticky;
char arrived, lost;
float color;		/* to calculate the RGB color */
unsigned char* buf;
long allocsize;
hsvtriplet hsv; rgbtriplet rgb;

char xofs[8] = {1,0,-1,0,1,1,-1,-1};
char yofs[8] = {0,1,0,-1,-1,1,1,-1};


	if (allocate_image (img, B_RGB, imgsize,imgsize,1)==-1)
		fatal_memerr ();
	img->minval.imin=0; img->maxval.imax=255;

	buf = calloc (imgsize*imgsize*4, sizeof (char));	/* Our buffer is twice the size */
	if (buf==NULL) fatal_memerr ();
	allocsize = 4*SQR(imgsize);

	xm = imgsize; ym=imgsize; BUF(xm,ym)=1; /* Midpoint of buffer = seed */
	maxrad = 1; n=0; nlost=0;
	updatecnt=0;
	dp (1, "Generating DLA with %d particles, omega=%4.2f\n",numparts, omega);

	do                           /* number of aggregated particles loop */
	{
		/* determine initial point (x,y) */

		phi = 2*PI*( rand() & 0xff ) / 256.0;	/* Starting angle between 0 and 2 PI */
		rad = maxrad+20;
		x = (int) (0.48 + xm + (rad*cos(phi)));	/* Starting point on circle with radius rad */
		y = (int) (0.49 + ym + (rad*sin(phi)));

		/* perform random walk for this particle */

		arrived=0; lost=0; walkcnt=0;
		do
		{
			walkcnt++;

			/* With non-unity sticky probability, we need to find a non-occupied lattice position */

			do
			{
				if (nb8!=0)
					direction = rand() & 0x0007;
				else
					direction = rand() & 0x0003;
			}
			while (BUF(x+xofs[direction],y+yofs[direction]));
			x+=xofs[direction];
			y+=yofs[direction];

			if ((SQR(x-xm) + SQR(y-ym)) > SQR (maxrad+40)) lost=1;	/* Lost if too far out */
			if (walkcnt>50000) lost=1;								/* Lost after 50,000 futile steps */
			if ((x<5) || (y<5) || (x>imgsize*2-5) || (y>imgsize*2-5))
				lost = 1;											/* Lost if within 5 pixels of border */

			if (!lost)
			{
				/* Test for attachment, 4 or 8 neightbors */

				if ((BUF(x-1,y)>0) || (BUF(x+1,y)>0)
					|| (BUF(x,y-1)>0) || (BUF(x,y+1)>0)) arrived=1;

				if (nb8 )
				{
					if ((BUF(x-1,y-1)>0) || (BUF(x+1,y+1)>0)
						|| (BUF(x+1,y-1)>0) || (BUF(x-1,y+1)>0)) arrived=1;
				}

				/* If the particle touches the cluster, determine if it sticks.
				   See Mon KK, Transitions in the large-scale structure of aggregation models, Phys Rev A 1986; 34(5): 4469-4471. */

				if ((omega>0) && (arrived))
				{
					sticky = 1.0 - pow ((float)n/(float)numparts, omega);	/* S(N) = 1-N^Omega */
					sticky = 1.0/pow ((float)n, omega);		/* KK Mon original: S(N) = (1/N)^Omega */

					if (sticky < 0.01) sticky = 0.01;	/* be reasonable. For omega->0 the process takes FOREVER */

					if ( (rand() & 0xff)/256.0 > sticky)
					{
						arrived=0;	/* Move along ... */
					}
				}

				if (arrived==1)
				{

					BUF(x,y)  = 1;		/* Set in main lattice */
					r = SQR (x-xm) + SQR(y-ym);
					if (r > SQR(maxrad)) maxrad = sqrt(r);

					x1 = (x-xm)+imgsize/2;	/* Pixel coordinates in image */
					y1 = (y-ym)+imgsize/2;
					if ((x1>=0) && (y1>=0) && (x1<imgsize) && (y1<imgsize))	/* If within image bounds only */
					{
						color=(300.0*n)/numparts;	/* To color by time of arrival */
						/* color = (100.0*walkcnt)/10000; to color by length of random walk */

						hsv.hue = color;
						hsv.sat = 255;
						hsv.val = 255;
						rgb = hsv_to_rgb (hsv);

						((rgbtriplet*)img->data)[x1+imgsize*y1] = rgb;
					}
				}
			}
		}
		while ((lost==0) && (arrived==0));
		if (arrived) n++;
		if (lost) nlost++;

	}
	while ((n<numparts) && (nlost < 100*numparts));
	free (buf);

	dp (1, "Done generating DLA. %d particles attached. %d particles lost.\n",n,nlost);

	/* Enable the following to draw exponentially expanding circles

	rad=10; rd=255; gn=255; bl=255;
	while (rad<200)
	{
		for (phi=0; phi<360.0; phi+=0.05)
		{

			x = (int) (0.48 + xm + (rad*cos(phi)));
			y = (int) (0.49 + ym + (rad*sin(phi)));
			((rgbtriplet*)img->data)[x+imgsize*y] = (rgbtriplet){rd,gn,bl};
		}

		if (rad>=171) rad=239;
		else if (rad>=122) rad=171;
		else if (rad>=87) rad=122;
		else if (rad>=62) rad=87;
		else if (rad>=44) rad=62;
		else if (rad>=31) rad=44;
		else if (rad>=22) rad=31;
		else if (rad>=15) rad=22;
		else rad=15;
	}
	*/

}

#undef BUF


int dla_8ngh;
GtkAdjustment *dla_imgsize_adj, *dla_numparts_adj, *dla_omega_adj;


void on_dla_neigh_clicked  (GtkButton *button, gpointer user_data)
{
	dla_8ngh = !dla_8ngh;
}


void on_dla_cancelbutton_clicked  (GtkButton *button, gpointer user_data)
{

	gtk_widget_destroy (GTK_WIDGET (user_data));
}


void on_dla_okbutton_clicked  (GtkButton *button, gpointer user_data)
{
int imgsize, numparts;
float omega;


	imgsize  = dla_imgsize_adj->value;
	numparts = dla_numparts_adj->value;
	omega = dla_omega_adj->value;

	make_dla (&mainimg, imgsize, numparts, dla_8ngh, omega);
	statusbar_display ("DLA generation finished");
	generator_file_ok ("DLA");

	gtk_widget_destroy (GTK_WIDGET (user_data));
}


void make_dla_gui ()
{
GtkWidget* wnd;
GtkWidget* vbox;
GtkWidget* table;
GtkWidget *label1, *label2, *label3, *label4;
GtkWidget *dla_imgsize, *dla_numparts, *dla_omega;
GtkWidget *okbutton, *cancelbutton;
GtkWidget *neigh_checkbutton;

	dla_8ngh=0;
	wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (wnd), "wnd001", wnd);
	gtk_container_set_border_width (GTK_CONTAINER (wnd), 3);
	gtk_window_set_title (GTK_WINDOW (wnd), "Create DLA cluster");
	gtk_window_set_modal (GTK_WINDOW (wnd), TRUE);
	vbox = vbox_in_window (wnd,"vbox001");
	table = MakeTable (wnd,vbox,"table001",2,5);
	label1 = MakeLabel (wnd,table,"Image size","label1",0,0);
	label2 = MakeLabel (wnd,table,"Number of particles","label2",0,1);
	label3 = MakeLabel (wnd,table,"Neighborhood","label3",0,3);
	label4 = MakeLabel (wnd,table,"Omega","label3",0,2);

	dla_imgsize_adj = GTK_ADJUSTMENT (gtk_adjustment_new (512, 128, 2048, 1, 128, 0));
	dla_imgsize = MakeHScale (dla_imgsize_adj,wnd,table,"dla_imgsize",1,0);
	gtk_scale_set_digits (GTK_SCALE (dla_imgsize), 0);

	dla_numparts_adj = GTK_ADJUSTMENT (gtk_adjustment_new (5000,100,20000, 1, 100, 0));
	dla_numparts = MakeHScale (dla_numparts_adj,wnd,table,"dla_numparts",1,1);
	gtk_scale_set_digits (GTK_SCALE (dla_numparts), 0);

	dla_omega_adj = GTK_ADJUSTMENT (gtk_adjustment_new (0.0, 0.0, 1.0, 0.01, 0.1, 0));
	dla_omega = MakeHScale (dla_omega_adj,wnd,table,"dla_omega",1,2);
	gtk_scale_set_digits (GTK_SCALE (dla_omega), 2);

	neigh_checkbutton = Checkbutton_in_table (wnd,table," 8-neighborhood ","dla_chkbutton",1,3,on_dla_neigh_clicked);
	okbutton = Button_in_table (wnd,table,"OK","dla_okbutton",0,4,NULL);
	cancelbutton = Button_in_table (wnd,table,"Cancel","dla_cancelbutton",1,4,NULL);

	gtk_signal_connect (GTK_OBJECT (wnd), "destroy",
                      GTK_SIGNAL_FUNC (on_dla_cancelbutton_clicked), wnd);
	gtk_signal_connect (GTK_OBJECT (cancelbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_dla_cancelbutton_clicked), wnd);
	gtk_signal_connect (GTK_OBJECT (okbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_dla_okbutton_clicked), wnd);

	gtk_widget_show (wnd);

}


/* Macro interpreter hookup. The syntax is as diverse as the module

	Each macro call starts with

		imgvar = GENERATE ("gentype", imgsize...

	The remaining parameters depend on "gentype" as follows:

		"sierp": 	Sierpinsky triangle, (no more parameters)
		"psyrings": Psychedelic rings, 				...frequency);
		"dla": 		DLA cluster 					...numparticles, neighborhood, omega);
		"majority": Majority rule automaton			... percent_seed);
		"oneineight":	One-in-eight automaton (no more parameters)
		"noise": Stack of Gaussian noise			...zsize, sd [,H]);
		"headphantom": Shepp-Logan head phantom		...contrastflg);

	Not included in the GUI, but available through macros:

		"oneover":	1/r function (PSF of unfiltered backprojection)



*/


variable_type mac_generator (unsigned char* s, int* i)
{
variable_type v,r;
int j, pa, pb;
float ra, rb;
char gentype[25];
int imgsize;

	j = *i;

	r = expr (s, &j, MT_STRING);		/* extract the type of generator to be used */
	strncpy (gentype, r.strg, 24);
	free (r.strg);
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); imgsize = r.intvalue;	/* Next one MUST be the image size */

	v.type = MT_IMG;		/* We will always return an image */
	pa=0; pb=0; ra=0; rb=0;	/* silence a compiler warning */

	if (!strncasecmp (gentype, "sierp",5))
	{
		make_sierp_2d (&v.imgptr, imgsize);
	}
	else if (!strncasecmp (gentype, "psyrings",8))
	{
		if (s[j++]!=',') macro_error (1,j);
		r = expr (s,&j, MT_INT); pa = r.intvalue;	/* Next one MUST be frequency */
		make_psy_rings (&v.imgptr, imgsize, pa);
	}
	else if (!strncasecmp (gentype, "dla",3))
	{
		if (s[j++]!=',') macro_error (1,j);
		r = expr (s,&j, MT_INT); pa = r.intvalue;	/* Next one MUST be number of particles */
		if (s[j++]!=',') macro_error (1,j);
		r = expr (s,&j, MT_INT); pb = r.intvalue;	/* Next one MUST be neighborhood flag */
		if (s[j++]!=',') macro_error (1,j);
		r = expr (s,&j, MT_REAL); rb = r.numvalue;	/* Next one MUST be omega */
		make_dla (&v.imgptr, imgsize, pa, pb, rb);
	}
	else if (!strncasecmp (gentype, "major",5))
	{
		if (s[j++]!=',') macro_error (1,j);
		r = expr (s,&j, MT_REAL); ra = r.numvalue;	/* Next one MUST be seed percentage */
		make_automaton_1 (&v.imgptr, imgsize,ra);
	}
	else if (!strncasecmp (gentype, "onein",5))
	{
		make_automaton_2 (&v.imgptr, imgsize);
	}
	else if (!strncasecmp (gentype, "noise",5))
	{
		if (s[j++]!=',') macro_error (1,j);
		r = expr (s,&j, MT_INT); pa = r.intvalue;	/* Next one MUST be number of Z slices */
		if (s[j++]!=',') macro_error (1,j);
		r = expr (s,&j, MT_INT); pb = r.intvalue;	/* Next one MUST be noise type [1..7] */
		if (pb>2)
		{
			if (s[j++]!=',') macro_error (1,j);
			r = expr (s,&j, MT_REAL); ra = r.numvalue;	/* If Gaussian or FBM, specify SD */
		}
		if (pb>3)
		{
			if (s[j++]!=',') macro_error (1,j);
			r = expr (s,&j, MT_REAL); rb = r.numvalue;	/* If FBM, specify H as well. */
		}


		make_noise (&v.imgptr, imgsize, pa /* Z */, pb /* Type */, ra /* SD */, rb /* H */ );
	}
	else if (!strncasecmp (gentype, "head",4))
	{
		if (s[j++]!=',') macro_error (1,j);
		r = expr (s,&j, MT_INT); pa = r.intvalue;	/* Next one MUST be high-contrast flag */
		make_headphantom (&v.imgptr, imgsize, pa, 0);
	}
	else if (!strncasecmp (gentype, "empty",5))
	{
		if (allocate_image (&v.imgptr, B_CHAR, imgsize, imgsize,1)==-1)
			fatal_memerr();
	}
	else if (!strncasecmp (gentype, "oneover", 7))
	{
		make_spec_function (&v.imgptr, imgsize, 0);
	}
	else macro_error (15,j);			/* TODO: We haven't covered them all yet */



	while (s[j]!=')') j++;
	*i = j+1;

	make_minmax (&v.imgptr);


	return v;

}






/**************************************************************

	This is the infrastructure we use to register with the
	main program. The pointer "menufunction" points to the
	main entry (the function to be called when the menu is
	activated). Initmodule is the one function that the
	main program calls (the name is fixed!) to initialize
	the module and to register the module with the menu.

***************************************************************/

void (*menufunction1)();
void (*menufunction2)();
void (*menufunction3)();
void (*menufunction4)();
void (*menufunction5)();
void (*menufunction6)();
void (*menufunction7)();
void (*menufunction8)();
void (*menufunction9)();
void (*menufunction10)();
variable_type (*macrofunction)(unsigned char* s, int* i);


void generator_make_sierp()
{

	dp (2, "Generator module activated\n");
	if (file_ok)
		ShowMessage ("Close main image first");
	else
		make_sierp_gui (1);
}

void generator_make_empty()
{

	dp (2, "Generator module activated\n");
	if (file_ok)
		ShowMessage ("Close main image first");
	else
		make_sierp_gui (0);
}

void generator_make_psyrings()
{

	dp (2, "Generator module activated\n");
	if (file_ok)
		ShowMessage ("Close main image first");
	else
		make_psyrings_gui ();
}


void generator_make_dla ()
{

	dp (2, "Generator module activated\n");
	if (file_ok)
		ShowMessage ("Close main image first");
	else
		make_dla_gui ();
}

void generator_make_automaton ()
{

	dp (2, "Generator module activated\n");
	if (file_ok)
		ShowMessage ("Close main image first");
	else
		make_automaton_gui ();
}

void generator_make_noise ()
{

	dp (2, "Generator module activated\n");
	if (file_ok)
		ShowMessage ("Close main image first");
	else
		make_noise_gui ();
}

void generator_make_fftmask ()
{

	dp (2, "Generator module activated\n");
	if (file_ok)
		ShowMessage ("Close main image first");
	else
		mkfftmask_gui ();
}

void generator_make_headphant ()
{

	dp (2, "Generator module activated\n");
	if (file_ok)
		ShowMessage ("Close main image first");
	else
		make_headphantom_gui ();
}


void generator_make_sinexy ()
{

	dp (2, "Generator module activated\n");
	if (file_ok)
		ShowMessage ("Close main image first");
	else
		make_sinexy_gui ();
}



void generator_make_grid()
{

	dp (2, "Generator module activated\n");
	if (file_ok)
		ShowMessage ("Close main image first");
	else
		make_grid_gui ();
}



void initmodule ()
{
GtkWidget* menu;

	dp (1, "Generator module: Initializer called\n");

	/* Register this module with the main menu */

	menufunction1 = &generator_make_sierp;
	menufunction2 = &generator_make_dla;
	menufunction3 = &generator_make_automaton;
	menufunction4 = &generator_make_noise;
	menufunction5 = &generator_make_fftmask;
	menufunction6 = &generator_make_psyrings;
	menufunction7 = &generator_make_headphant;
	menufunction8 = &generator_make_sinexy;
	menufunction9 = &generator_make_empty;
	menufunction10 = &generator_make_grid;
	menu = dso_menu_register_with_submenu ("Generate");
	dso_submenu_register (menu, "Empty Image", menufunction9);
	dso_submenu_register (menu, "Sierpinsky triangle", menufunction1);
	dso_submenu_register (menu, "DLA cluster", menufunction2);
	dso_submenu_register (menu, "Cellular Automata", menufunction3);
	dso_submenu_register (menu, "3D noise", menufunction4);
	dso_submenu_register (menu, "FFT filter masks", menufunction5);
	dso_submenu_register (menu, "Psychedelic rings", menufunction6);
	dso_submenu_register (menu, "Shepp-Logan Head Phantom", menufunction7);
	dso_submenu_register (menu, "Sinusoids", menufunction8);
	dso_submenu_register (menu, "Rectangular Grid", menufunction10);

	/* Register the module with the macro interpreter  */

	macrofunction = &mac_generator;
	register_dso_macro ("GENERATE",5,MT_IMG,macrofunction);



}

