

/* fileio.c:
	High-level file related functions
	including the file save GUI
	This file is new in rev 1.0 and contains parts of what used to be in tiffstuff.c
*/



/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */




#include "defs.h"
#include "interface.h"
#include "functions.h"
#include "application.h"
#include "dlgbox.h"
#include "gtkhelpers.h"
#include "tiffstuff.h"





/************************************************************************************

	In addition to is_tiff and is_vsrt, this function get_filetype tries to
	obtain the type of file by its header info. It encompasses is_tiff and
	is_vsrt.


************************************************************************************/

char get_filetype (const char* filename)
{
FILE* FP;
char hdr[256];
int numelems;
char ftype;

	FP = fopen (filename, "rb");
	if (!FP) return 0;
	
	numelems = fread (&hdr, sizeof (char), sizeof (hdr)-1, FP);
	fclose (FP);
	if (numelems < 128) return -1;		/* If this file is *that* short, it's not likely an image file */
	
	if (!strncasecmp (hdr, "VSRT", 4)) 
		ftype= FTYPE_VSRT;
	else if (!strncasecmp (hdr, "VFLT", 4)) 
		ftype= FTYPE_VFLT;
	else if (!strncmp (hdr, "II", 2)) 
		ftype= FTYPE_TIFFI; 			/* Little-endian, (II for Intel) */
	else if (!strncmp (hdr, "MM", 2)) 
		ftype= FTYPE_TIFFM;				/* Big-endian (MM for Motorola) */
	else if (!strncmp (hdr+0x80, "DICM", 4)) 
		ftype= FTYPE_DCM;				/* Dicom file has code at offset 128 */
	else if (!strncmp (hdr, "P5", 2)) 
		ftype= FTYPE_PGM;				/* PGM - portable graymap. One-byte binary grayscale */
	else if (!strncmp (hdr, "P6", 2)) 
		ftype= FTYPE_PPM;				/* PPM - portable bytemap. 24-bit RGB */
	else
		ftype=0;

	dp (1, "Probed file type. Type is %d ",ftype);
	if (ftype < NFILETYPES) dp (1, "(%s)", filetype_names[(unsigned char)ftype]);
	dp (1, "\n");


	return ftype;
	
}







/************************************************************************************

	File save dialog
	(moved here in rev. 1.0)
	
	
************************************************************************************/

int save_type, compression_mode;
GtkAdjustment *savedlg_startslice_adj;
GtkAdjustment *savedlg_endslice_adj;
GtkAdjustment *savedlg_zscale_adj;
GtkWidget *savedlg_filename;
char savefilename[2048];
image_type* saveimg;
GtkWidget *savedlg_label2a, *savedlg_label2b;
GtkWidget *savedlg_zscale;
GtkWidget *savedlg_optionmenu2;



void on_savedlg_browse_clicked (GtkButton *button, gpointer user_data)
{
gchar* selectedfile; 

	ChooseFile ("Filename to save", window1, 1, &selectedfile);
	if (selectedfile)
	{
		strcpy (savefilename, selectedfile);
		gtk_entry_set_text (GTK_ENTRY(savedlg_filename), savefilename); 
	}
}


void on_savedlg_ok_clicked (GtkButton *button, gpointer user_data)
{
gchar* savefile;
int err;
int stslice, enslice;
float zscale;

	savefile = gtk_entry_get_text (GTK_ENTRY(savedlg_filename));
	err=0;		/* prevent a warning */
	stslice = (int) (0.5+savedlg_startslice_adj->value);
	savedlg_startslice_adj->value = stslice;
	enslice = (int) (0.5+savedlg_endslice_adj->value);
	if (enslice<stslice) enslice = stslice;
	savedlg_endslice_adj->value = enslice;
	zscale = savedlg_zscale_adj->value;
	
	if (save_type == 0)		/* TIFF */
	{
		dp (1, "Saving %s as TIFF, compression %d\n",savefile,compression_mode);
		err=tiffwrite (savefile, saveimg, compression_mode, "Made with Crystal Image 1.xx (C) M. Haidekker", stslice, enslice);
	}
	else if (save_type==1)		/* RAW */
	{
		dp (1, "Saving %s as RAW\n",savefile);
		err=raw_write (saveimg, savefile, stslice, enslice);
	}
	else if (save_type==2)		/* VSRT */
	{
		dp (1, "Saving %s as VSRT\n",savefile);
		err=vsrtwrite (saveimg, savefile, stslice, enslice);
	}
	else if (save_type==3)		/* OpenDX */
	{
		dp (1, "Saving %s as DX native\n",savefile);
		err=DX_write (saveimg, savefile, stslice, enslice, zscale);
	}
	else if (save_type==4)		/* PGM and PPM */
	{
		dp (1, "Saving %s as Portable Graymap PGM\n",savefile);
		if (stslice != enslice)
			ShowMessage ("Error: PGM cannot save stacks");
		else
			err=pnm_write (saveimg, savefile, stslice, enslice);
	}
	else if (save_type==5)		/* RAW ASCII */
	{
		dp (1, "Saving %s as raw ASCII file\n",savefile);
		err=ascii_write (saveimg, savefile, stslice, enslice);
	}
	
	
	if (err==0)
		statusbar_display ("File has been saved");
	else 
		statusbar_display ("File save possibly failed");
	set_display_title (savefile);

	gtk_widget_destroy (GTK_WIDGET(user_data));
}


void on_savedlg_cancel_clicked (GtkButton *button, gpointer user_data)
{
	gtk_widget_destroy (GTK_WIDGET(user_data));
}

void savedlg_set_type_menu (GtkWidget *widget, gpointer user_data)
{
	save_type = *(int*) user_data;
	if (save_type == 0)
	{
		gtk_widget_show (savedlg_label2a);
		gtk_widget_show (savedlg_optionmenu2);
		gtk_widget_hide (savedlg_label2b);
		gtk_widget_hide (savedlg_zscale);
	}
	else if (save_type == 3)
	{
		gtk_widget_hide (savedlg_label2a);
		gtk_widget_hide (savedlg_optionmenu2);
		gtk_widget_show (savedlg_label2b);
		gtk_widget_show (savedlg_zscale);
	}
	else
	{
		gtk_widget_hide (savedlg_label2a);
		gtk_widget_hide (savedlg_optionmenu2);
		gtk_widget_hide (savedlg_label2b);
		gtk_widget_hide (savedlg_zscale);
	}
	
}

void savedlg_set_compression_menu (GtkWidget *widget, gpointer user_data)
{
	compression_mode = *(int*) user_data;
}



void create_savedlg_wnd (image_type* img)
{
  GtkWidget *savedlg_wnd;
  GtkWidget *savedlg_table1;
  GtkWidget *savedlg_label1;
  GtkWidget *savedlg_optionmenu1;
  GtkWidget *savedlg_optionmenu1_menu;
  GtkWidget *glade_menuitem;
  GtkWidget *savedlg_browse;
  GtkWidget *savedlg_optionmenu2_menu;
  GtkWidget *savedlg_label3;
  GtkWidget *savedlg_label4;
  GtkWidget *savedlg_startslice;
  GtkWidget *savedlg_endslice;
  GtkWidget *savedlg_okbutton;
  GtkWidget *savedlg_cancel;

	savedlg_wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (savedlg_wnd), "savedlg_wnd", savedlg_wnd);
	gtk_window_set_title (GTK_WINDOW (savedlg_wnd), "Save file as");
	gtk_window_set_modal (GTK_WINDOW (savedlg_wnd), TRUE);

	savedlg_table1 = gtk_table_new (7, 2, FALSE);
	gtk_widget_ref (savedlg_table1);
	gtk_object_set_data_full (GTK_OBJECT (savedlg_wnd), "savedlg_table1", savedlg_table1,
                              (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (savedlg_table1);
	gtk_container_add (GTK_CONTAINER (savedlg_wnd), savedlg_table1);

	savedlg_label1 = MakeLabel (savedlg_wnd,savedlg_table1,"File Format","savedlg_label1",0,0);
	savedlg_label2b = MakeLabel (savedlg_wnd,savedlg_table1,"Z anisotropy","savedlg_label2b",0,3);
	gtk_widget_hide (savedlg_label2b);
	savedlg_label2a = MakeLabel (savedlg_wnd,savedlg_table1,"TIFF compression","savedlg_label2a",0,2);
	savedlg_label3 = MakeLabel (savedlg_wnd,savedlg_table1,"Start Slice","savedlg_label3",0,4);
	savedlg_label4 = MakeLabel (savedlg_wnd,savedlg_table1,"End Slice","savedlg_label3",0,5);

	savedlg_browse = Button_in_table (savedlg_wnd,savedlg_table1,"Browse Filename","savedlg_browse",0,1,on_savedlg_browse_clicked);
	savedlg_okbutton = Button_in_table (savedlg_wnd,savedlg_table1,"Save","savedlg_okbutton",0,6,on_savedlg_ok_clicked);
	savedlg_cancel = Button_in_table (savedlg_wnd,savedlg_table1,"Cancel","savedlg_cancel",1,6,on_savedlg_cancel_clicked);


	savedlg_optionmenu1 = gtk_option_menu_new ();
	gtk_widget_ref (savedlg_optionmenu1);
	gtk_object_set_data_full (GTK_OBJECT (savedlg_wnd), "savedlg_optionmenu1", savedlg_optionmenu1,
                              (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (savedlg_optionmenu1);
	gtk_table_attach (GTK_TABLE (savedlg_table1), savedlg_optionmenu1, 1, 2, 0, 1,
                      (GtkAttachOptions) (0), (GtkAttachOptions) (0), 3, 3);
	gtk_container_set_border_width (GTK_CONTAINER (savedlg_optionmenu1), 6);

	savedlg_optionmenu1_menu = gtk_menu_new ();
	glade_menuitem = Option_menuitem (savedlg_optionmenu1_menu,savedlg_set_type_menu,"TIFF",&menuopt0);
	glade_menuitem = Option_menuitem (savedlg_optionmenu1_menu,savedlg_set_type_menu,"RAW",&menuopt1);
	glade_menuitem = Option_menuitem (savedlg_optionmenu1_menu,savedlg_set_type_menu,"RAW ASCII",&menuopt5);
	if ((img->imgtype == B_SHORT) || (img->imgtype == B_FLOAT))
  	  glade_menuitem = Option_menuitem (savedlg_optionmenu1_menu,savedlg_set_type_menu,"VSRT/VFLT",&menuopt2);
	if (img->imgtype != B_RGB)
  	  glade_menuitem = Option_menuitem (savedlg_optionmenu1_menu,savedlg_set_type_menu,"DX",&menuopt3);
	if (img->imgtype == B_CHAR)
  	  glade_menuitem = Option_menuitem (savedlg_optionmenu1_menu,savedlg_set_type_menu,"PGM",&menuopt4);
	else if (img->imgtype == B_RGB)
  	  glade_menuitem = Option_menuitem (savedlg_optionmenu1_menu,savedlg_set_type_menu,"PPM",&menuopt4);

	gtk_option_menu_set_menu (GTK_OPTION_MENU (savedlg_optionmenu1), savedlg_optionmenu1_menu);

	savedlg_optionmenu2 = gtk_option_menu_new ();
	gtk_widget_ref (savedlg_optionmenu2);
	gtk_object_set_data_full (GTK_OBJECT (savedlg_wnd), "savedlg_optionmenu2", savedlg_optionmenu2,
                              (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (savedlg_optionmenu2);
	gtk_table_attach (GTK_TABLE (savedlg_table1), savedlg_optionmenu2, 1, 2, 2, 3,
                      (GtkAttachOptions) (0),
                      (GtkAttachOptions) (0), 0, 0);
	gtk_container_set_border_width (GTK_CONTAINER (savedlg_optionmenu2), 6);

	savedlg_optionmenu2_menu = gtk_menu_new ();

	/* The choice of TIFF compression schemes actually depends on the image type we have */

	switch (img->imgtype)
	{
		case B_CHAR:
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"NONE",&menuopt0);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"LZW",&menuopt1);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"Packbits",&menuopt2);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"CCITT FAX 3",&menuopt3);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"CCITT FAX 4",&menuopt4);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"JPEG",&menuopt5);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"CCITT RLE",&menuopt6);
			break;
  		case B_SHORT:
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"NONE",&menuopt0);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"LZW",&menuopt1);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"Packbits",&menuopt2);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"JPEG",&menuopt5);
			break;
  		case B_FLOAT:
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"NONE",&menuopt0);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"LZW",&menuopt1);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"Packbits",&menuopt2);
			break;
  		case B_RGB:
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"NONE",&menuopt0);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"LZW",&menuopt1);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"Packbits",&menuopt2);
  			glade_menuitem = Option_menuitem (savedlg_optionmenu2_menu,savedlg_set_compression_menu,"JPEG",&menuopt5);
			break;
	}		

	gtk_option_menu_set_menu (GTK_OPTION_MENU (savedlg_optionmenu2), savedlg_optionmenu2_menu);

	savedlg_filename = gtk_entry_new ();
	gtk_widget_ref (savedlg_filename);
	gtk_object_set_data_full (GTK_OBJECT (savedlg_wnd), "savedlg_filename", savedlg_filename,
                              (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (savedlg_filename);
	gtk_table_attach (GTK_TABLE (savedlg_table1), savedlg_filename, 1, 2, 1, 2,
                      (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                      (GtkAttachOptions) (0), 6, 6);

	savedlg_zscale_adj = GTK_ADJUSTMENT(gtk_adjustment_new (1, 0, 25, 0.1, 1, 0));
	savedlg_zscale = MakeHScale (savedlg_zscale_adj, savedlg_wnd, savedlg_table1,
  									  "savedlg_zscale", 1,3);
	gtk_widget_hide (savedlg_zscale);

	savedlg_startslice_adj = GTK_ADJUSTMENT(gtk_adjustment_new (0, 0, img->zmax-1, 1, 10, 0));
	savedlg_startslice = MakeHScale (savedlg_startslice_adj, savedlg_wnd, savedlg_table1,
  									  "savedlg_startslice", 1,4);

	savedlg_endslice_adj = GTK_ADJUSTMENT(gtk_adjustment_new (img->zmax-1, 0, img->zmax-1, 1, 10, 0));
	savedlg_endslice = MakeHScale (savedlg_endslice_adj, savedlg_wnd, savedlg_table1,
  									  "savedlg_startslice", 1,5);

	gtk_scale_set_digits (GTK_SCALE (savedlg_startslice), 0);
	gtk_scale_set_digits (GTK_SCALE (savedlg_endslice), 0);

	gtk_signal_connect (GTK_OBJECT (savedlg_wnd), "destroy",
                    	GTK_SIGNAL_FUNC (on_savedlg_cancel_clicked),savedlg_wnd);

	if (mainimg_name[0]) gtk_entry_set_text (GTK_ENTRY(savedlg_filename), mainimg_name);

	gtk_widget_show ( savedlg_wnd);
}


void save_main (image_type* img)
{

	save_type=0;
	compression_mode=0;
	saveimg = img;
	
	create_savedlg_wnd (img);
	
}


/*******************************************************************************/


/* A function to save an image to a temporary file.
	We assume that a path to a temp directory is geven (undo_path)
	and the file name is composed of "Img_undo_", the instance's PID and a 3-digit counter.
	The image file is stored in TIFF format.

	A helper function generates the temp file name; this can be used for other purposes as well.
	Each call of that helper function increments undo_cnt so that (combined with the PID)
	the names are unique, yet systematic (somewhat akin to 'mktemp')
*/



gchar *create_tempfn ()
{
gchar *tempnam;
gchar *fn;
int i;


	i = strlen(undo_path) + 35;
	tempnam = calloc (i, sizeof (gchar));
	fn = calloc (i, sizeof (gchar));

	sprintf (tempnam, "Img_undo_%04x_%03d.tmp", cimage_pid, undo_cnt);
	undo_cnt++; if (undo_cnt>999) undo_cnt=0;

	if (strlen (undo_path)>0)
	{
		strcpy (fn, undo_path);
		if (fn[strlen(fn)-1]!='/') strcat (fn, "/");
		strcat (fn, tempnam);
	}
	else
		strcpy (fn, tempnam);

	free (tempnam);
	return fn;			/* Returns the pointer; freeing needs to be done by the caller */

}






int store_temp_file (image_type *img, const gchar *tempfilename)
{
int err, z;


	/* Save secondary image in temporary image path */

	dp (1, "Temp file: Saving as %s\n", tempfilename);

	z = 0; if (img->zmax>1) z = img->zmax-1;

	err=tiffwrite (tempfilename, img, 2, "cimage TEMP file", 0, z);
	if (err < 0)
	{
		dp (0, "Error saving temporary file %s\n", tempfilename);
		statusbar_display ("Error saving temporary file");
		return -1;
	}

	return 0;

}






/*******************************************************************************/


