
/*
 * Misc Op: Miscellaneous operators
 */
 

/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */

 
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>


#include "application.h"
#include "defs.h"
#include "dlgbox.h"
#include "functions.h"
#include "functions2.h"
#include "gtkhelpers.h"
#include "imgbuf.h"
#include "interface.h"
#include "miscop.h"
#include "randgen.h"
#include "roi.h"
#include "viz.h"







/*************************************************************

	Rotate: Rotate an image by theta degrees
	
**************************************************************/

/* Original and rotated image have the same dimensions.
   whatever does not fit will be cropped. Undefined areas
   are set to zero. 
   THETA IS GIVEN IN DEGREES! 
   
   If rotation angles are between 45 and 135� or between 
   225 and 315 degrees, we flip image's X and Y extensions
   
*/


void rotate_img (image_type* img, double theta)
{
double st, ct;	/* cosine and sine of theta */
int x,y,z;		/* target image coords */
double x1,y1; 		/* Original image coords */
double xh,yh;	/* Half-image coordinates */
float* p;
rgbtriplet* prgb;
long i, idx, zofs;
int new_xmax, new_ymax;


	p = (float*) calloc (img->xmax * img->ymax, sizeof (float));
	if (!p) fatal_memerr();
	prgb = (rgbtriplet*)p; /* !!! This assumes that sizeof (float) > sizeof (rgbtriplet) ...or SIGSEGV */

	new_xmax = img->xmax; 
	new_ymax = img->ymax; 
	if ( ((fabs(theta)>45) && (fabs(theta)<135)) || ((fabs(theta)>225) && (fabs(theta)<315)) ) 
	{
		new_xmax = img->ymax;
		new_ymax = img->xmax;
	}

	st = sin(PI*theta/180.0);
	ct = cos(PI*theta/180.0);
	xh = new_xmax/2.0;
	yh = new_ymax/2.0;

	for (z=0; z<img->zmax; z++)		/* slice by slice */
	{
		zofs = new_xmax * new_ymax * z;
		dp (1, "Rotating slice %d at offset %li\n", z, zofs);
		for (y=0; y < new_ymax; y++)
		{
			idx = y*new_xmax;
			for (x=0; x < new_xmax; x++)
			{
				x1 = (x-xh)*ct + (y-yh)*st + img->xmax/2.0;
				y1 = -(x-xh)*st + (y-yh)*ct + img->ymax/2.0;
				if ((y1<0) || (x1<0) || (x1>=img->xmax) || (y1>=img->ymax))
				{
					if (img->imgtype==B_RGB)
						prgb[idx+x]=(rgbtriplet){0,0,0};
					else
						p[idx+x]=0;
				}
				else
				{
					if (img->imgtype==B_RGB)
						prgb [idx+x] = rgbireadbuf (img, x1,y1,z, intp_pref);
					else
						p [idx+x] = ireadbuf (img, x1,y1,z, intp_pref);
				}
			}
		}
		
		/* copy the slice back into the original image */
		
		switch (img->imgtype)
		{
			case B_CHAR:
			{
				unsigned char* p1; float buf;
				p1 = img->data;
				for (i=0; i < new_xmax * new_ymax; i++)
				{
					buf = IROUND(p[i]); 
					if (buf<0) buf=0; else if (buf>255) buf=255;
					p1[i+zofs] = (int)buf;
				}
				break;
			}
			case B_SHORT:
			{
				short* p1;
				p1 = img->data;
				for (i=0; i < new_xmax * new_ymax; i++)
					p1[i+zofs] = IROUND(p[i]);
				break;
			}
			case B_FLOAT:	/* Beauty! We don't have to round here!!! */
			{
				float* p1;
				p1 = img->data;
				for (i=0; i < new_xmax * new_ymax; i++)
					p1[i+zofs] = p[i];
				break;
			}						
			case B_RGB:	
			{
				rgbtriplet* p1;
				p1 = img->data;
				for (i=0; i < new_xmax * new_ymax; i++)
					p1[i+zofs] = prgb[i];
				break;
			}
		}
	}
	
	img->xmax = new_xmax; img->ymax = new_ymax;
	img->uid = next_uid++;
	free (p);
}


/* Rescale image, make its new dimensions (xnew, ynew).
	This is a 2D operation and is executed slice by slice. */


void zoom2D (image_type* img, int xnew, int ynew)
{
image_type tempimg;
int x,y,z, ival;
float dx,dy,fy,fx;
long idx;



	if (allocate_image (&tempimg, img->imgtype, xnew,ynew,img->zmax)==-1)
		fatal_memerr();
	
	/* interpolation step size */
	
	dx = (float)img->xmax / (float)xnew;
	dy = (float)img->ymax / (float)ynew;
	
	/* Now transfer. Go via float, because we have to interpolate anyway */
	
	idx=0;
	for (z=0; z<tempimg.zmax; z++)
	{
		for (y=0, fy=0; y<tempimg.ymax; y++, fy+=dy)
			for (x=0, fx=0; x<tempimg.xmax; x++, fx+=dx)
			{
				switch (tempimg.imgtype)
				{
					case B_CHAR:
					{
						unsigned char *p = tempimg.data;
						ival = IROUND(ireadbuf (img, fx,fy,z,intp_pref));		/* Intermediate integer value */
						ival = CLAMP (ival, 0, 255);							/* force small rounding errors back in range */
						p[idx] = ival;
						break;
					}
					case B_SHORT:
					{
						short *p = tempimg.data;
						ival = IROUND(ireadbuf (img, fx,fy,z,intp_pref));
						ival = CLAMP (ival, -32768, 32767);
						p[idx] = ival;
						break;
					}
					case B_FLOAT:
					{
						float *p = tempimg.data;
						p[idx] = ireadbuf (img, fx,fy,z,intp_pref);
						break;
					}
					case B_RGB:
					{
						rgbtriplet* prgb = tempimg.data;
						prgb [idx] = rgbireadbuf (img, fx,fy,z, intp_pref);
						break;
					}
				}
				
				idx++;
			}
	}

	handback (img, &tempimg);

}






/* Interpolate a number of Z slices into existing ones. 
   numslices is the number of slices to create between two existing ones */

/* TODO: Cannot yet handle RGB. Remember that we have a function "rgbireadbuf" !!! */

void interpol_slice (image_type* img, int numslices)
{
image_type target;
image_type oneslice;
int znew,c;
int x,y,z; long idx;
float buf;

	if (img->imgtype == B_RGB) return;
	if (img->zmax==1) return;

	znew = (img->zmax-1)*(numslices+1)+1;	/* New z dimensions */
	
	/* Allocate the intermediate and final image */
	
	if (allocate_image (&oneslice,B_FLOAT,img->xmax,img->zmax,1)==-1)
		fatal_memerr();

	if (allocate_image (&target,img->imgtype,img->xmax,img->ymax,znew)==-1)
		fatal_memerr();


	for (y=0; y<img->ymax; y++)		/* X-dominant interpolation */
	{
	
		/* Fill the x-z slice we want to interpolate from */
		
		for (x=0; x<img->xmax; x++)
			for (z=0; z<img->zmax; z++)
			{
				buf = readbuf_flt (img,x,y,z);	/* Interpolating means float, anyway */
				((float*)oneslice.data)[x+oneslice.xmax*z] = buf;
			}
		oneslice.uid = next_uid++;
		
		/* Fill and interpolate the destination */

		for (x=0; x<img->xmax; x++)
			for (z=0; z<znew; z++)
			{
				idx = x + target.xmax*(y + target.ymax*z);
				buf = 0.5*ireadbuf (&oneslice,x, (double)z/(double)(numslices+1),1,intp_pref);
				switch (target.imgtype)
				{
					case B_CHAR:
						c = (int)(0.5+buf);
						if (c<0) c=0; else if (c>255) c=255;
						((unsigned char*)target.data)[idx]=c;
						break;
					case B_SHORT:
						c = (int)(0.5+buf);
						((short*)target.data)[idx]=c;
						break;
					case B_FLOAT:
						((float*)target.data)[idx]=buf;
						break;
				}
			}
	}
	
	freebuf (&oneslice);
	if (allocate_image (&oneslice,B_FLOAT,img->ymax,img->zmax,1)==-1)
		fatal_memerr();

	for (x=0; x<img->xmax; x++)		/* Y-dominant interpolation */
	{
	
		/* Fill the y-z slice we want to interpolate from */
		
		for (y=0; y<img->ymax; y++)
			for (z=0; z<img->zmax; z++)
			{
				buf = readbuf_flt (img,x,y,z);	/* Interpolating means float, anyway */
				((float*)oneslice.data)[y+oneslice.xmax*z] = buf;
			}
		oneslice.uid = next_uid++;
		
		/* Fill and interpolate the destination */

		for (y=0; y<img->ymax; y++)
			for (z=0; z<znew; z++)
			{
				idx = x + target.xmax*(y + target.ymax*z);
				buf = 0.5*ireadbuf (&oneslice,y, (double)z/(double)(numslices+1),1,intp_pref);
				switch (target.imgtype)
				{
					case B_CHAR:
						c = (int)(0.5+buf);
						if (c<0) c=0; else if (c>255) c=255;
						((unsigned char*)target.data)[idx]+=c;
						break;
					case B_SHORT:
						c = (int)(0.5+buf);
						((short*)target.data)[idx]+=c;
						break;
					case B_FLOAT:
						((float*)target.data)[idx]+=buf;
						break;
				}
			}
	}

	handback (img, &target);
	freebuf (&oneslice);

}



/* Flip the x and z coordinates (rotate around Y)
   to allow different access to the stack */


void flip_xz (image_type* img)
{
int x,y,z,e,i;
long sidx, didx;
image_type target;
unsigned char *p1, *p2;

	if (img->zmax==1) return;	/* can do this on stacks only */

	if (allocate_image (&target,img->imgtype,img->zmax,img->ymax,img->xmax)==-1)
		fatal_memerr();
	e = DATALEN(img->imgtype);
	p1 = img->data;
	p2 = target.data;

	for (z=0; z<img->zmax; z++)
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				sidx = e*(x + img->xmax*(y + img->ymax*z));
				didx = e*(z + target.xmax*(y + target.ymax*x));
				for (i=0; i<e; i++) p2[didx+i] = p1[sidx+i];
			}
	handback (img, &target);

}


/******************************************************

	Rot Zoom: GUI for rotate and zoom operation
	
*******************************************************/


GtkAdjustment *rotzoom_rotspin_adj;
GtkAdjustment *rotzoom_xscalespin_adj;
GtkAdjustment *rotzoom_yscalespin_adj;
GtkAdjustment *rotzoom_zscalespin_adj;
image_type* rotzoomimg;


void on_rotzoom_cancel_clicked (GtkButton *button, gpointer user_data)
{


	gtk_widget_destroy (GTK_WIDGET(user_data));
}


void on_rotzoom_ok_clicked (GtkButton *button, gpointer user_data)
{
double theta;
int xnew, ynew, zadd;

	undo_nexus (&mainimg);

	theta = rotzoom_rotspin_adj->value;
	xnew = rotzoom_xscalespin_adj->value;
	ynew = rotzoom_yscalespin_adj->value;
	zadd = rotzoom_zscalespin_adj->value;

	/* Sequence of operations: Zoom first, then rotate, then add Z slices */

	if ((xnew != rotzoomimg->xmax) || (ynew != rotzoomimg->ymax))
		zoom2D (rotzoomimg, xnew,ynew);	

	if (theta !=0.0) rotate_img (rotzoomimg, theta);

	if (zadd>0)
	{
		interpol_slice (rotzoomimg, zadd);
		gtk_adjustment_init_val ((GtkAdjustment*)slice_spinbutton_adj, 0, 0, mainimg.zmax-1);
	}

	make_imlib_buffer_from_main_image (cur_slice);
	redraw (1);

	gtk_widget_destroy (GTK_WIDGET(user_data));
}





void create_rotzoom_wnd (image_type* img)
{
  GtkWidget *rotzoom_wnd;
  GtkWidget* vbox1;
  GtkWidget *table1;
  GtkWidget *rotzoom_label1, *rotzoom_label2, *rotzoom_label3, *rotzoom_label4;
  GtkWidget *rotzoom_cancel;
  GtkWidget *rotzoom_ok;
  GtkWidget *rotzoom_rotspin;
  GtkWidget *rotzoom_xscalespin;
  GtkWidget *rotzoom_yscalespin;
  GtkWidget *rotzoom_zscalespin;
  GtkWidget* hsep1;
  GtkWidget* hsep2;

	rotzoomimg = img;

	rotzoom_wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (rotzoom_wnd), "rotzoom_wnd", rotzoom_wnd);
	gtk_window_set_title (GTK_WINDOW (rotzoom_wnd), "Rotate - Zoom");
	gtk_window_set_modal (GTK_WINDOW (rotzoom_wnd), TRUE);

	vbox1 = vbox_in_window (rotzoom_wnd, "rotzoom_vbox1");
	table1 = MakeTable (rotzoom_wnd, vbox1, "table1", 2,7);
	rotzoom_label1 = MakeLabel (rotzoom_wnd,table1,"Rotate (degrees)","rotzoom_label1",0,0);
	rotzoom_label2 = MakeLabel (rotzoom_wnd,table1,"New X dimension","rotzoom_label2",0,1);
	rotzoom_label3 = MakeLabel (rotzoom_wnd,table1,"New Y dimension","rotzoom_label3",0,2);
	rotzoom_label4 = MakeLabel (rotzoom_wnd,table1,"Add new Z slices","rotzoom_label4",0,3);
	gtk_misc_set_alignment (GTK_MISC (rotzoom_label1), 0, 0.5);
	gtk_misc_set_alignment (GTK_MISC (rotzoom_label2), 0, 0.5);
	gtk_misc_set_alignment (GTK_MISC (rotzoom_label3), 0, 0.5);

	rotzoom_rotspin_adj = GTK_ADJUSTMENT(gtk_adjustment_new (0, -359, 359, 1, 10, 0));
	rotzoom_rotspin = MakeSpinbutton (rotzoom_rotspin_adj,rotzoom_wnd,table1,"rotzoom_rotspin",1,0);

	rotzoom_xscalespin_adj = GTK_ADJUSTMENT(gtk_adjustment_new (img->xmax, 16, 32768, 1, 10, 0));
	rotzoom_xscalespin = MakeSpinbutton (rotzoom_xscalespin_adj,rotzoom_wnd,table1,"rotzoom_xscalespin",1,1);

	rotzoom_yscalespin_adj = GTK_ADJUSTMENT(gtk_adjustment_new (img->ymax, 16, 32768, 1, 10, 0));
	rotzoom_yscalespin = MakeSpinbutton (rotzoom_yscalespin_adj,rotzoom_wnd,table1,"rotzoom_yscalespin",1,2);

	rotzoom_zscalespin_adj = GTK_ADJUSTMENT(gtk_adjustment_new (0, 0, 10, 1, 10, 0));
	rotzoom_zscalespin = MakeSpinbutton (rotzoom_zscalespin_adj,rotzoom_wnd,table1,"rotzoom_zscalespin",1,3);

	hsep1 = MakeHsep (rotzoom_wnd,table1,"hsep1",0,4);
	hsep2 = MakeHsep (rotzoom_wnd,table1,"hsep1",1,4);

	rotzoom_cancel = Button_in_table (rotzoom_wnd,table1,"Cancel","rotzoom_cancel",1,5,NULL);
	rotzoom_ok = Button_in_table (rotzoom_wnd,table1,"OK","rotzoom_ok",0,5,NULL);

	gtk_signal_connect (GTK_OBJECT (rotzoom_cancel), "clicked",
                    	GTK_SIGNAL_FUNC (on_rotzoom_cancel_clicked), rotzoom_wnd);
	gtk_signal_connect (GTK_OBJECT (rotzoom_ok), "clicked",
                    	GTK_SIGNAL_FUNC (on_rotzoom_ok_clicked), rotzoom_wnd);
	gtk_signal_connect (GTK_OBJECT (rotzoom_wnd), "destroy",
                    	GTK_SIGNAL_FUNC (on_rotzoom_cancel_clicked), rotzoom_wnd);

	gtk_widget_show (rotzoom_wnd);
}


void rotzoom_main ()
{

	create_rotzoom_wnd (&mainimg);
	
}



/*************************************************************

	Rebinning function. Has a similar effect as zoom,
	but zooms by integer factors through binning
	Rebinning has the advantage of noise reduction
	so if you need to reduce by a factor of e.g., 2, rebin.

**************************************************************/

/* Note: Rebinning will take place without respect of size,
	Image values are assumed to continue constant beyond image boundaries
*/




void rebin_img (image_type *img, int rebin_ngh)
{
int x,x1,y,y1,z,xm,ym,x2,y2,x3,y3,c;
image_type target;
float buf,r,g,b; int cnt;
rgbtriplet rgb;

	/* First calculate target image size and allocate it */

	if (rebin_ngh<=1) return;		/* Should not happen */

	xm = img->xmax / rebin_ngh;		/* Truncating division is intentional */
	ym = img->ymax / rebin_ngh;

	if (allocate_image (&target, img->imgtype, xm,ym,img->zmax)==-1)
		fatal_memerr();

	for (z=0; z<img->zmax; z++)
	{
		x3=0; y3=0;
		for (y=0; y<img->ymax; y+=rebin_ngh)
		{
			x3=0;
			for (x=0; x<img->xmax; x+=rebin_ngh)
			{

				buf= r = g = b = 0.0; cnt=0;
				for (y1=0; y1<rebin_ngh; y1++)
					for (x1=0; x1<rebin_ngh; x1++)
					{
						x2 = MIN(x+x1, img->xmax-1);
						y2 = MIN(y+y1, img->ymax-1);
						if (img->imgtype==B_RGB)
						{
							rgb = readbuf_rgb (img,x2,y2,z);
							r+=rgb.red; b+=rgb.blue; g+=rgb.green; cnt++;
						}
						else
						{
							buf += readbuf_flt (img,x2,y2,z);
							cnt++;
						}
					}

				if (img->imgtype==B_RGB)
				{
					r/=cnt; b/=cnt; g/=cnt;
				}
				else
					buf /= cnt; 

				if ((x3<xm) && (y3<ym))
				{
					switch (img->imgtype)
					{
						case B_CHAR:
							c = IROUND(buf); if (c<0) c=0; else if (c>255) c=255;
							((unsigned char *)target.data)[x3+xm*(y3+ym*z)] = c;
							break;
						case B_SHORT:
							c = IROUND(buf);
							((short *)target.data)[x3+xm*(y3+ym*z)] = c;
							break;
						case B_FLOAT:
							((float *)target.data)[x3+xm*(y3+ym*z)] = buf;
							break;
						case B_RGB:
							c = IROUND(r); if (c<0) c=0; else if (c>255) c=255;
							rgb.red=c;
							c = IROUND(g); if (c<0) c=0; else if (c>255) c=255;
							rgb.green=c;
							c = IROUND(b); if (c<0) c=0; else if (c>255) c=255;
							rgb.blue=c;
							((rgbtriplet *)target.data)[x3+xm*(y3+ym*z)] = rgb;
							break;
					}
				}

				x3++;
			}
			y3++;
		}
	}
	make_minmax (&target);
	handback (img, &target);

}



int rebin_func;


void on_rebin_cancel_clicked (GtkButton *widget, gpointer user_data)
{
	gtk_widget_destroy (GTK_WIDGET(user_data));
}


void on_rebin_ok_clicked (GtkButton *widget, gpointer user_data)
{
	undo_nexus (&mainimg);
	rebin_img (&mainimg, rebin_func);
	make_imlib_buffer_from_main_image (cur_slice);
	redraw (1);
	gtk_widget_destroy (GTK_WIDGET(user_data));
}



void rebin_menu_handler (GtkWidget *widget, gpointer user_data)
{
int i;

	i = *(int*) user_data;
	rebin_func = i;
}



void create_rebin_wnd (image_type* img)
{
  GtkWidget *rebin_wnd;
  GtkWidget* vbox1;
  GtkWidget *table1;
  GtkWidget *rebin_label1;
  GtkWidget *rebin_cancel;
  GtkWidget *rebin_ok;
  GtkWidget *menuitem;
  GtkWidget *rebin_menu;
  GtkWidget *rebin_menu_menu;
  GtkWidget* hsep1;
  GtkWidget* hsep2;

	rebin_func=2;

	rebin_wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (rebin_wnd), "rebin_wnd", rebin_wnd);
	gtk_window_set_title (GTK_WINDOW (rebin_wnd), "Image Binning");
	gtk_window_set_modal (GTK_WINDOW (rebin_wnd), TRUE);

	vbox1 = vbox_in_window (rebin_wnd, "rebin_vbox1");
	table1 = MakeTable (rebin_wnd, vbox1, "table1", 2,1);
	rebin_label1 = MakeLabel (rebin_wnd,table1,"Rebinning group","rebin_label1",0,0);
	rebin_menu = Make_Optionmenu (rebin_wnd,table1,"rebin_opt",1,0);
	rebin_menu_menu = gtk_menu_new ();
	menuitem = Option_menuitem (rebin_menu_menu, GTK_SIGNAL_FUNC(rebin_menu_handler), "2 x 2", (gpointer) &menuopt2);
	menuitem = Option_menuitem (rebin_menu_menu, GTK_SIGNAL_FUNC(rebin_menu_handler), "3 x 3", (gpointer) &menuopt3);
	menuitem = Option_menuitem (rebin_menu_menu, GTK_SIGNAL_FUNC(rebin_menu_handler), "4 x 4", (gpointer) &menuopt4);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (rebin_menu), rebin_menu_menu);


	hsep1 = MakeHsep (rebin_wnd,table1,"hsep1",0,4);
	hsep2 = MakeHsep (rebin_wnd,table1,"hsep1",1,4);

	rebin_cancel = Button_in_table (rebin_wnd,table1,"Cancel","rebin_cancel",1,5,NULL);
	rebin_ok = Button_in_table (rebin_wnd,table1,"OK","rebin_ok",0,5,NULL);

	gtk_signal_connect (GTK_OBJECT (rebin_cancel), "clicked",
                    	GTK_SIGNAL_FUNC (on_rebin_cancel_clicked), rebin_wnd);
	gtk_signal_connect (GTK_OBJECT (rebin_ok), "clicked",
                    	GTK_SIGNAL_FUNC (on_rebin_ok_clicked), rebin_wnd);
	gtk_signal_connect (GTK_OBJECT (rebin_wnd), "destroy",
                    	GTK_SIGNAL_FUNC (on_rebin_cancel_clicked), rebin_wnd);

	gtk_widget_show (rebin_wnd);
}


void rebin_main ()
{

	create_rebin_wnd (&mainimg);
	
}





/************************************************************+

	Horizontal and vertical mirroring, slice by slice
	
**************************************************************/



void mirror_img (image_type *img, int hor_vert)
{
int x,x1,y,y1,z,xm,ym;
unsigned char c;
short s;
image_type target;
float buf,r,g,b; int cnt;
rgbtriplet rgb;


	/* First calculate target image size and allocate it */


	xm = img->xmax; 		/* used as shorthand notation */
	ym = img->ymax;

	if (allocate_image (&target, img->imgtype, xm,ym,img->zmax)==-1)
		fatal_memerr();

	for (z=0; z<img->zmax; z++)
	{
		for (y=0; y<img->ymax; y++)
		{
			y1=y; if (hor_vert) y1=ym-y-1;
			for (x=0; x<img->xmax; x++)
			{
				x1=x; if (!hor_vert) x1=xm-x-1;

				switch (img->imgtype)
				{
					case B_CHAR:
						c = ((unsigned char *)img->data)[x+xm*(y+ym*z)];
						((unsigned char *)target.data)[x1+xm*(y1+ym*z)] = c;
						break;
					case B_SHORT:
						s = ((short *)img->data)[x+xm*(y+ym*z)];
						((short *)target.data)[x1+xm*(y1+ym*z)] = s;
						break;
					case B_FLOAT:
						buf = ((float *)img->data)[x+xm*(y+ym*z)];
						((float *)target.data)[x1+xm*(y1+ym*z)] = buf;
						break;
					case B_RGB:
						rgb = ((rgbtriplet *)img->data)[x+xm*(y+ym*z)];
						((rgbtriplet *)target.data)[x1+xm*(y1+ym*z)] = rgb;
						break;
				}
			}
		}
	}

	handback (img, &target);
}






/************************************************************+

	Edit the clickstack
	
**************************************************************/


typedef struct
{
	GtkAdjustment *xadj;	/* The adjustment part of a spin button */
	GtkWidget* xspin;		/* A spin button widget */
	GtkAdjustment *yadj;	/* The adjustment part of a spin button */
	GtkWidget* yspin;		/* A spin button widget */
	GtkAdjustment *zadj;	/* The adjustment part of a spin button */
	GtkWidget* zspin;		/* A spin button widget */
	GtkWidget* label;		/* Its label text */
}
cstedit_type;				/* Used in edit_clickstack */
cstedit_type cst[CLICKSTACK_MAX];


int cst_opts[CLICKSTACK_MAX];	/* To identify the row in the "on...changed functions */




void on_cst_cancel_clicked (GtkButton *button, gpointer user_data)
{
	gtk_widget_destroy (GTK_WIDGET(user_data));
}


void on_cst_x_adj (GtkWidget* widget, gpointer user_data)
{
int i;

	i = *(int*)user_data;
	clickstack[i].x = (int)cst[i].xadj->value;
}

void on_cst_y_adj (GtkWidget* widget, gpointer user_data)
{
int i;

	i = *(int*)user_data;
	clickstack[i].y = (int)cst[i].yadj->value;

}

void on_cst_z_adj (GtkWidget* widget, gpointer user_data)
{
int i;

	i = *(int*)user_data;
	clickstack[i].z = (int)cst[i].zadj->value;

}


void edit_clickstack (image_type* img)
{

GtkWidget *cst_wnd;
GtkWidget* vbox1;
GtkWidget *table1;
GtkWidget *l1,*l2,*l3, *l4;
GtkWidget* cst_cancelbutton;

int i,j,k;
char s[100];

	j = MIN (CLICKSTACK_MAX,clickstack_last);
	if (!j) return;	/* Nothing to edit */
	
	for (i=0; i<CLICKSTACK_MAX; i++) cst_opts[i]=i;

	cst_wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (cst_wnd), "cst_wnd", cst_wnd);
	gtk_window_set_title (GTK_WINDOW (cst_wnd), "Edit Clickstack");
	gtk_window_set_modal (GTK_WINDOW (cst_wnd), TRUE);

	vbox1 = vbox_in_window (cst_wnd, "cst_vbox1");
	table1 = MakeTable (cst_wnd, vbox1, "table1", 4,j+1);

	l1 = MakeLabel (cst_wnd,table1,"Stack position","l1",0,0);
	l2 = MakeLabel (cst_wnd,table1,"X coordinate","l2",1,0);
	l3 = MakeLabel (cst_wnd,table1,"Y coordinate","l3",2,0);
	l4 = MakeLabel (cst_wnd,table1,"Z coordinate","l4",3,0);

	for (i=0; i<j; i++)
	{
		k=j-i;
		sprintf (s, "%2d", k);
		if (i==0) strcat (s, " (Bottom)");
		cst[i].label = MakeLabel (cst_wnd,table1,s,s,0,k);
		
		cst[i].xadj = GTK_ADJUSTMENT(gtk_adjustment_new (clickstack[i].x,0,img->xmax-1,1,1,0));
		cst[i].xspin= MakeSpinbutton (cst[i].xadj,cst_wnd,table1,s,1,k);
		gtk_signal_connect (GTK_OBJECT (cst[i].xadj), "value_changed",
                      GTK_SIGNAL_FUNC (on_cst_x_adj), &cst_opts[i]);
		
		cst[i].yadj = GTK_ADJUSTMENT(gtk_adjustment_new (clickstack[i].y,0,img->ymax-1,1,1,0));
		cst[i].yspin= MakeSpinbutton (cst[i].yadj,cst_wnd,table1,s,2,k);
		gtk_signal_connect (GTK_OBJECT (cst[i].yadj), "value_changed",
                      GTK_SIGNAL_FUNC (on_cst_y_adj), &cst_opts[i]);
		
		cst[i].zadj = GTK_ADJUSTMENT(gtk_adjustment_new (clickstack[i].z,0,img->zmax-1,1,1,0));
		cst[i].zspin= MakeSpinbutton (cst[i].zadj,cst_wnd,table1,s,3,k);
		gtk_signal_connect (GTK_OBJECT (cst[i].zadj), "value_changed",
                      GTK_SIGNAL_FUNC (on_cst_z_adj), &cst_opts[i]);
	}

	cst_cancelbutton = Button_in_table (cst_wnd,vbox1,"Dismiss","cst_cancel",-1,-1,NULL);
	
	gtk_signal_connect (GTK_OBJECT (cst_cancelbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_cst_cancel_clicked), cst_wnd);
	gtk_signal_connect (GTK_OBJECT (cst_wnd), "destroy",
                      GTK_SIGNAL_FUNC (on_cst_cancel_clicked), cst_wnd);
	gtk_widget_show (cst_wnd);
}








/******************************************************

	ARTIFACTS
	With these functions, we can spoil images with
	artifacts, such as noise etc.

******************************************************/


void add_gaussian_noise (image_type* img, float amplitude, int slice, int roiflg)
{
int x,y,z,i;
int c;
float f;
int bx1,by1,bx2,by2;

	roi_bbox (img, &bx1,&by1, &bx2, &by2);
	if (roiflg) 
	{
		roiflg=1;	/* Adjust to pure 0/1 */
		bx1=0; by1=0; bx2=img->xmax; by2=img->ymax;
	}
	
	z = slice * img->xmax * mainimg.ymax;
	switch (mainimg.imgtype)
	{
		case B_CHAR:
		{
			unsigned char* p = mainimg.data;
			for (y=by1; y<=by2; y++)
				for (x=bx1; x<=bx2; x++)
					if (in_roi(x,y)^roiflg)
					{
						c=(int) (amplitude * gauss_rnd() );	/* Note: Amplitude is stdev here */
						c = p[x + mainimg.xmax*y + z] + c;
						if (c<0) c=0; else if (c>255) c=255;
						p[x + mainimg.xmax*y + z] = c;
				}
			break;
		}
		case B_SHORT:
		{
			short* p = mainimg.data;
			for (y=by1; y<=by2; y++)
				for (x=bx1; x<=bx2; x++)
					if (in_roi(x,y)^roiflg)
					{
						c=(int) (amplitude * gauss_rnd() );
						c = p[x + mainimg.xmax*y + z] + c;
						p[x + mainimg.xmax*y + z] = c;
					}
			break;
		}
		case B_RGB:
		{
			unsigned char* p = mainimg.data;
			for (i=0; i<3; i++)
				for (y=by1; y<=by2; y++)
					for (x=bx1; x<=bx2; x++)
						if (in_roi(x,y)^roiflg)
						{
							c=(int) (amplitude * gauss_rnd() );
							c = p[i+3*(x + mainimg.xmax*y + z)] + c;
							if (c<0) c=0; else if (c>255) c=255;
							p[i+3*(x + mainimg.xmax*y + z)] = c;
						}
			break;
		}
		case B_FLOAT:
		{
			float* p = mainimg.data;
			for (y=by1; y<=by2; y++)
				for (x=bx1; x<=bx2; x++)
					if (in_roi(x,y)^roiflg)
					{
						f= (amplitude * gauss_rnd() );
						f = p[x + mainimg.xmax*y + z] + f;
						p[x + mainimg.xmax*y + z] = f;
					}
			break;
		}
	}
	
	img->uid = next_uid++;

}


void add_shot_noise (image_type* img, float probability, int slice, int roiflg)
{
int x,y,z,i;
int c, ct;
float f,a,b;
int bx1,by1,bx2,by2;

	a = 200; b=-0.5*a;
	ct = a*0.01*probability; 	/* a good value for prob is 5 (5%) so that ct==10 */
	roi_bbox (img, &bx1,&by1, &bx2, &by2);
	if (roiflg) 
	{
		roiflg=1;	/* Adjust to pure 0/1 */
		bx1=0; by1=0; bx2=img->xmax; by2=img->ymax;
	}

	if (file_ok) 
	{
		z = slice * img->xmax * img->ymax;
		switch (img->imgtype)
		{
			case B_CHAR:
			{
				unsigned char* p = img->data;
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
						if (in_roi(x,y)^roiflg)
						{
							c=(int) (b+a*rand()/(RAND_MAX+1.0));
							if ((c>0) && (c<ct))
							{
								c = img->maxval.imax;
								p[x + img->xmax*y + z] = c;
							}
							else if ((c<0) && (c>-ct))
							{
								c = img->minval.imin;
								p[x + img->xmax*y + z] = c;
							}

						}
				break;
			}
			case B_SHORT:
			{
				short* p = img->data;
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
						if (in_roi(x,y)^roiflg)
						{
							c=(int) (b+a*rand()/(RAND_MAX+1.0));
							if ((c>0) && (c<ct))
							{
								c = img->maxval.imax;
								p[x + img->xmax*y + z] = c;
							}
							else if ((c<0) && (c>-ct))
							{
								c = img->minval.imin;
								p[x + img->xmax*y + z] = c;
							}
						}
				break;
			}
			case B_RGB:
			{
				unsigned char* p = img->data;
				for (i=0; i<3; i++)
					for (y=0; y<img->ymax; y++)
						for (x=0; x<img->xmax; x++)
							if (in_roi(x,y)^roiflg)
							{
								c=(int) (b+a*rand()/(RAND_MAX+1.0));
								if ((c>0) && (c<ct))
								{
									c = img->maxval.imax;
									p[i + 3*(x + img->xmax*y + z)] = c;
								}
								else if ((c<0) && (c>-ct))
								{
									c = img->minval.imin;
									p[i + 3*(x + img->xmax*y + z)] = c;
								}
							}
				break;
			}
			case B_FLOAT:
			{
				float* p = img->data;
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
						if (in_roi(x,y)^roiflg)
						{
							c=(int) (b+a*rand()/(RAND_MAX+1.0));
							if ((c>0) && (c<ct))
							{
								f = img->maxval.fmax;
								p[x + img->xmax*y + z] = f;
							}
							else if ((c<0) && (c>-ct))
							{
								f = img->minval.fmin;
								p[x + img->xmax*y + z] = f;
							}
						}
				break;
			}
		}

	}
	img->uid = next_uid++;

}


/* A rudiment related to the above code to add a line artifact 
   maybe we can use it somewhere?

	if ((x & 0x0f) ==1)
	{
		c=(int) (40 + 2*rand()/(RAND_MAX+1.0));
		p[x + img->xmax*y + z] += c;
	}

*/

void chg_image_values (image_type* img, float value, int slice, int roiflg)
{
int x,y,z,i;
int c;
int bx1,by1,bx2,by2;

	roi_bbox (img, &bx1,&by1, &bx2, &by2);
	c = (int)value;
	if (roiflg) 
	{
		roiflg=1;	/* Adjust to pure 0/1 */
		bx1=0; by1=0; bx2=img->xmax-1; by2=img->ymax-1;
	}
	
	z = slice * img->xmax * img->ymax;
	switch (img->imgtype)
	{
		case B_CHAR:
		{
			unsigned char* p = img->data;
			for (y=by1; y<=by2; y++)
				for (x=bx1; x<=bx2; x++)
				{
					if (in_roi(x,y)^roiflg)
					{
						p[x + img->xmax*y + z] = c;
					}
				}
			break;
		}
		case B_SHORT:
		{
			short* p = img->data;
			for (y=by1; y<=by2; y++)
				for (x=bx1; x<=bx2; x++)
					if (in_roi(x,y)^roiflg)
					{
						p[x + img->xmax*y + z] = c;
					}
			break;
		}
		case B_RGB:
		{
			unsigned char* p = img->data;
			for (i=0; i<3; i++)
				for (y=by1; y<=by2; y++)
					for (x=bx1; x<=bx2; x++)
						if (in_roi(x,y)^roiflg)
						{
							p[i+3*(x + img->xmax*y + z)] = c;
						}
			break;
		}
		case B_FLOAT:
		{
			float* p = img->data;
			for (y=by1; y<=by2; y++)
				for (x=bx1; x<=bx2; x++)
					if (in_roi(x,y)^roiflg)
					{
						p[x + img->xmax*y + z] = value;
					}
			break;
		}
	}
	
	img->uid = next_uid++;

	
}




/******************************************************

	GUI for the Edit -> Values function

*******************************************************/

GtkAdjustment* edval_value_adj;
int edval_roi, edval_function, edval_allslices;
GtkWidget *edval_menu;


void on_edval_cancel_clicked (GtkButton *button, gpointer user_data)
{
	gtk_widget_destroy (GTK_WIDGET(user_data));
}


void on_edval_ok_clicked (GtkButton *button, gpointer user_data)
{
float min, max;
int z,z1,z2;
gchar res;
	

	undo_nexus (&mainimg);

	imgminmax (&mainimg, &min, &max);
	if (edval_allslices)
	{
		z1=0; z2=mainimg.zmax-1;
	}
	else
		z1=z2=cur_slice;

	switch (edval_function)
	{
		case 0:	
			for (z=z1; z<=z2; z++)
				chg_image_values (&mainimg, 0, z, edval_roi);
			break;
		case 1:
			for (z=z1; z<=z2; z++)
				chg_image_values (&mainimg, max, z, edval_roi);
			break;
		case 2:
			for (z=z1; z<=z2; z++)
				chg_image_values (&mainimg, min, z, edval_roi);
			break;
		case 3:
			for (z=z1; z<=z2; z++)
				chg_image_values (&mainimg, edval_value_adj->value, z, edval_roi);
			break;
		case 4:
			for (z=z1; z<=z2; z++)
				add_gaussian_noise (&mainimg, edval_value_adj->value, z, edval_roi);
			break;
		case 5:
			for (z=z1; z<=z2; z++)
				add_shot_noise (&mainimg, edval_value_adj->value, z, edval_roi);
	}
	
	make_minmax (&mainimg);
	make_imlib_buffer_from_main_image (cur_slice);
	redraw (0);
	
	gtk_widget_destroy (GTK_WIDGET(user_data));
}


void on_edval_checkbutton1_toggled (GtkButton *button, gpointer user_data)
{
	edval_roi = (*(GtkToggleButton*)user_data).active;

}


void on_edval_checkbutton2_toggled (GtkButton *button, gpointer user_data)
{
	edval_allslices = (*(GtkToggleButton*)user_data).active;

}


void edval_set_menu (GtkWidget *widget, gpointer user_data)
{

	edval_function = *(int*)user_data;
	
}

void on_edval_value_changed (GtkWidget *widget, gpointer user_data)
{

	edval_function = menuopt3;
	gtk_option_menu_set_history (GTK_OPTION_MENU (edval_menu), edval_function);
	
}



void create_editval_wnd (image_type* img)
{
  GtkWidget *edval_wnd;
  GtkWidget* vbox1;
  GtkWidget *table1;
  GtkWidget *edval_label1;
  GtkWidget *edval_label2;
  GtkWidget *edval_label3;
  GtkWidget *edval_label4;
  GtkWidget *edval_roisel;
  GtkWidget *edval_slicesel;
  GtkWidget *edval_cancel;
  GtkWidget *edval_ok;
  GtkWidget *edval_value;
  GtkWidget* hsep1;
  GtkWidget* hsep2;
  GtkWidget *edval_menu_menu;
  GtkWidget *glade_menuitem;
 
  float min, max;
  
  	switch (img->imgtype)
	{
		case B_CHAR:
		case B_RGB:
			min = 0; max = 255;
			break;
		case B_SHORT:
			min = -32767; max = 32767;
			break;
		case B_FLOAT:
			min = MIN(img->minval.fmin,-65535); max = MAX(img->maxval.fmax,65535);
			break;
		default:
			min=0; max=0; 	/* kill a compiler warning */
	}

	edval_roi=0;	edval_allslices=0; edval_function=0;		

	edval_wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (edval_wnd), "edval_wnd", edval_wnd);
	gtk_window_set_title (GTK_WINDOW (edval_wnd), "Edit Values");
	gtk_window_set_modal (GTK_WINDOW (edval_wnd), TRUE);

	vbox1 = vbox_in_window (edval_wnd, "edval_vbox1");
	table1 = MakeTable (edval_wnd, vbox1, "table1", 2,6);
	edval_label1 = MakeLabel (edval_wnd,table1,"Region","edval_label1",0,0);
	edval_label2 = MakeLabel (edval_wnd,table1,"Slice","edval_label2",0,1);
	edval_label3 = MakeLabel (edval_wnd,table1,"Function","edval_label3",0,2);
	edval_label4 = MakeLabel (edval_wnd,table1,"Value","edval_label4",0,3);

	edval_roisel = Checkbutton_in_table (edval_wnd, table1, "Outside ROI","edval_check1",1,0, NULL);
	edval_slicesel = Checkbutton_in_table (edval_wnd, table1, "All slices","edval_check2",1,1, NULL);
		
	edval_menu = Make_Optionmenu (edval_wnd,table1,"edval_menu",1,2);
	edval_menu_menu = gtk_menu_new ();
	glade_menuitem = Option_menuitem (edval_menu_menu, GTK_SIGNAL_FUNC(edval_set_menu), "Zero", (gpointer) &menuopt0);
	glade_menuitem = Option_menuitem (edval_menu_menu, GTK_SIGNAL_FUNC(edval_set_menu), "Maximum", (gpointer) &menuopt1);
	glade_menuitem = Option_menuitem (edval_menu_menu, GTK_SIGNAL_FUNC(edval_set_menu), "Minimum", (gpointer) &menuopt2);
	glade_menuitem = Option_menuitem (edval_menu_menu, GTK_SIGNAL_FUNC(edval_set_menu), "Specified value", (gpointer) &menuopt3);
	glade_menuitem = Option_menuitem (edval_menu_menu, GTK_SIGNAL_FUNC(edval_set_menu), "Gaussian noise", (gpointer) &menuopt4);
	glade_menuitem = Option_menuitem (edval_menu_menu, GTK_SIGNAL_FUNC(edval_set_menu), "Shot noise", (gpointer) &menuopt5);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (edval_menu), edval_menu_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (edval_menu), 0);

	edval_value_adj = GTK_ADJUSTMENT(gtk_adjustment_new (0, min, max, 1, 1, 0));
	edval_value = MakeSpinbutton (edval_value_adj,edval_wnd,table1,"edval_value",1,3);
	if (img->imgtype==B_FLOAT) gtk_spin_button_set_digits (GTK_SPIN_BUTTON(edval_value), 1);

	hsep1 = MakeHsep (edval_wnd,table1,"hsep1",0,4);
	hsep2 = MakeHsep (edval_wnd,table1,"hsep2",1,4);
	
	edval_cancel = Button_in_table (edval_wnd,table1,"Cancel","edval_cancel",1,5,NULL);
	edval_ok = Button_in_table (edval_wnd,table1,"OK","edval_ok",0,5,NULL);

	gtk_signal_connect (GTK_OBJECT (edval_wnd), "destroy", GTK_SIGNAL_FUNC (on_edval_cancel_clicked), edval_wnd);
	gtk_signal_connect (GTK_OBJECT (edval_ok), "clicked", GTK_SIGNAL_FUNC (on_edval_ok_clicked), edval_wnd);
	gtk_signal_connect (GTK_OBJECT (edval_roisel), "toggled", GTK_SIGNAL_FUNC (on_edval_checkbutton1_toggled), edval_roisel);
	gtk_signal_connect (GTK_OBJECT (edval_value_adj), "value_changed", GTK_SIGNAL_FUNC (on_edval_value_changed), edval_value_adj);
	gtk_signal_connect (GTK_OBJECT (edval_slicesel), "toggled", GTK_SIGNAL_FUNC (on_edval_checkbutton2_toggled), edval_slicesel);
	gtk_signal_connect (GTK_OBJECT (edval_cancel), "clicked", GTK_SIGNAL_FUNC (on_edval_cancel_clicked), edval_wnd);

	gtk_widget_show (edval_wnd);
}


void editval_main ()
{

	if (!current_roi.roi_mode)
		ShowMessage ("This function requires a ROI");
	else
		create_editval_wnd (&mainimg);

}





/* A function to test whether a file exists */

int file_exists (const char *filename)
{
struct stat buf;
int err;

	if (strlen(filename)==0) return 0;		/* No filename doesn't exist :-) */

	err = stat (filename, &buf);
	dp (2, "STAT: File mode is %d\n", buf.st_mode);
	if (err==0)
		return 1;				/* No error -> file exists */


	return 0;
}

	


/*******************************************************************/

/*	A general-purpose 2D flood-fill function. It works as follows:
		- x and y are the seed values in slice z
		- newval is the new value to flood-fill with
		- level needs to be passed (initially level=0) to keep track of the recursion level

		2D flood fill distinguishes two conditions. If newval is zero, 
			it flood-fills a contigugous region with zeros, i.e., makes a cluster vanish
		If newval is >0, it flood-fills a contigous region of zero-values pixels with 
			the new value (filling holes, for example)
*/

void rg_floodfill (image_type* img, float newval, int x, int y, int z, int* level)
{

	if (*level==0) dp (1, "Flood-filling from (%d,%d) in slice %d with value %f\n",x,y,z,newval);

	/* Abort recursion if out-of-range or non-background or already visited */
	
	if ((x<0) || (y<0) || (x>=img->xmax) || (y>=img->ymax)) return;

	/* Abort recursion if
		1) newval is zero and the current pixel is also zero
		2) newval is nonzero and the current pixel is nonzero */

	if (F_ISZERO(newval))
	{
		if (F_ISZERO(readbuf_flt (img,x,y,z))) return;
	}
	else
	{
		if (!F_ISZERO(readbuf_flt (img,x,y,z))) return;
	}

	
	(*level)++;
	
	/* Set the current point to newval */

	switch (img->imgtype)
	{
		case B_CHAR:
		{
			unsigned char *p = img->data;
			int c;
			if (newval<0) c=0; else if (newval>255) c=255; else c=IROUND(newval);
			p[x + img->xmax*(y + img->ymax*z)]=c;
			break;
		}
		case B_SHORT:
		{
			short *p = img->data;
			int c;
			c=IROUND(newval);
			p[x + img->xmax*(y + img->ymax*z)]=c;
			break;
		}
		case B_FLOAT:
		{
			float *p = img->data;
			p[x + img->xmax*(y + img->ymax*z)]=newval;
			break;
		}
		case B_RGB:
		{
			rgbtriplet *p = img->data;
			int c;
			if (newval<0) c=0; else if (newval>255) c=255; else c=IROUND(newval);
			p[x + img->xmax*(y + img->ymax*z)]= (rgbtriplet){c,c,c};
			break;
		}
	}


	/* and visit the 8 neighbors */
	
	rg_floodfill (img,newval,x+1,y,z,level);
	rg_floodfill (img,newval,x-1,y,z,level);
	rg_floodfill (img,newval,x,y+1,z,level);
	rg_floodfill (img,newval,x,y-1,z,level);
	rg_floodfill (img,newval,x+1,y+1,z,level);
	rg_floodfill (img,newval,x-1,y+1,z,level);
	rg_floodfill (img,newval,x+1,y-1,z,level);
	rg_floodfill (img,newval,x-1,y-1,z,level);
	(*level)--;
}




/*******************************************************************/

/* Combine images. Two different functions for use with the macro interpreter.
	The first function combines two images to form a stack. The second
	combines two images into a color image, very much like the rgbload
	function in the import module.
*/


int combine1 (image_type *img1, image_type *img2)
{
image_type dest;
long size;
unsigned char *p;
int e;
	
	if ((img1->xmax != img2->xmax) || (img1->ymax != img2->ymax))
		return -1;													/* Error: Size mismatch */

	/* Combine1 should be somewhat flexible with the target type, but for now
		we simply reject any image type mismatch */
	
	if (img1->imgtype != img2->imgtype) return -1;

	if (allocate_image (&dest, img1->imgtype, img1->xmax, img1->ymax, img1->zmax+img2->zmax)==-1)
		fatal_memerr();

	/* Because we have matching elements, we can brutally memcpy to the target pointers */

	p = dest.data; e = DATALEN(dest.imgtype); size = e*img1->xmax*img1->ymax*img1->zmax;
	memcpy (p, img1->data, size);
	p +=size;
	memcpy (p, img2->data, e*img2->xmax*img2->ymax*img2->zmax);

	/* Free the original images, then hand back the dest through img1. */

	freebuf (img2);
	handback (img1, &dest);
	return 0;
}



int combine2 (image_type *img1, image_type *img2)
{
image_type dest;
rgbtriplet *p;
int x,y,z;
hsvtriplet hsv;
rgbtriplet rgb;
float buf1, buf2;
float max1,min1, max2, min2, val, sat;

	
	if ((img1->xmax != img2->xmax) || (img1->ymax != img2->ymax) || (img1->zmax!=img2->zmax))
		return -1;													/* Error: Size mismatch */

	imgminmax (img1, &min1, &max1);
	imgminmax (img2, &min2, &max2);

	if (allocate_image (&dest, B_RGB, img1->xmax, img1->ymax, img1->zmax)==-1)
		fatal_memerr();
	p = dest.data;

	for (z=0; z<img1->zmax; z++)
		for (y=0; y<img1->ymax; y++)
			for (x=0; x<img1->xmax; x++)
			{
				/* Scale buf1 and buf2 to fall within [0...1], then compute val */

				buf1 = (readbuf_flt (img1, x,y,z)-min1)/(max1-min1);
				buf2 = (readbuf_flt (img2, x,y,z)-min2)/(max2-min2);
				val = 0.707*sqrt (SQR(buf1)+SQR(buf2));

				/* Sat is the imbalance between buf1 and buf2, but if val is very small,
					make this pixel black. */

				if (val < 0.01)
				{
					rgb.red=rgb.green=rgb.blue=0;
				}
				else
				{
					if (buf1 > buf2)
					{
						hsv.hue = 60;
						sat = (buf1-buf2)/val;
					}
					else
					{
						hsv.hue = 240;
						sat = (buf2-buf1)/val;
					}
					hsv.val = (int)(255.0*val);
					hsv.sat = (int)(255.0*sat);
					rgb = hsv_to_rgb (hsv);
				}

				p[x + dest.xmax*(y + dest.ymax*z)] = rgb;

			}
		

	/* Free the original images, then hand back the dest through img1. */

	freebuf (img2);
	handback (img1, &dest);
	return 0;
}




/*************************************************************

	Karhunen-Loeve-Transform (Principal component analysis of
	the RGB space of a RGB image)

	The K-L transform acts on RGB images by interpreting
	each triplet as a point in three-dimensional RGB space.
	The K-L transform performs a coordinate system trans-
	formation so that one axis has the highest contrast
	(the principal component). The principal component
	is then returned as grayscale (luminance) value.

	NEW in 0.9.7: Can also handle 3-stacks of arbitrary type,
	and it returns all three principal components.

*************************************************************/


/* Compute covariance matrix in C */

void kl_comp_covar_mat (image_type* img, MFLOAT* C, int slice, int oorflg)
{
MFLOAT mr,mg,mb,sum;
MFLOAT c1, c2, c3;			/* R,G,B components in non-RGB types */
int i,j,x,y;
long l,z,cnt;
unsigned char* p;
rgbtriplet* prgb;
rgbtriplet rgb;


	p = img->data;

	/* Step 1: Compute the image mean for each pane */

	z = slice * img->xmax * img->ymax;

	if (img->imgtype == B_RGB)
	{
		prgb=img->data;
		mr=0.0; mg=0.0; mb=0.0; cnt=0;
		for (l=0; l<img->xmax * img->ymax; l++)
		{
			rgb = prgb[l+z];
			if ((!oorflg) || (rgb.red>0) || (rgb.green>0) || (rgb.blue>0))
			{
				mr += rgb.red; mg += rgb.green; mb += rgb.blue;
				cnt++;
			}
		}
	}
	else if ( (img->zmax %3) ==0)
	{
		mr=0.0; mg=0.0; mb=0.0; cnt=0;
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				c1 = readbuf_flt (img,x,y,slice);
				c2 = readbuf_flt (img,x,y,slice+1);
				c3 = readbuf_flt (img,x,y,slice+2);
				mr += c1; mg += c2; mb += c3;
				cnt++;
			}
	}
	else
	{
		ShowMessage ("Error: Incompatible stack size");
		return;
	}

	if (cnt>0)
	{
		mr/=cnt; mb/=cnt; mg/=cnt;
	}
	else
	{
		ShowMessage ("Cannot compute image mean - no data?");
		return;
	}


	/* Step 2: Compute the covariance matrix */
		
	cnt=0; sum=0.0;
	if (img->imgtype == B_RGB)
	{
		for (l=0; l<img->xmax * img->ymax; l++)
		{
			rgb = prgb[l+z];
			if ((!oorflg) || (rgb.red>0) || (rgb.green>0) || (rgb.blue>0))
			{
				C[0] += (rgb.red   - mr)*(rgb.red   - mr);	/* RR = 0,0 */
				C[1] += (rgb.red   - mr)*(rgb.green - mg);	/* RG = 0,1 */
				C[2] += (rgb.red   - mr)*(rgb.blue  - mb);	/* RB = 0,2 */
				C[4] += (rgb.green - mg)*(rgb.green - mg);	/* GG = 1,1 */
				C[5] += (rgb.green - mg)*(rgb.blue  - mg);	/* GB = 1,2 */
				C[8] += (rgb.blue  - mg)*(rgb.blue  - mg);	/* BB = 2,2 */
				cnt++;
			}
		}
	}
	else			/* Case of not a three-stack has already been trapped */
	{
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				c1 = readbuf_flt (img,x,y,slice);
				c2 = readbuf_flt (img,x,y,slice+1);
				c3 = readbuf_flt (img,x,y,slice+2);
				C[0] += (c1   - mr)*(c1  - mr);	/* RR = 0,0 */
				C[1] += (c1   - mr)*(c2  - mg);	/* RG = 0,1 */
				C[2] += (c1   - mr)*(c3  - mb);	/* RB = 0,2 */
				C[4] += (c2   - mg)*(c2  - mg);	/* GG = 1,1 */
				C[5] += (c2   - mg)*(c3  - mg);	/* GB = 1,2 */
				C[8] += (c3   - mg)*(c3  - mg);	/* BB = 2,2 */
				cnt++;
			}
	}

	C[3] = C[1];		/* RG = GR */
	C[7] = C[5];		/* GB = BG */
	C[6] = C[2];		/* BR = RB */
	for (i=0; i<9; i++) C[i] /= cnt;
	
	dp (2, "K-L Transform: This is the covariance matrix:\n");
	for (i=0; i<3; i++)
	{
		for (j=0; j<3; j++)
			dp (2, "%7.2f   ",C[3*i+j]);
		dp (2, "\n");
	}
	dp (2,"----------------------------------------\n");
}



/* Step 3: Compute the eigenvectors and the transposed and sorted
   eigenmatrix */

void kl_comp_transform_mat (MFLOAT* C)
{
MFLOAT D[3], E[3];
int i,j;

	householder_reduction (C,3,D,E);
	eigen_ql (C,3,D,E);
	
	dp (2, "K-L Transform: Eigenvectors ");
	for (i=0; i<3; i++) dp (2, "%f   ",D[i]);
	dp (2,"\n");
	dp (2, "K-L Transform: This is the transformation matrix:\n");
	for (i=0; i<3; i++)
	{
		for (j=0; j<3; j++)
			dp (2, "%7.2f   ",C[3*i+j]);
		dp (2, "\n");
	}
	dp (2,"----------------------------------------\n");

}


void karhunen_loeve_transform (image_type* img, int oorflg)
{
MFLOAT C[9];
image_type buf;
MFLOAT V[3];
int x,y,z,zinc;
long zidx;
rgbtriplet* p1;
float* p2;
rgbtriplet rgb;


	
	/* Allocate the output buffer (float grayscale) */

	if (img->imgtype == B_RGB)			/* Three stack slices per RGB slice */
	{
		if (allocate_image (&buf, B_FLOAT, img->xmax, img->ymax, 3*img->zmax) ==-1)
			fatal_memerr();
		zinc = 1;
	}
	else
	{
		if (allocate_image (&buf, B_FLOAT, img->xmax, img->ymax, img->zmax) ==-1)
			fatal_memerr();
		zinc = 3;
	}


	/* Transform slice by slice (TODO: is that a good idea?? Or should we
	   compute one transformation matrix for the entire stack?) */

	for (z=0; z<img->zmax; z+=zinc)
	{
		dp (1, "Karhunen-Loeve-Transform: Processing slice %d\n",z);

		kl_comp_covar_mat (img, C, z, oorflg);
		kl_comp_transform_mat (C);

		if (img->imgtype == B_RGB)
		{
			p1 = img->data;
			p2 = buf.data;
			zidx = img->xmax*img->ymax;
			
			/* Pixel-by-pixel conversion of RGB image into a stack */
			
			for (y=0; y<img->ymax; y++)
				for (x=0; x<img->xmax; x++)
				{
					rgb = p1[x+img->xmax*y+z*zidx];
					V[0] = (float)rgb.red;
					V[1] = (float)rgb.green;
					V[2] = (float)rgb.blue;
					matmult (C,V,3);
					p2[x+img->xmax*y+3*z*zidx	]		= V[0]; 	/* First principal component */
					p2[x+img->xmax*y+(3*z+1)*zidx] 	= V[1]; 	/* Second principal component */
					p2[x+img->xmax*y+(3*z+2)*zidx] 	= V[2]; 	/* Third principal component */
				}
		}
		else		/* Any non-RGB type */
		{
			p2 = buf.data;
			zidx = img->xmax*img->ymax;
			
			/* Pixel-by-pixel conversion of stack into a stack of same size */
			
			for (y=0; y<img->ymax; y++)
				for (x=0; x<img->xmax; x++)
				{
					V[0] = readbuf_flt (img,x,y,z);
					V[1] = readbuf_flt (img,x,y,z+1);
					V[2] = readbuf_flt (img,x,y,z+2);
					matmult (C,V,3);
					p2[x+img->xmax*y+z*zidx]		= V[0]; 	/* First principal component */
					p2[x+img->xmax*y+(z+1)*zidx] 	= V[1]; 	/* Second principal component */
					p2[x+img->xmax*y+(z+2)*zidx] 	= V[2]; 	/* Third principal component */
				}
		}

	}
	
	/* return the new b&w image */
	
	handback (img, &buf);

}






/*************************************************************/


