

/*
 * Prefs: The preferences (settings) dialog
 */


/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.

		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>


#include "application.h"
#include "defs.h"
#include "dlgbox.h"
#include "functions.h"
#include "functions2.h"
#include "gtkhelpers.h"
#include "imgbuf.h"
#include "interface.h"
#include "miscop.h"
#include "randgen.h"
#include "roi.h"
#include "viz.h"












/******************************************************************

	Preferences Dialog

******************************************************************/


struct
{
	int tiffreport;
	int show_rulers;
	int intp_prefs;
	int autobinning;
	int debuglevel;
	int skip_license;
	UNDO_TYPE undo_function;
	char browspath[1024];
	char helppath[1024];
	char dsopath[1024];
	char undo_path[1024];
	char gnuplotpath[1024];
	char imager_sv_path[1024];
}
prefs;

GtkWidget* prefs_entry1;
GtkWidget* prefs_entry2;
GtkWidget* prefs_entry3;
GtkWidget* prefs_entry4;
GtkWidget* prefs_entry5;
GtkWidget* prefs_entry6;







int read_prefs ()
{
FILE* FP;
char fname [256];
char s[256];
char s1[10];
char* home;
int line,i,j;

	/* Do this silently because gtk is not set up at this point */

	home = getenv ("HOME");
	if (!home) return -1;

	strcpy (fname, home);
	strcat (fname, "/.cimagerc");
	FP = fopen (fname, "rt");
	if (!FP)
	{
		strcpy (fname, "/etc/cimage.conf");
		FP = fopen (fname, "rt");
	}
	if (!FP) return -1;			/* no preferences file. Too bad. */

	/* Read the file line by line */

	line=0;
	while (fgets (s, 255, FP) != NULL)
	{
		line++;
		for (i=0; i<9; i++) s1[i]=0;
		if ((!isempty(s)) && (s[0]!='#'))	/* not a comment line */
		{
			if (!strncasecmp (s, "tiffinfo", 8))
			{
				i=8;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				strncpy (s1, &s[i], MIN(j-i,10));
				tiffreport = atoi (s1);
				if ((tiffreport!=0) && (tiffreport!=1))
				{
					dp (0,"Warning: Illegal value %d in .cimagerc, line %d\n", tiffreport,line);
					tiffreport=0;
				}
			}
			else if (!strncasecmp (s, "showrulers", 10))
			{
				i=10;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				strncpy (s1, &s[i], MIN(j-i,10));
				show_rulers = atoi (s1);
				if ((show_rulers!=0) && (show_rulers!=1))
				{
					dp (0,"Warning: Illegal value %d in .cimagerc, line %d\n", show_rulers,line);
					show_rulers=0;
				}
			}
			else if (!strncasecmp (s, "interpolation", 13))
			{
				i=13;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				strncpy (s1, &s[i], MIN(j-i,10));
				intp_pref = atoi (s1);
				if ((intp_pref<0) || (intp_pref>6))
				{
					dp (0,"Warning: Illegal value %d in .cimagerc, line %d\n", intp_pref,line);
					intp_pref=1;
				}
			}
			else if (!strncasecmp (s, "auto_binning", 12))
			{
				i=12;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				strncpy (s1, &s[i], MIN(j-i,10));
				autobinning = atoi (s1);
				if ((autobinning!=0) && (autobinning!=1))
				{
					dp (0,"Warning: Illegal value %d in .cimagerc, line %d\n", autobinning,line);
					autobinning=1;
				}
			}
			else if (!strncasecmp (s, "expand_contrast", 15))		/* Deprecated! */
			{
				i=15;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				strncpy (s1, &s[i], MIN(j-i,10));
				autoscale = atoi (s1);
				if ((autoscale<0) || (autoscale>1))
				{
					dp (0,"Warning: Illegal value %d in .cimagerc, line %d\n", autoscale,line);
					autoscale=1;
				}
			}
			else if (!strncasecmp (s, "debuglevel", 10))
			{
				i=10;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				strncpy (s1, &s[i], MIN(j-i,10));
				debuglevel = atoi (s1);
				if ((debuglevel<0) || (debuglevel>3))
				{
					dp (0,"Warning: Illegal value %d in .cimagerc, line %d\n", debuglevel,line);
					debuglevel=0;
				}
			}
			else if (!strncasecmp (s, "skip_license", 12))
			{
				i=12;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				strncpy (s1, &s[i], MIN(j-i,10));
				skip_license = atoi (s1);
				if ((debuglevel<0) || (debuglevel>3))
				{
					dp (0,"Warning: Illegal value %d in .cimagerc, line %d\n", skip_license,line);
					skip_license=0;
				}
			}
			else if (!strncasecmp (s, "undo_function", 13))
			{
				i=13;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				strncpy (s1, &s[i], MIN(j-i,13));
				undo_function = atoi (s1);
				if ((undo_function<0) || (undo_function>2))
				{
					dp (0,"Warning: Illegal value %d in .cimagerc, line %d\n", debuglevel,line);
					undo_function=0;
				}
			}
			else if (!strncasecmp (s, "path_to_browser", 15))
			{
				i=15;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				memset (browspath, 0, 1020);
				strncpy (browspath, &s[i], MIN(1000,j-i));
			}
			else if (!strncasecmp (s, "path_to_gnuplot", 15))
			{
				i=15;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				memset (gnuplotpath, 0, 1020);
				strncpy (gnuplotpath, &s[i], MIN(1000,j-i));
			}
			else if (!strncasecmp (s, "path_to_help", 12))
			{
				i=12;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				memset (helppath, 0, 1020);
				strncpy (helppath, &s[i], MIN(1000,j-i));
			}
			else if (!strncasecmp (s, "path_to_dsos", 12))
			{
				i=12;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				memset (dsopath, 0, 1020);
				strncpy (dsopath, &s[i], MIN(1000,j-i));
			}
			else if (!strncasecmp (s, "path_to_undo", 12))
			{
				i=12;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				memset (undo_path, 0, 1020);
				strncpy (undo_path, &s[i], MIN(1000,j-i));
			}
			else if (!strncasecmp (s, "sv_viewer_name", 14))
			{
				i=14;
				while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t'))) i++;
				j=i+1;
				while ((j<strlen(s)) && ((s[j]!=' ') && (s[j]!='\t') && (s[j]!='\n'))) j++;
				memset (imager_sv_path, 0, 1020);
				strncpy (imager_sv_path, &s[i], MIN(1000,j-i));
			}
			else
				dp (0,"Warning: Illegal keyword in .cimagerc, line %d\n",line);
		}
	}

	fclose (FP);
	return 0;
}



/* This function creates the initial Crystal Image .rc file.  */

void save_default_prefs ()
{
FILE* FP;
char fname [256];
char* home;

	prefs.tiffreport = 0;
	prefs.show_rulers = 0;
	prefs.intp_prefs = INTP_BILIN;
	prefs.debuglevel = 1;
	prefs.autobinning = 1;
	prefs.skip_license=1;
	prefs.undo_function = 1;
	strcpy (prefs.browspath, "/usr/bin/rekonq");		/* TODO: Should query system browser */
	strcpy (prefs.helppath, "/usr/local/cimage/doc");
	prefs.dsopath[0] = 0;								/* Keep empty b/c we look into /usr/local anyway */
	strcpy (prefs.undo_path, "/var/tmp/");
	strcpy (prefs.imager_sv_path, imager_sv_path);


	home = getenv ("HOME");
	if (!home)
		ShowMessage ("Cannot determine home directory");
	dp (2, "Home directory is %s\n", home);

	strcpy (fname, home);
	strcat (fname, "/.cimagerc");
	dp (2, "Saving preferences in %s\n",fname);
	FP = fopen (fname, "wt");
	if (!FP) ShowMessage ("Error saving preferences");

	fprintf (FP, "Tiffinfo %d\n",prefs.tiffreport);
	fprintf (FP, "Showrulers %d\n",prefs.show_rulers);
	fprintf (FP, "Interpolation %d\n",prefs.intp_prefs);
	fprintf (FP, "Auto_Binning %d\n",prefs.autobinning);
	fprintf (FP, "Debuglevel %d\n",prefs.debuglevel);
	fprintf (FP, "Skip_license %d\n",prefs.skip_license);
	fprintf (FP, "Undo_function %d\n",prefs.undo_function);
	fprintf (FP, "Path_to_undo  %s\n", prefs.undo_path);
	fprintf (FP, "Path_to_Browser %s\n",prefs.browspath);
	fprintf (FP, "Path_to_Gnuplot %s\n",prefs.gnuplotpath);
	fprintf (FP, "Path_to_Help %s\n",prefs.helppath);
	fprintf (FP, "Path_to_DSOs %s\n",prefs.dsopath);
	fprintf (FP, "sv_viewer_name %s\n",prefs.imager_sv_path);
	fclose (FP);
	dp (1, "Preferences saved.\n");
}




void save_prefs ()
{
FILE* FP;
char fname [256];
char* home;


	home = getenv ("HOME");
	if (!home)
		ShowMessage ("Cannot determine home directory");
	dp (2, "Home directory is %s\n", home);

	strcpy (fname, home);
	strcat (fname, "/.cimagerc");
	dp (2, "Saving preferences in %s\n",fname);
	FP = fopen (fname, "wt");
	if (!FP) ShowMessage ("Error saving preferences");

	fprintf (FP, "Tiffinfo %d\n", tiffreport);
	fprintf (FP, "Showrulers %d\n", show_rulers);
	fprintf (FP, "Interpolation %d\n",intp_pref);
	fprintf (FP, "Auto_Binning %d\n", autobinning);
	fprintf (FP, "Debuglevel %d\n", debuglevel);
	fprintf (FP, "Skip_license %d\n", skip_license);
	fprintf (FP, "Undo_function %d\n", undo_function);
	fprintf (FP, "Path_to_undo %s\n", undo_path);
	fprintf (FP, "Path_to_Browser %s\n", browspath);
	fprintf (FP, "Path_to_Gnuplot %s\n", gnuplotpath);
	fprintf (FP, "Path_to_Help %s\n", helppath);
	fprintf (FP, "Path_to_DSOs %s\n", dsopath);
	fprintf (FP, "sv_viewer_name %s\n", imager_sv_path);
	fclose (FP);
	dp (1, "Preferences saved.\n");
}



void on_interp_menu_changed (GtkWidget *widget, gpointer user_data)
{
	prefs.intp_prefs = *(int*) user_data;
	dp (2, "Interpolation method set to %d\n",prefs.intp_prefs);
}


void on_undo_menu_changed (GtkWidget *widget, gpointer user_data)
{
	prefs.undo_function = *(int*) user_data;
	dp (2, "Undo function set to %d\n",prefs.undo_function);
}


void on_prefs_tiffbutton_clicked (GtkToggleButton *togglebutton, gpointer user_data)
{
	prefs.tiffreport = gtk_toggle_button_get_active ((GtkToggleButton*) user_data);
	dp (2, "Tiffreport set to %d\n",prefs.tiffreport);
}


void on_prefs_showrulers_clicked (GtkToggleButton *togglebutton, gpointer user_data)
{
	prefs.show_rulers = gtk_toggle_button_get_active ((GtkToggleButton*) user_data);
	dp (2, "Show Rulers set to %d\n",prefs.show_rulers);
}


void on_prefs_autobinbutton_clicked (GtkToggleButton *togglebutton, gpointer user_data)
{
	prefs.autobinning = gtk_toggle_button_get_active ((GtkToggleButton*) user_data);
	dp (2, "Histogram auto-binning set to %d\n",prefs.autobinning);
}


void on_prefs_licensebutton_clicked (GtkToggleButton *togglebutton, gpointer user_data)
{
	prefs.skip_license = gtk_toggle_button_get_active ((GtkToggleButton*) user_data);
	dp (2, "Skip-license set to %d\n",prefs.skip_license);
}


void on_debug_spinner_changed (GtkEditable *editable, gpointer user_data)
{
GtkAdjustment* adj;

	adj = user_data;
	prefs.debuglevel = adj->value;
	dp (2, "Debuglevel set to %d\n",prefs.debuglevel);
}

void on_prefs_cancelbutton_clicked ( GtkButton * button, gpointer data )
{
	gtk_widget_destroy ( GTK_WIDGET ( data ) );
}


void on_prefs_okbutton_clicked ( GtkButton * button, gpointer data )
{
	tiffreport=prefs.tiffreport;
	intp_pref=prefs.intp_prefs;
	debuglevel=prefs.debuglevel;
	skip_license = prefs.skip_license;
	show_rulers = prefs.show_rulers;
	autobinning = prefs.autobinning;

	if (undo_function != prefs.undo_function) undo_stack_invalidate ();
	undo_function=prefs.undo_function;

	memset (prefs.browspath,0,1020);
	strncpy (prefs.browspath,gtk_entry_get_text (GTK_ENTRY(prefs_entry1)),1000);
	strcpy (browspath,prefs.browspath);

	memset (prefs.helppath,0,1020);
	strncpy (prefs.helppath,gtk_entry_get_text (GTK_ENTRY(prefs_entry2)),1000);
	strcpy (helppath,prefs.helppath);

	memset (prefs.dsopath,0,1020);
	strncpy (prefs.dsopath,gtk_entry_get_text (GTK_ENTRY(prefs_entry3)),1000);
	strcpy (dsopath,prefs.dsopath);

	memset (prefs.undo_path,0,1020);
	strncpy (prefs.undo_path,gtk_entry_get_text (GTK_ENTRY(prefs_entry4)),1000);
	strcpy (undo_path,prefs.undo_path);

	memset (prefs.gnuplotpath,0,1020);
	strncpy (prefs.gnuplotpath,gtk_entry_get_text (GTK_ENTRY(prefs_entry6)),1000);
	strcpy (gnuplotpath,prefs.gnuplotpath);

	memset (prefs.imager_sv_path,0,1020);
	strncpy (prefs.imager_sv_path,gtk_entry_get_text (GTK_ENTRY(prefs_entry5)),1000);
	strcpy (imager_sv_path,prefs.imager_sv_path);

	/* Save the new preferences */

	save_prefs ();

	gtk_widget_destroy ( GTK_WIDGET ( data ) );
}


void create_prefswnd (void)
{
  GtkWidget *prefswnd;
  GtkWidget *vbox1;
  GtkWidget *table1;
  GtkWidget *label1, *label2, *label3, *label4, *label5, *label6, *label7;
  GtkWidget *label8, *label9, *label10, *label11, *label12, *label13;
  GtkWidget *hseparator1, *hseparator2;
  GtkWidget *okbutton;
  GtkWidget *cancelbutton;
  GtkWidget *tiffinfobutton;
  GtkWidget *showrulerbutton;
  GtkWidget *autobinbutton;
  GtkWidget *licensebutton;
  GtkWidget *interp_menu;
  GtkWidget *interp_menu_menu;
  GtkWidget *undo_menu;
  GtkWidget *undo_menu_menu;
  GtkWidget *glade_menuitem;
  GtkObject *debug_spinner_adj;
  GtkWidget *debug_spinner;
  int y;

	prefswnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (prefswnd), "prefswnd", prefswnd);
	gtk_window_set_title (GTK_WINDOW (prefswnd), "Preferences");

	vbox1 = vbox_in_window (prefswnd, "vbox1");
	label12 = MakeLabel (prefswnd,vbox1,"(Most preferences require cimage restart to take effect)","label12",-1,-1);
	hseparator1 = MakeHsep (prefswnd,vbox1,"hsep0",-1,-1);

	table1 = MakeTable (prefswnd, vbox1, "table1",2,12);
	y=0;

	label11 = MakeLabel (prefswnd,table1,"Show main image rulers","label11",0,y);
	showrulerbutton = Checkbutton_in_table (prefswnd,table1," ","prefs_showrulers",1,y,NULL);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (showrulerbutton), show_rulers);
	gtk_signal_connect (GTK_OBJECT (showrulerbutton), "toggled", GTK_SIGNAL_FUNC (on_prefs_showrulers_clicked), showrulerbutton);
  	y++;

	label1 = MakeLabel (prefswnd,table1,"Show file import info","label1",0,y);
	tiffinfobutton = Checkbutton_in_table (prefswnd,table1," ","prefs_tiffinfo",1,y,NULL);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (tiffinfobutton), tiffreport);
	gtk_signal_connect (GTK_OBJECT (tiffinfobutton), "toggled", GTK_SIGNAL_FUNC (on_prefs_tiffbutton_clicked), tiffinfobutton);
  	y++;

	label13 = MakeLabel (prefswnd,table1,"Histogram Auto-Binning","label13",0,y);
	autobinbutton = Checkbutton_in_table (prefswnd,table1," ","prefs_autobin",1,y,NULL);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (autobinbutton), autobinning);
	gtk_signal_connect (GTK_OBJECT (autobinbutton), "toggled", GTK_SIGNAL_FUNC (on_prefs_autobinbutton_clicked), autobinbutton);
  	y++;

	label2 = MakeLabel (prefswnd,table1,"Preferred interpolation method","label2",0,y);
	interp_menu = Make_Optionmenu (prefswnd,table1,"interp_menu",1,y);
	interp_menu_menu = gtk_menu_new ();
	glade_menuitem = Option_menuitem (interp_menu_menu,on_interp_menu_changed,"Nearest Neighbor",&menuopt0);
	glade_menuitem = Option_menuitem (interp_menu_menu,on_interp_menu_changed,"Bilinear",&menuopt1);
	glade_menuitem = Option_menuitem (interp_menu_menu,on_interp_menu_changed,"Bicubic",&menuopt2);
	glade_menuitem = Option_menuitem (interp_menu_menu,on_interp_menu_changed,"Spline (2nd order)",&menuopt3);
	glade_menuitem = Option_menuitem (interp_menu_menu,on_interp_menu_changed,"Spline (3rd order)",&menuopt4);
	glade_menuitem = Option_menuitem (interp_menu_menu,on_interp_menu_changed,"Spline (4th order)",&menuopt5);
	glade_menuitem = Option_menuitem (interp_menu_menu,on_interp_menu_changed,"Spline (5th order)",&menuopt6);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (interp_menu), interp_menu_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (interp_menu), intp_pref);
	y++;

	label3 = MakeLabel (prefswnd,table1,"Debug message level (0-3)","label3",0,y);
	debug_spinner_adj = gtk_adjustment_new (debuglevel, 0, 3, 1, 1, 0);
	debug_spinner = MakeSpinbutton (GTK_ADJUSTMENT (debug_spinner_adj),prefswnd,table1,"prefsspin1",1,y);
	gtk_signal_connect (GTK_OBJECT (debug_spinner), "value-changed", GTK_SIGNAL_FUNC (on_debug_spinner_changed),debug_spinner_adj);
	y++;

	label8 = MakeLabel (prefswnd,table1,"Undo function","label8",0,y);
	undo_menu = Make_Optionmenu (prefswnd,table1,"undo_menu",1,y);
	undo_menu_menu = gtk_menu_new ();
	glade_menuitem = Option_menuitem (undo_menu_menu,on_undo_menu_changed,"UNDO disabled",&menuopt0);
	glade_menuitem = Option_menuitem (undo_menu_menu,on_undo_menu_changed,"UNDO in memory",&menuopt1);
	glade_menuitem = Option_menuitem (undo_menu_menu,on_undo_menu_changed,"UNDO stack on disk",&menuopt2);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (undo_menu), undo_menu_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (undo_menu), undo_function);
	y++;

	label7 = MakeLabel (prefswnd,table1,"Temp path for undo","label7",0,y);
	prefs_entry4 = Entry_in_table (prefswnd,table1,prefs.undo_path,"prefs_entry4",1,y);
	y++;

	label4 = MakeLabel (prefswnd,table1,"Path to browser","label4",0,y);
	prefs_entry1 = Entry_in_table (prefswnd,table1,prefs.browspath,"prefs_entry1",1,y);
	y++;
	label4 = MakeLabel (prefswnd,table1,"Path to Gnuplot","label4",0,y);
	prefs_entry6 = Entry_in_table (prefswnd,table1,prefs.gnuplotpath,"prefs_entry1",1,y);
	y++;
	label5 = MakeLabel (prefswnd,table1,"Path to help files","label5",0,y);
	prefs_entry2 = Entry_in_table (prefswnd,table1,prefs.helppath,"prefs_entry2",1,y);
	y++;
	label6 = MakeLabel (prefswnd,table1,"Path to DSO modules","label6",0,y);
	prefs_entry3 = Entry_in_table (prefswnd,table1,prefs.dsopath,"prefs_entry3",1,y);
	y++;
	label9 = MakeLabel (prefswnd,table1,"Path+name of 3D viewer","label9",0,y);
	prefs_entry5 = Entry_in_table (prefswnd,table1,prefs.imager_sv_path,"prefs_entry5",1,y);
	y++;

	label10 = MakeLabel (prefswnd,table1,"Suppress license splash","label10",0,y);
	licensebutton = Checkbutton_in_table (prefswnd,table1," ","prefs_license",1,y,NULL);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (licensebutton), skip_license);
	gtk_signal_connect (GTK_OBJECT (licensebutton), "toggled", GTK_SIGNAL_FUNC (on_prefs_licensebutton_clicked), licensebutton);
	y++;

	hseparator1 = MakeHsep (prefswnd,table1,"hsep1",0,y);
	hseparator2 = MakeHsep (prefswnd,table1,"hsep2",1,y);
	y++;

	okbutton = Button_in_table (prefswnd,table1,"OK","prefs_ok",0,y,on_prefs_okbutton_clicked);
	cancelbutton = Button_in_table (prefswnd,table1,"Cancel","prefs_cancel",1,y,on_prefs_cancelbutton_clicked);

	gtk_widget_show (prefswnd);

}


void preferences_main ()
{

	prefs.tiffreport=tiffreport;
	prefs.show_rulers = show_rulers;
	prefs.skip_license=skip_license;
	prefs.intp_prefs=intp_pref;
	prefs.autobinning = autobinning;
	prefs.debuglevel=debuglevel;
	prefs.undo_function = undo_function;
	strcpy (prefs.dsopath, dsopath);
	strcpy (prefs.browspath,browspath);
	strcpy (prefs.helppath,helppath);
	strcpy (prefs.gnuplotpath,gnuplotpath);
	strcpy (prefs.undo_path, undo_path);
	strcpy (prefs.imager_sv_path, imager_sv_path);
	create_prefswnd ();
}


