

/*
	wvlt_core.c : Core functions for wavelet transforms
	Please note that this module is somewhat experimental
	and may still contain hidden bugs.

	Credits: The core functions of the DWT, particularly the 
	subband filters, are leaning heavily on Numerical Recpies in C,
	second edition, chapter 13.

 */


/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif


#include <stdio.h>
#include <string.h>
#include <math.h>

#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "application.h"
#include "dlgbox.h"
#include "filter.h"
#include "functions.h"
#include "functions2.h"
#include "graphdsp.h"
#include "gtkhelpers.h"
#include "imgbuf.h"
#include "secondwnd.h"
#include "viz.h"
#include "wvlt_core.h"






/***************************************************************************************************************

  ***    *   *   ****    ****      *     *   *   ****            *****   *       *****
 *   *   *   *   *   *   *   *    * *    *   *   *   *           *       *         *
 *       *   *   *   *   *   *   *   *   **  *   *   *           *       *         *
  ***    *   *   ****    ****    *   *   * * *   *   *           ****    *         *
     *   *   *   *   *   *   *   *****   *  **   *   *           *       *         *
 *   *   *   *   *   *   *   *   *   *   *   *   *   *           *       *         *
  ***     ***    ****    ****    *   *   *   *   ****            *       *****     *

*/



/* The subband filters perform one step of DWT decomposision or synthesis.
	These subband filters operate on an 1D array of float data (length
	must be a power of two or - less strictly - an even number) and return
	the scaling filter coefficients in the lower half of the array and the 
	wavelet coefficients in the upper half of the array.
	The variable idwt determines whether we want a forward DWT (idwt==0)
	or an inverse DWT (idwt!=0).

	A note on the use of float versus double: Considerable rounding errors
	exist when float is used throughout. For this reason, the actual
	convolution is performed with double precision. Only when compatibility
	with images (B_FLOAT) is needed, do we revert to single-precision.
*/



void dead_subband (double *data, unsigned long n, DWTDIRECTION idwt, wavefilt dummy)
{
	dp (0, "Dead subband filter. Should not happen.\n");
}



/*
	Initializing routine for general_subband. Here, the wavefilt structure is initialized and
	filled with the wavelet and scale parameters.
	This function MUST be called before general_subband can be used.
	Note that the standalone functions in the wavelet *module* make use of specific 
	parameter set lengths and are therefore faster than general_subband.
*/


/* w00t! Check out http://wavelets.pybytes.com/ */


wavefilt wavelet_prepare (WAVELET_TYPE wlt)
{
int k;
double sgn;
wavefilt wfilt;

	/* scale filter (lowpass) parameters h(k). */

static double c_haar[2] = 	{	0.7071067118655,    0.7071067118655 };				/* Haar */

static double c_daub4[4] =	{	0.4829629131445341, 0.8365163037378079,				/* DAUB4, 2 vanishing moments */
								0.2241438680420134,-0.1294095225512604 };

static double c_daub8[8] = 	{	 0.23037781330890,  0.71484657055292, 				/* DAUB8, 4 vanishing moments */
								 0.63088076792986, -0.02798376941686,
								-0.18703481171909,  0.03084138183556,
								 0.03288301166689, -0.01059740178507 };

static double c_daub12[12]=	{	0.111540743350, 0.494623890398, 0.751133908021,		/* DAUB12, 6 vanishing moments */
								0.315250351709,-0.226264693965,-0.129766867567,
								0.097501605587, 0.027522865530,-0.031582039318,
								0.000553842201, 0.004777257511,-0.001077301085 };

static double c_daub20[20] =	{	 0.026670057901, 0.188176800078, 0.527201188932,	/* DAUB20, 10 vanishing moments */
									 0.688459039454, 0.281172343661,-0.249846424327,
									-0.195946274377, 0.127369340336, 0.093057364604,
									-0.071394147166,-0.029457536822, 0.033212674059,
									 0.003606553567,-0.010733175483, 0.001395351747,
									 0.001992405295,-0.000685856695,-0.000116466855,
									 0.000093588670,-0.000013264203 };

static double c_coif6[6] = 	{	-0.07273261951285,  0.33789766245781,				/* COIF6, 2 vanishing moments */
								 0.85257202021226,  0.38486484686420,
								-0.07273261951285, -0.01565572813546 };

static double c_coif12[12] =	{	 0.016387336463, -0.041464936781,				/* COIF12, 4 vanishing moments */
									-0.067372554722,  0.386110066823,
									 0.812723635449,  0.417005184423,
									-0.076488599078, -0.059434418646,
									 0.023680171946,  0.005611434819,
									-0.001823208870, -0.000720549446 };

static double c_symlet4[8] =	{	 0.032223100604042702, -0.012603967262037833,
									-0.099219543576847216,  0.29785779560527736,
									 0.80373875180591614,   0.49761866763201545,
									-0.02963552764599851,  -0.075765714789273325};


static double c_symlet8[16] =	{	 0.0018899503327594609,  -0.0003029205147213668,	/* Symlet */
									-0.014952258337048231,    0.0038087520138906151,
									 0.049137179673607506,   -0.027219029917056003,
									-0.051945838107709037,    0.3644418948353314,
									 0.77718575170052351,     0.48135965125837221, 
									-0.061273359067658524,   -0.14329423835080971,
									 0.0076074873249176054,   0.031695087811492981,
									-0.00054213233179114812, -0.0033824159510061256 };

static double c_symlet12[24] =	{	-0.00017906658697508691, -1.8158078862617515e-05, 
									 0.0023502976141834648,   0.00030764779631059454,
									-0.014589836449234145,   -0.0026043910313322326,
									 0.057804179445505657,    0.01530174062247884,
									-0.17037069723886492,    -0.07833262231634322,
									 0.46274103121927235,     0.76347909778365719,
									 0.39888597239022,       -0.022162306170337816,
									-0.035848830736954392,    0.049179318299660837,
									 0.0075537806116804775,  -0.024220722675013445,
									-0.0014089092443297553,   0.007414965517654251,
									 0.00018021409008538188, -0.0013497557555715387,
									-1.1353928041541452e-05,  0.00011196719424656033};



	wfilt.n_param=0;
	
	if (wlt==W_HAAR) 
	{
		wfilt.n_param=2;
		wfilt.hk=c_haar;
	}
	else if (wlt==W_DAUB4) 
	{
		wfilt.n_param=4;
		wfilt.hk=c_daub4;
	}
	else if (wlt==W_DAUB8) 
	{
		wfilt.n_param=8;
		wfilt.hk=c_daub8;
	}
	else if (wlt==W_DAUB12) 
	{
		wfilt.n_param=12;
		wfilt.hk=c_daub12;
	}
	else if (wlt==W_DAUB20) 
	{
		wfilt.n_param=20;
		wfilt.hk=c_daub20;
	}
	else if (wlt==W_COIF6) 
	{
		wfilt.n_param=6;
		wfilt.hk=c_coif6;
	}
	else if (wlt==W_COIF12) 
	{
		wfilt.n_param=12;
		wfilt.hk=c_coif12;
	}
	else if (wlt==W_SYM4) 
	{
		wfilt.n_param=8;
		wfilt.hk=c_symlet4;
	}
	else if (wlt==W_SYM8) 
	{
		wfilt.n_param=16;
		wfilt.hk=c_symlet8;
	}
	else if (wlt==W_SYM12) 
	{
		wfilt.n_param=24;
		wfilt.hk=c_symlet12;
	}
	else 
		ShowMessage ("Error setting up wavelet parameters");

	wfilt.gk = (double*) calloc (wfilt.n_param, sizeof (double));

	/* Generate the wavelet filter parameters g(k) from the scale filter h(k) */
	
	sgn = -1.0;
	for (k=0; k<wfilt.n_param; k++) 
	{
		wfilt.gk[wfilt.n_param-1-k]=sgn*wfilt.hk[k];
		sgn = -sgn;
	}
	
	dp (2, "Setting up filter parameters: You have:\n");
	dp (2, "Lowpass:  ");
	for (k=0; k<wfilt.n_param; k++) dp (2, "%7.4f ",wfilt.hk[k]); dp (2,"\n");
	dp (2, "Highpass: ");
	for (k=0; k<wfilt.n_param; k++) dp (2, "%7.4f ",wfilt.gk[k]); dp (2,"\n");
	

/*	
	These values center the "support" of the wavelets at each level. Alternatively, the "peaks"
	of the wavelets can be approximately centered by the choices hk_offs=-2 and gk_offs=-n+2.
	Note that daub4 and this function with n=4 use different default centerings.
*/
	wfilt.hk_offs = wfilt.gk_offs =  0; //-(n >> 1);
	return wfilt;

}



/*
	Generalized one-stage subband filter: applies an arbitrary wavelet filter to data  (for idwt==0)
	or applies its transpose (for idwt==1). Used hierarchically by the actual DWT functions.
	The actual filter is determined by a preceding (a required) call to wavelet_prepare, which initializes
	the structure wfilt.
*/


void general_subband (double *data, unsigned long n, DWTDIRECTION idwt, wavefilt wfilt)
{
double *buf;
unsigned long i,j,k,bitmask,ni,nj,nh,nmod;

	if (n < 4) return;
	
	buf= (double *) calloc (n+1, sizeof (double));		/* temporary storage buffer */

	nmod=wfilt.n_param*n; 				/* A positive constant equal to zero mod n. */
	bitmask=n-1;						/* Mask of all bits, since n is a power of 2. */
	nh = n >> 1;						/* Start of highpass data */

	if (idwt == W_FORWARD)				/* Forward subband coding filter (forward one-level DWT). */
	{
		j=0;
		for (i=0; i<n; i+=2) 			/* Coefficient loop */
		{
			ni = i+nmod+wfilt.hk_offs;				/* Index into data to be incremented and wrapped-around. */
			nj = i+nmod+wfilt.gk_offs;				/* essentially i + a large multiple of n + offset */

			for (k=0; k<wfilt.n_param; k++)			/* Convolution loop */
			{
				buf[j]    += wfilt.hk[k]*data[(ni+k) & bitmask];	/* Scale filter (lowpass) data */
				buf[j+nh] += wfilt.gk[k]*data[(nj+k) & bitmask];	/* Wavelet filter (highpass) data */
			}
			j++;
		}
	} 
	else 			/* Inverse filter (inverse DWT). */
	{
		j=0;
		for (i=0; i<n; i+=2) 
		{
			ni = i+nmod+wfilt.hk_offs;
			nj = i+nmod+wfilt.gk_offs;
			for (k=0; k<wfilt.n_param; k++) 
			{
				buf[(ni+k) & bitmask] += wfilt.hk[k]*data[j];
				buf[(nj+k) & bitmask] += wfilt.gk[k]*data[j+nh];
			}
			j++;
		}
	}
	
	for (i=0; i<n; i++) data[i]=buf[i];        /* Copy the results back from workspace.*/
	free (buf);
}



/* This function initializes the subband() function pointer.
*/


WAVELET_TYPE set_subband (int wavelet)
{
WAVELET_TYPE wtype;
	
	wtype = W_NULL;
	switch (wavelet)
	{
		case 0:						/* Haar */
				wtype=W_HAAR;
				subband = &general_subband;
				strcpy (wvlt_used, "Haar");
				break;
		case 1:						/* DAUB4 */
				wtype=W_DAUB4;
				subband = &general_subband;
				strcpy (wvlt_used, "Daub4");
				break;
		case 2:						/* DAUB8 */
				wtype=W_DAUB8;
				subband = &general_subband;
				strcpy (wvlt_used, "Daub8");
				break;
		case 3:						/* DAUB12 */
				wtype=W_DAUB12;
				subband = &general_subband;
				strcpy (wvlt_used, "Daub12");
				break;
		case 4:						/* DAUB20 */
				wtype=W_DAUB20;
				subband = &general_subband;
				strcpy (wvlt_used, "Daub20");
				break;
		case 5:						/* Coiflet-6 */
				wtype=W_COIF6;
				subband = &general_subband;
				strcpy (wvlt_used, "Coif6");
				break;
		case 6:						/* Coiflet-12 */
				wtype=W_COIF12;
				subband = &general_subband;
				strcpy (wvlt_used, "Coif12");
				break;
		case 7:						/* Symlet 4 */
				wtype=W_SYM4;
				subband = &general_subband;
				strcpy (wvlt_used, "Symlet4");
				break;
		case 8:						/* Symlet 8 */
				wtype=W_SYM8;
				subband = &general_subband;
				strcpy (wvlt_used, "Symlet8");
				break;
		case 9:						/* Symlet12 */
				wtype=W_SYM12;
				subband = &general_subband;
				strcpy (wvlt_used, "Symlet12");
				break;

		default:
				subband = &dead_subband;
				return 0;
	}
	return wtype;
}



/*************************************************************************

 ####    #   #   #####
 #   #   #   #     #
 #   #   #   #     #
 #   #   # # #     #
 #   #   # # #     #
 #   #   # # #     #
 ####     # #      #

*************************************/


/*
	Here are the pyramidal deconstruction/reconstruction algorithms
	for 1D data, n-dimensional data etc. */


/*
	One-dimensional discrete wavelet transform. This routine implements the pyramid algorithm,
	replacing data[0..n-1] by its wavelet transform (for idwt=0), or performing the inverse operation
	(for idwt=1). Note that n MUST be an integer power of 2. The routine wt_subband, whose
	actual name must be supplied in calling this routine, is the underlying wavelet filter. Examples
	of wt_subband are daub4 and (preceded by wavelet_prepare) general_subband.
*/

void dwt_1d (double *data, unsigned long n, DWTDIRECTION idwt, wavefilt wfilt, void (*wt_subband)(double *, unsigned long, DWTDIRECTION, wavefilt))
{
unsigned long i;


	if (n < 4) return;

	if (!idwt)						/* Forward Wavelet transform. */
	{
		for (i=n; i>=4; i>>=1) 
			(*wt_subband)(data, i, idwt, wfilt);		/* Start at largest hierarchy, and work towards smallest.  */
		
	} 
	else 							/* Inverse wavelet transform.*/
	{
		for (i=4; i<=n; i<<=1) 
			(*wt_subband)(data, i, idwt, wfilt);		/* Start at smallest hierarchy, and work towards largest. */
	}
}


/* The same as above, but with a limited number of levels */

void dwt_1dn (double *data, unsigned long n, DWTDIRECTION idwt, int levels, wavefilt wfilt, void (*wt_subband)(double *, unsigned long, DWTDIRECTION, wavefilt))
{
unsigned long i;
int l;


	if (n < 4) return;

	if (!idwt)						/* Forward Wavelet transform. */
	{
		i=n; l=levels;
		while ((i>=4) && (l>1))
		{
			(*wt_subband)(data, i, idwt, wfilt);		/* Start at largest hierarchy, and work towards smallest.  */
			i >>= 1;
			l--;
		}
		
	} 
	else 							/* Inverse wavelet transform.*/
	{
		i = n; l=levels;
		while (l>1) { i>>=1; l--; }
		l = levels;
		while ((i<=n) && (l>1))
		{
			(*wt_subband)(data, i, idwt, wfilt);		/* Start at smallest hierarchy, and work towards largest. */
			i <<= 1;
			l--;
		}
	}
}



/* Simplistic slice-by-slice 2D wavelet transform. This function
	transforms the current image and returns a float image. 
	NOTE: We have two versions here. The first one is incorrect as it subdivides
	the L/H and H/L areas. The second version performs a subdivision in 2D by one level only,
	then restricts (halves) the image size, performs a 1-level subdivision on the
	remaining area etc.
*/


void dwt_2d_messedup (image_type *img, DWTDIRECTION idwt, int decomp_level, wavefilt wfilt, void (*wt_subband)(double *, unsigned long, DWTDIRECTION, wavefilt))
{
image_type result;
double *buf1, *buf2;
float *p;
int nx, ny;
int x,y,z,i, level;


	/* Expand to nearest power of two */
	
	nx = 1; while (nx<img->xmax) nx=nx<<1;
	ny = 1; while (ny<img->ymax) ny=ny<<1;
	if ((nx<4) || (ny<4))
	{
		ShowMessage ("Not enough data for DWT");
		return;
	}

	/* Get memory */
	
	if (allocate_image (&result, B_FLOAT,nx, ny, img->zmax) ==-1)
		fatal_memerr();
	p = result.data;
	buf1 = calloc (nx, sizeof(double));
	buf2 = calloc (ny, sizeof(double));

	if (idwt==W_INVERSE) 
		dp (1, "2D inverse wavelet transform, %d level(s)...\n", decomp_level);
	else
		dp (1, "2D wavelet transform, %d level(s)...\n", decomp_level);


	for (z=0; z<img->zmax; z++)
	{

		/* Full pyramidal transform in X direction */

		for (y=0; y<img->ymax; y++)
		{
			for (x=0; x<nx; x++) buf1[x]=0;
			for (x=0; x<img->xmax; x++) buf1[x] = readbuf_flt (img, x,y,z);
		
			if (idwt==W_FORWARD)				/* Forward Wavelet transform. */
			{
				i = nx; level=decomp_level;
				while ((i>=4) && (level>0))
				{
					dp (3, "2D DWT in X: Subband filter, level %d, size %d\n",level,i);
					(*wt_subband) (buf1, i, idwt, wfilt);		/* Start at largest hierarchy, and work towards smallest.  */
					i = i>>1;							/* Divide subband active range by two */
					level--;
				}
				
			} 
			else 							/* Inverse wavelet transform.*/
			{
				i=nx; level = decomp_level;
				while (level>1) { i>>=1; level--; }
				level = decomp_level;
				while ((i<=nx) && (level>0))
				{
					dp (3, "2D IDWT in X: Subband filter, level %d, size %d\n",level,i);
					(*wt_subband) (buf1, i, idwt, wfilt);		/* Start at smallest hierarchy, and work towards largest. */
					i = i<<1;
					level--;
				}
			}
	
			for (x=0; x<nx; x++) p[x + nx*(y+ny*z)] = buf1[x];
		}
	
		/* Full pyramidal transform in Y direction */
	
		for (x=0; x<nx; x++)
		{
			for (y=0; y<ny; y++) buf2[y]=p[x + nx*(y+ny*z)];
		
			if (idwt==W_FORWARD)						/* Forward Wavelet transform. */
			{
				i = ny; level=decomp_level;
				while ((i>=4) && (level>0))
				{
					dp (3, "2D DWT in Y: Subband filter, level %d, size %d\n",level,i);
					(*wt_subband) (buf2, i, idwt, wfilt);		/* Start at largest hierarchy, and work towards smallest.  */
					i = i>>1;							/* Divide subband active range by two */
					level--;
				}
				
			} 
			else 							/* Inverse wavelet transform.*/
			{
				i=ny; level = decomp_level;
				while (level>1) { i>>=1; level--; }
				level = decomp_level;
				while ((i<=ny) && (level>0))
				{
					dp (3, "2D IDWT in Y: Subband filter, level %d, size %d\n",level,i);
					(*wt_subband) (buf2, i, idwt, wfilt);		/* Start at smallest hierarchy, and work towards largest. */
					i = i<<1;
					level--;
				}
			}
	
			for (y=0; y<ny; y++) p[x + nx*(y+ny*z)] = buf2[y];
		}
	}

	make_minmax (&result);
	handback (img, &result);

}


/* The second version is based on repeated single-level 2D decompositions.
	dwt_2d_onelevel performs the two orthogonal decompositions in an image-sized buffer.
	The transform is done in-place. The buffer may have the full image size, and the stride
	parameter takes care that the correct row start is used, and that row-spaced elements are 
	selected for the y-direction.

	One additional advantage of one-level orthogonal transforms is that we remain within
	double precision for both directions.
*/


void dwt_2d_onelevel (double *data, DWTDIRECTION idwt, int nx, int ny, int stride, wavefilt wfilt, void (*wt_subband)(double*, unsigned long, DWTDIRECTION, wavefilt))
{
double *buf1, *buf2;
int x,y,i;


	buf1 = calloc (nx, sizeof(double));
	buf2 = calloc (ny, sizeof(double));		/* Mere laziness. nx may be different from ny. */

	if (idwt==W_INVERSE) 
		dp (2, "     2D one-level inverse wavelet transform, (0,0) to (%d,%d) with stride %d\n", nx-1, ny-1, stride);
	else
		dp (2, "     2D one-level wavelet transform, (0,0) to (%d,%d) with stride %d\n", nx-1, ny-1, stride);


	/* One-level transform in X direction. Since the y-direction is the slower
		direction, the pointer increments not by nx, but by stride (which is the
		xmax value of the full image). This goes for both directions. */

	for (y=0; y<ny; y++)			/* Line-by-line */
	{
		for (x=0; x<nx; x++) buf1[x] = data[x + stride*y];
		(*wt_subband) (buf1, nx, idwt, wfilt);
		for (x=0; x<nx; x++) data[x + stride*y] = buf1[x];
	}

	/* One-level transform in Y direction. */

	for (x=0; x<nx; x++)			/* Column-by-column */
	{
		for (y=0; y<ny; y++) buf2[y] = data[x + stride*y];
		(*wt_subband) (buf2, ny, idwt, wfilt);
		for (y=0; y<ny; y++) data[x + stride*y] = buf2[y];
	}

	free (buf1); free (buf2);

}


/* 2D DWT/IDWT for a 2D array of double-precision pixels. Used for denoising and for the independent
	DWT/IDWT from an image buffer. 

	The DWT is executed in-place, and data is modified.
	data is assumed to have power-of-two dimensions, and padding is not required (and not done).
*/

void dwt_2d (double *data, DWTDIRECTION idwt, int xmax, int ymax, int decomp_level, wavefilt wfilt, void (*wt_subband)(double *, unsigned long, DWTDIRECTION, wavefilt))
{
int nx, ny;
int x,y, level;


	if (idwt==W_INVERSE) 
		dp (1, "2D inverse wavelet transform, %d level(s), data size (%d, %d)...\n", decomp_level, xmax, ymax);
	else
		dp (1, "2D wavelet transform, %d level(s), data size (%d, %d)...\n", decomp_level, xmax, ymax);

	nx = xmax;
	ny = ymax;

	if (idwt==W_FORWARD)				/* Forward Wavelet transform. */
	{
		level=decomp_level;
		while ((nx>=4) && (ny>=4) && (level>0))
		{
			dwt_2d_onelevel (data, idwt, nx, ny, xmax, wfilt, wt_subband);
			nx = nx >> 1;								/* Divide subband active range by two */
			ny = ny >> 1;
			level--;
		}

	} 
	else 							/* Inverse wavelet transform.*/
	{
		level = decomp_level;

		//while (level>1) 						/* Which of the dry runs is it?? */
		while ((nx>=4) && (ny>=4) && (level>1))
		{ 
			nx >>= 1; ny >>= 1;
			level--;
		}

		level = decomp_level;
		while ((nx <= xmax) && (ny <= ymax) && (level>0))
		{
			dwt_2d_onelevel (data, idwt, nx, ny, xmax, wfilt, wt_subband);
			nx = nx << 1;								/* Expand subband active range by factor of two */
			ny = ny << 1;
			level--;
		}
	}

}




void dwt_2d_image (image_type *img, DWTDIRECTION idwt, int decomp_level, wavefilt wfilt, void (*wt_subband)(double *, unsigned long, DWTDIRECTION, wavefilt))
{
image_type result;
float *p;
double *buf;
int nx, ny;
int x,y,z,x1,y1, level;


	/* Expand to nearest power of two */
	
	nx = 1; while (nx<img->xmax) nx=nx<<1;
	ny = 1; while (ny<img->ymax) ny=ny<<1;
	if ((nx<4) || (ny<4))
	{
		ShowMessage ("Not enough data for DWT");
		return;
	}

	/* Get memory for result image. The onelevel transform operates in-place.
		We *could* let it operate in-place if the source image is guaranteed B_FLOAT,
		but it isn't, so we first need to dumb-copy the source image to the target (result).
		Dumb, because it doesn't do well with RGB images. As usual... sigh.
	*/
	
	if (allocate_image (&result, B_FLOAT, nx, ny, img->zmax) ==-1)
		fatal_memerr();
	p = result.data;

	buf = calloc (nx*ny, sizeof(double));		/* 2D slice buffer for actual DWT */


	if (idwt==W_INVERSE) 
		dp (1, "2D-image inverse wavelet transform, %d level(s)...\n", decomp_level);
	else
		dp (1, "2D-image wavelet transform, %d level(s)...\n", decomp_level);


	for (z=0; z<img->zmax; z++)
	{
		/* First, copy the current z-slice into the result image buffer, converting
			it from float to double. We remain at double precision for the entire 2D
			decomposition / reconstruction and only stuff it back in float for the result image.
			We can assume that buf is zero-padded due to the calloc used above. */

		nx = result.xmax; ny = result.ymax;
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++) 
				buf[x + nx*y] = readbuf_flt (img, x,y,z);

		/* Use the data-based dwt_2d function for the actual DWT/IDWT */

		dwt_2d (buf, idwt, nx,ny, decomp_level, wfilt, wt_subband);

		/* Copy the one-slice buffer back into the result image buffer,
			thereby losing the double precision. */

		for (y=0; y<result.ymax; y++)
			for (x=0; x<result.xmax; x++) 
				p[x + result.xmax*(y+result.ymax*z)] = buf[x + result.xmax*y];

	}

	make_minmax (&result);
	handback (img, &result);

}









/***************************************************************************************************************

 *   *   *   *   *       *****            ***    *   *   ****    ***   *   *   *   *   ***   *   *    ***
 *   *   *   *   *         *             *   *   *   *   *   *    *    *   *   *  *     *    *   *   *   *
 *   *   *   *   *         *             *       *   *   *   *    *    **  *   * *      *    **  *   *
 * * *   *   *   *         *              ***    *****   ****     *    * * *   **       *    * * *   *  **
 * * *   *   *   *         *                 *   *   *   * *      *    *  **   * *      *    *  **   *   *
 * * *    * *    *         *             *   *   *   *   *  *     *    *   *   *  *     *    *   *   *   *
  * *      *     *****     *              ***    *   *   *   *   ***   *   *   *   *   ***   *   *    ***
 
*/





/* Numerical Recipes median finder. Function to help finding the n-th quantile etc.
	The input data array has n values (from 0 to n-1). This function
	returns the k-th smallest value in the float array data. The input array will be rearranged
	to have this value in location data[k], with all smaller elements moved to data[0..k-2] 
	(in arbitrary order) and all larger elements in data[k..n-1] (also in arbitrary order).
	NOTE: FUNCTION CHANGES ELEMENT ORDER OF INPUT ARRAY.
*/

#define WVSWAP(A,B) temp=(A);(A)=(B);(B)=temp;

double element_finder (double *data, int k, int n)
{
int i,rght,j,lft,mid;
double a, temp;
unsigned char abort;
	
	lft=0;
	rght=n-1;
	abort=FALSE;
	
	while (TRUE)					/* Exit this loop through the RETURN below */
	{
		if (rght <= lft+1) 				/* Active partition contains 1 or 2 elements. */
		{
			if ((rght == lft+1) && (data[rght] < data[lft])) 		/* Case of 2 elements. */
			{
				WVSWAP (data[lft], data[rght])
			}
			return data[k];
		} 
		else 
		{
			mid=(lft+rght) >> 1;					/* Choose median of left, center, and right elements... */
			WVSWAP(data[mid],data[lft+1]);			/* ... as partitioning element a. */
			
			if (data[lft] > data[rght]) 			/* Also, rearrange so that data[l]  data[l+1], data[ir]  data[l+1] */
				WVSWAP(data[lft],data[rght]);
			
			if (data[lft+1] > data[rght])
				WVSWAP(data[lft],data[rght]);
			
			if (data[lft] > data[lft+1]) 
				WVSWAP(data[lft],data[lft+1]);

			i=lft+1;								/* Initialize pointers for partitioning. */
			j=rght;
			a=data[lft+1];							/* Partitioning element. */
			abort=FALSE;
			while (!abort)							/* Begin innermost loop */
			{
				do 
					i++; 							/* Scan up to find element > a.*/
				while (data[i] < a);
				do 
					j--; 							/* Scan down to find element < a. */
				while (data[j] > a);
				if (j < i) 
					abort=TRUE;					/* Pointers crossed. Partitioning complete. */
				else
					WVSWAP(data[i],data[j]);
			}

			data[lft+1]=data[j];					/* Insert partitioning element. */
			data[j]=a;
			if (j >= k) rght=j-1;					/* Keep active the partition that contains the kth element. */
			if (j <= k) lft=i;
		}
	}
}


#undef WVSWAP




/* Shrinking and thresholding function. Acts on element elem with threshold t.
	Choice is 
				hardsoft=0: Soft thresholding (=shrinking in the strict sense)
				hardsoft=1: hard thresholding
	Threshold t must be positive!
	Returns the shrunk/thresholded element
*/


double wv_thresh (double elem, double t, int hardsoft)
{
	if (!hardsoft)			/* ==0 -> soft thresholding */
	{
		if (fabs(elem) <= fabs (t))
			return 0.0;
		else
			return SIGN(elem)*(fabs(elem)-fabs(t));
	}
	else					/* hard thresholding */
	{
		if (fabs(elem) > fabs (t))
			return elem;
		else
			return 0.0;
	}
}



#define C06745		1.48258			/* this is 1/0.6745 */


/* Estimate Sigma according to Donoho from the highpass coefficients
	of the first decomposition level (i.e., upper half of the signal).
	In its 2D version, we use the H/H square *only* . Therefore, we don't need
	any x,y delimiters, because the H/H area goes from xmax/2 to xmax and
	from ymax/2 to ymax.
*/

double estimate_sigma_1d (double *data, int len)
{
int i,j;
double *p;
double w;


	p = calloc (len/2+10, sizeof (double));
	j=0;
	for (i=len/2; i<len; i++)					/* Find minimum and maximum absolute value in data stream */
	{
		p[j++] = fabs(data[i]);
	}

	w = element_finder (p, j/2, j);	/* Note that this implies an odd n (which we don't have), 
												but with large n the difference is minimal */
	free (p);
	return w*C06745;						/* Donohan & Johnstone estimated standard deviation */
}


/* estimate_sigma_2d is the corresponding 2D function, which operates on a
	single slice of dimensions xmax, ymax. data is supposed to be a 2D decomposition,
	and only the H/H quadrant is examined. data is double-precision and can be 
	used independent from the image structure. */


double estimate_sigma_2d (double *data, int xmax, int ymax)
{
int i,n,x,y,xm,ym;
double *p;
double w;

	xm = xmax/2;
	ym = ymax/2;

	n = (xmax-xm+1)*(ymax-ym+1);			/* Number of elements, generously over-estimated */
	p = calloc (n, sizeof (double));

	i=0;
	for (y=ym; y<ymax; y++)					/* H/H area of level 1 only */
		for (x=xm; x<xmax; x++)
		{
			p[i++] = fabs(data[x + xmax*y]);
		}

	w = element_finder (p, i/2, i);
	free (p);
	return w*C06745;						/* Donohan & Johnstone estimated standard deviation */
}


/* estimate_sigma_2d_image is a variation of the above, in which the source data is
	taken from an image structure. img is supposed to be the result of a DWT
	(thus, B_FLOAT, although the function is fault-tolerant). Only the H/H region 
	is examined. It is not worth filling a buffer and passing it to estimate_sigma_2d,
	so this function is standalone. */


double estimate_sigma_2d_image (image_type *img, int z)
{
int i,n,x,y,xm,ym,xn,yn;
double *p;
double w;

	xn = img->xmax; xm = xn/2;
	yn = img->ymax; ym = yn/2;

	n = (xn-xm+1)*(yn-ym+1);						/* Number of elements, generously over-estimated */
	p = calloc (n, sizeof (double));

	i=0;
	for (y=ym; y<yn; y++)				/* H/H area of level 1 only */
		for (x=xm; x<xn; x++)
		{
			p[i++] = fabs(readbuf_flt (img,x,y,z));
		}

	w = element_finder (p, i/2, i);		/* Note that this implies an odd n (which we don't have), */
	free (p);
	return w*C06745;						/* Donohan & Johnstone estimated standard deviation */
}





/* Wavelet shrinking threshold finder. We pass a 1D array of wavelet coefficients
	(length N) as well as a noise estimate sigma. Sigma can be provided from
	a priori knowledge, or the above Donoho median coefficient estimator can
	be used.

	This function uses the entire 1D dataset, so any calling function needs
	to copy the relevant subband and pass it (or use a suitable pointer).
	Different thresholds can be found for different decomposition levels.
	wshrinkfunc determines the threshold algorithm:
		0: Donoho universal threshold 
		1: BayesShrink, with universal threshold as upper limit
		2: SUREShrink threshold
		3: GCV minimizing threshold

	NOTE that the input field, data, must not be destroyed!

	TODO: Scaling behavior is wrong. Required: If image values scale by a, then the
	threshold also must scale by a. What we have is:
		Universal/VisuShrink: 		Scales correctly. This is due to sigma scaling correctly.
		BayesShrink:				Scaling factor cancels out, does not scale at all.
		SUREShrink:					Scales correctly (implicit in choosing the wavelet coefficient with minimum risk)
		GCV:						Scales correctly (this is implicit in the minimum finder)


*/

double gcv_eval (double *data, int len, double T)			/* Calculates the GCV for a given threshold T */
{															/* T must be a positive value! */
int i, cnt;
double sumxx, buf, buf2;

	sumxx = 0.0; cnt=0;
	for (i=0; i<len; i++)
	{
		buf = fabs(data[i]);
		buf2 = buf - T;					/* Soft thresholding (required) */
		if (buf2 <= 0.0)				/* Smaller than threshold, clamp to zero */
		{
			buf2 = 0.0;
			cnt++;						/* Count zeroed coefficients */
		}
		sumxx += SQR(buf-buf2);			/* Add up the squared difference between original & shrunk coef */
	}

	buf = (double)(cnt) * cnt;			/* To avoid integer overflow, separate the division or convert cnt to double */

	dp (3, "   GCV eval: T=%f, sumsqr=%f, zeroed: %d, gcv=%f\n", T, sumxx, cnt, (cnt==0? 4.3e9 : len * sumxx / buf ));

	if (cnt)							/* Some coefficient were zeroed -> GCV is defined. */
		return len * sumxx / buf;
	else
		return 4.294967296e9;			/* Invalid GCV, return a large value that can never become the minimum */
}



double find_wsthreshold (double *data, int len, double sigma, int wshrinkfunc)
{
int  i,j,cnt,idx;
double T, w, wsigma,min,max, buf, buf2;
double t1, t2, t3, t4, golden_ratio;				/* For the minimum search algorithm */
double gcv3, gcv4, risk, minrisk, sumxx;
double *p;



	T = sigma * sqrt (2*SAFELOG(len));		/* Donoho's universal threshold, which is used as an upper
												bound in most cases. Which log? In Donoho's WaveLab 850,
												ln (log-e) is used. */

	switch (wshrinkfunc)
	{
		case 0:							/* Universal threshold: We can directly return T */

			dp (2, "      %d data points yield 'universal' threshold %f\n",len,T);
			return T;

		case 1:							/* Bayes shrink (level-adaptive) */
										/* BayesShrink calculates the threshold depending on the assumed
											image noise component and the subband's signal variance. Let's denote
											V(n) the noise component. We can use the median-absolute detail estimator
											(median absolute H/H coefficient, see estimate_sigma_2d() ). We also
											denote V(s) the subband signal variance. To obtain V(s), we need
											the subband variance, which is the mean of squared subband coefficients,
											i.e. Vj = sum(wj)/N where wj are the wavelet coefficients of subband j
											and N is the number of coefficients in the subband. Then, we obtain
											V(s) = Vj - V(n) as the subband variance reduced by the noise
											variance. Finally, the threshold is T = V(n) / V(s).
											NOTE: This formulation does NOT scale with image values, but it should.
											If we assume that the data stream is normalized to sigma=1 (true??)
											then we can multiply with sigma (again, sigma^3) to have proportional 
											scaling. But this conflicts with V(s) = Vj - V(n).
										*/
			wsigma = 0; w=0;
			for (i=0; i<len; i++)
				wsigma += SQR(data[i]);			/* sum of subband's squared wavelet coefficients */

			wsigma /= len;						/* Now we get the estimate of the subband's variance */

			w = wsigma-SQR(sigma);				/* Now we get the estimate of noise-free subband variance, w */
			if (w <= 0.0)						/* If <=0, we have practically no variance and shouldn't filter. */
				t1 = 0;
			else
			{
				w = sigma*sigma / w;			/* Bayes threshold: Noise variance / signal variance */
				t1 = MIN(w,T);					/* Limit maximum to universal threshold */
			}

			if (t1==T)
				dp (2, "      BayesShrink, est. subband var=%f. Threshold (%d points) defaults to UT %f\n",wsigma,len,t1);
			else
				dp (2, "      BayesShrink, est. subband var=%f. Threshold for %d data points: %f\n",wsigma,len,t1);

			return t1;
			break;

		case 2:							/* SUREShrink */
										/* SUREshrink assumes that the input data are normalized for sigma=1.
											Let's do it in a clean way, then: Copy the input data, then sort it.
											Calculate the squared coefficients (and now normalized), and from those
											the cumulative coefficient.
										*/
			
			p = calloc (len, sizeof (double));		/* Make a copy of the input data, because it gets sorted */
			for (i=0; i<len; i++)
				p[i] = fabs(data[i]);				/* |wavelet coefficients|, not yet normalized */
			combsort_d (p, NULL, len);				/* Requires |wavelet coefficients| in ascending order */

			sumxx = 0;								/* Reset cumulative coefficients */
			for (i=0, j=len-1; i<len; i++, j--)		/* Loop over all coefficients to compute their associated risk */
			{
				buf = SQR(p[i] / sigma);			/* Normalized coefficient, squared */
				sumxx += buf;						/* Cumulative squared coefficients */
				w = sumxx + j*buf;					/* Intermediate risk term;  j counts backwards down to zero */
				risk = (len - 2*(i+1) + w)/len;		/* Stein's Risk Estimator; index i must start with 1 and include len */
				dp (3, "      Coeff. %f, cumulative %f. Stein's: %f,  Risk %f\n", sqrt(p[i]), sumxx, w, risk);
				if (i==0)
				{
					minrisk=risk; idx=0;			/* Initialize risk with first data point */
				}
				else if (minrisk > risk)
				{
					minrisk=risk; idx=i;			/* Keep track of lowest risk, index idx */
				}
			}

			w = p[idx];								/* Return absolute value of coefficient with lowest risk */
			free (p);
			T = MIN (w,T);							/* Keep threshold below universal threshold */
			if (w==T)
				dp (2, "      SUREShrink threshold for %d data points: %f with risk %f\n",len,T, minrisk);
			else
				dp (2, "      SUREShrink threshold for %d data points: %f with risk %f limited to UT %f\n",len,w, minrisk, T);

			return T;
			break;

		case 3:									/* General Cross-Validation (GCV) */

			max=data[0];
			for (i=0; i<len; i++)
				max = MAX (max,fabs(data[i]));			/* Find maximum absolute value in data stream */

			/* A constant-interval brute-force search is not sufficiently precise (or too slow),
				so we implement a bracketing search (see, for example, Golden Section Search,
					https://en.wikipedia.org/wiki/Golden-section_search
				or Fibonacci search).
			*/

			t1 = 0; t2 = max;						/* Initialize search bracket */
			cnt = 0;								/* Keep track of the iterations */
			golden_ratio = (sqrt(5.0)+1)/2.0;		/* Hence, "Golden Ratio Search" */
			t3 = t2 - (t2-t1)/golden_ratio;
			t4 = t1 + (t2-t1)/golden_ratio;			/* the initial two inner bracketing points */
			while ( fabs(t2-t1) > 0.01 )			/* Define convergence */
			{
				cnt++;
				gcv3 = gcv_eval (data, len, t3);	/* Get GCV values on inner bracket */
				gcv4 = gcv_eval (data, len, t4);
				dp (3, "GCV finder: gcv(%f)=%f and gcv(%f)=%f. Iteration %d\n", t3, gcv3, t4, gcv4, cnt);
				if (gcv3 < gcv4)
					t2 = t4;						/* Keep the lower value at t3 */
				else
					t1 = t3;						/* keep the lower value at t4 */

				t3 = t2 - (t2-t1)/golden_ratio;		/* Recompute the inner bracket points to avoid loss */
				t4 = t1 + (t2-t1)/golden_ratio;		/* of precision with potentially incorrect results */
			}

			T = (t1+t2)/2.0;						/* Final threshold after convergence */
			dp (2, "    --> GCV finder: Minimum GCV %f at threshold %f\n",gcv3, T);
			return T;
			break;

	}
	return 0;
}






/**********************************************************************************

 ####    #####   #   #    #*#     #*#     #*#     #*#    #   #    #*#   
 #   #   #       #   #   #   #     #     #   #     #     #   #   #   #  
 #   #   #       ##  #   #   #     #     #         #     ##  #   #      
 #   #   ####    # # #   #   #     #      #*#      #     # # #   #  ##  
 #   #   #       #  ##   #   #     #         #     #     #  ##   #   #  
 #   #   #       #   #   #   #     #     #   #     #     #   #   #   #  
 ####    #####   #   #    #*#     #*#     #*#     #*#    #   #    #*#   

*/



/* This function performs denoising (= coefficient shrinking) on a 1D time series.
	It decomposes the signal into a "reasonable" number of dyadic levels and then
	applies a shrinking function to each level except the lowest lowpass subband.
	"reasonable" means that the lowest band will have 16 data points.

	wfunc is the mother wavelet function; see "set_subband" for how that works.
	wshrinkfunc is the threshold finder and shrinking function:
		0: Universal threshold, soft thresholding
		1: Universal threshold, hard thresholding
		2: Bayes shrink, soft thresholding
		3: Bayes shrink, hard thresholding
		4: SUREShrink, soft thresholding
		5: SUREShrink, hard thresholding
		6: GCV, soft thresholding
		7: GCV, hard thresholding
	tmult is a threshold multiplier for arbitrary proportional correction of the threshold

*/

void wvlt_1D_denoising  (double *data, int len, int wfunc, int wshrinkfunc, double tmult)
{
int M;							/* Needed to zero-pad */
double *paddata;
double T, sigma;
int i;
int xleft, xright, level, decomp_level;
WAVELET_TYPE wtype;
wavefilt wfilt;	



	M=2; decomp_level = -3;
	while (M<len)
	{
		M+=M;						/* Nearest power-of two */
		decomp_level++;				/* Keep track of reasonable decomposition levels */
	}

	if (decomp_level < 1) return;			/* Too small a dataset, don't do anything */

	paddata = calloc (M, sizeof(double));
	for (i=0; i<len; i++) paddata[i] = data[i];

	wtype = set_subband (wfunc);
	if (wtype!=W_NULL) wfilt = wavelet_prepare (wtype);

	dwt_1d (paddata, M, W_FORWARD, wfilt, (*subband));			/* Wavelet transform */

	/* Estimate noise component */

	sigma = estimate_sigma_1d (paddata, M);
	dp (2, "1D wavelet denoising: Estimated sigma %f\n", sigma);

	/* Begin dyadic level-by-level coefficient shrinking */

	xright = M; xleft = M/2;

	for (level=0; level<decomp_level; level++)
	{
		T = find_wsthreshold (&paddata[xleft], xleft, sigma, wshrinkfunc >> 1);

		for (i=xleft; i<xright; i++)
		{
			paddata[i] = wv_thresh (paddata[i], T*tmult, wshrinkfunc & 0x01);
		}

		xright >>=1; xleft >>= 1;
	}

	dwt_1d (paddata, M, W_INVERSE, wfilt, (*subband));		/* Inverse Wavelet transform */
	for (i=0; i<len; i++) data[i] = paddata[i];				/* Return data in-place */
	free (paddata);											/* so we no longer need this buffer */

}


/* A stand-alone function to obtain a noise estimate. This is a 1D function,
	but we can feed it 2D data, interpreted as 1D (not recommended as it loses the spatial relationship!)
	We keep some coefficients fixed: We always use DAUB4 and one decomp_level
*/


double wvlt_noise_estimate  (double *data, int len)
{
int M;							/* Needed to zero-pad */
double *paddata;
double T, sigma;
int i;
int xleft, xright, level, decomp_level;
WAVELET_TYPE wtype;
wavefilt wfilt;	


	M=2; decomp_level = -3;
	while (M<len)
	{
		M+=M;						/* Nearest power-of two */
		decomp_level++;				/* Keep track of reasonable decomposition levels */
	}

	if (decomp_level < 1) return 0;			/* Too small a dataset, don't do anything */

	paddata = calloc (M, sizeof(double));
	for (i=0; i<len; i++) paddata[i] = data[i];

	wtype = set_subband (1);					/* This sets DAUB4 */
	wfilt = wavelet_prepare (wtype);

	dwt_1dn (paddata, M, W_FORWARD, 1, wfilt, (*subband));		/* Wavelet transform, one level only */

	/* Estimate noise component */

	sigma = estimate_sigma_1d (paddata, M);
	dp (2, "1D wavelet noise detection: Estimated sigma %f\n", sigma);
	free (paddata);
	return sigma;
}



/* Noise estimate of an image slice. This is the 2D analog function of wvlt_noise_estimate.
	Strictly, we can't use 1D for image noise estimation (as stipulated above).
	Some coefficients fixed: We always use DAUB8 and one decomp_level
*/


double img_noise_estimate  (image_type *img, int slice)
{
int M;							/* Needed to zero-pad */
double *padimg;
double sigma;
int x, y, xs, ys;
WAVELET_TYPE wtype;
wavefilt wfilt;


	xs = 8;
	while (xs < img->xmax) xs <<= 1;
	ys = 8;
	while (ys < img->ymax) ys <<= 1;
	dp (2, "Wavelet noise estimator: Padding image size (%d, %d) to (%d, %d)\n", img->xmax, img->ymax, xs,ys);


	padimg = calloc (xs*ys, sizeof(double ));

	/* Copy slice into the data buffer. For RGB types, use YUV */

	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
		{
			padimg[x + xs*y] = readbuf_flt (img, x,y, slice);
		}

	wtype = set_subband (2);					/* This sets DAUB8 */
	wfilt = wavelet_prepare (wtype);

	/* Estimate noise component, requires DWT-transformed image */

	dwt_2d (padimg, W_FORWARD, xs, ys, 1, wfilt, *subband);
	sigma = estimate_sigma_2d (padimg, xs, ys);

	dp (2, "2D image wavelet noise detection in slice %d: Estimated sigma %f\n", slice, sigma);
	free (padimg);
	return sigma;
}




/*+**********************************+******************+

  ####    #    #   #*#
 #    #   #    #    #
 #        #    #    #
 #   ##   #    #    #
 #    #   #    #    #
 #    #   #    #    #
  ####     ####    #*#

This GUI only for DWT and IDWT. Other (experimental) functions are provided in a module.

*/

struct
{
	DWTDIRECTION dwtdir;
	WAVELET_TYPE wvlt_select;
	void (*subband)(double *, unsigned long, DWTDIRECTION, wavefilt);	/* Pointer to the selected subband filter function */
	int decomp_level;
}
dwt_control;



void on_wvlt_go_clicked  (GtkButton *button, gpointer user_data)
{
WAVELET_TYPE wtype;
wavefilt wfilt;


	undo_nexus (&mainimg);

	wtype = set_subband (dwt_control.wvlt_select);
	if (wtype!=W_NULL) wfilt = wavelet_prepare (wtype);
	dwt_2d_image (&mainimg, dwt_control.dwtdir, dwt_control.decomp_level, wfilt, *subband);

	make_imlib_buffer_from_main_image (cur_slice);
	redraw(0);

	gtk_widget_destroy (GTK_WIDGET(user_data));
}



void wvlt_decomp_changed (GtkWidget* widget, gpointer user_data)
{ 
float buf;
GtkAdjustment *w;


	w = GTK_ADJUSTMENT(widget);			/* Funny way to typecast */
	buf= w->value;
	dwt_control.decomp_level = IROUND(buf);
	dp (2, "DWT decomposition level set to %d\n", dwt_control.decomp_level);
}



void on_wvlt_func_menu (GtkWidget *widget, gpointer user_data)
{
int func;

	func = *(int*) user_data;
	dwt_control.dwtdir = func;

}



void on_wvlt_type_menu (GtkWidget *widget, gpointer user_data)
{
int func;

	func = *(int*) user_data;
	dwt_control.wvlt_select = func;
	dp (2, "Changed wavelet to %d\n", func);

}

void on_wvlt_cancel  (GtkButton *button, gpointer user_data)
{
	gtk_widget_destroy (GTK_WIDGET(user_data));
}


void create_wvlt_gui (image_type *img)
{
  GtkWidget *wvlt_window;
  GtkWidget *vbox1, *hbox1, *wvlt_table;
  GtkWidget *label1, *label2, *label3;
  GtkWidget *hsep1;
  GtkWidget *wvlt_okbutton;
  GtkWidget *wvlt_showbutton;
  GtkWidget *wvlt_testbutton;
  GtkWidget *wvlt_cancelbutton;
  GtkWidget *wvlt_typemenu, *wvlt_typemenu_menu;
  GtkWidget *wvlt_funcmenu, *wvlt_funcmenu_menu;
  GtkWidget *menuitem;
  GtkWidget *wvlt_decomp;
  GtkAdjustment *wvlt_decomp_adj;
  GtkTooltips *tt_wvltcancel, *tt_wvltdecomp, *tt_wvltfn, *tt_wvltok, *tt_wvltshow, *tt_wvltshrink, *tt_wvlttest, *tt_wvlttmult, *tt_wvlttype;

  float min,max;
  int y;

	imgminmax (img, &min, &max);
	dwt_control.dwtdir = W_FORWARD;
	dwt_control.wvlt_select = W_HAAR;
	dwt_control.decomp_level = 2;

	wvlt_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (wvlt_window), "wvlt_gui", wvlt_window);
	gtk_window_set_title (GTK_WINDOW (wvlt_window), "Wavelet Transform");

	vbox1 = vbox_in_window (wvlt_window,"vbox1"); 
	wvlt_table = MakeTable (wvlt_window, vbox1, "table1", 2,6);

	y=0;
	label3 = MakeLabel (wvlt_window,wvlt_table,"Function","wvlt_label3",0,y);
	wvlt_funcmenu = Make_Optionmenu (wvlt_window, wvlt_table, "wvlt_funcmenu", 1,y);
	wvlt_funcmenu_menu = gtk_menu_new ();
	menuitem = Option_menuitem (wvlt_funcmenu_menu, GTK_SIGNAL_FUNC(on_wvlt_func_menu), "Forward DWT", (gpointer) &menuopt0);
	menuitem = Option_menuitem (wvlt_funcmenu_menu, GTK_SIGNAL_FUNC(on_wvlt_func_menu), "Inverse DWT", (gpointer) &menuopt1);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (wvlt_funcmenu), wvlt_funcmenu_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (wvlt_funcmenu), 0); 
	gtk_container_set_border_width (GTK_CONTAINER (wvlt_funcmenu), 5);
	tt_wvltfn = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_wvltfn, wvlt_funcmenu, "Select operation, discrete wavelet transform\nor inverse discrete wavelet transform.", NULL);

	y++;
	label1 = MakeLabel (wvlt_window,wvlt_table,"Wavelet type","wvlt_label1",0,y);
	wvlt_typemenu = Make_Optionmenu (wvlt_window, wvlt_table, "wvlt_typemenu", 1,y);
	wvlt_typemenu_menu = gtk_menu_new ();
	menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(on_wvlt_type_menu), "Haar", (gpointer) &menuopt0);
	menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(on_wvlt_type_menu), "Daubechies-4", (gpointer) &menuopt1);
	menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(on_wvlt_type_menu), "Daubechies-8", (gpointer) &menuopt2);
	menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(on_wvlt_type_menu), "Daubechies-12", (gpointer) &menuopt3);
	menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(on_wvlt_type_menu), "Daubechies-20", (gpointer) &menuopt4);
	menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(on_wvlt_type_menu), "Coiflet-6", (gpointer) &menuopt5);
	menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(on_wvlt_type_menu), "Coiflet12", (gpointer) &menuopt6);
	menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(on_wvlt_type_menu), "Symlet-4", (gpointer) &menuopt7);
	menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(on_wvlt_type_menu), "Symlet-8", (gpointer) &menuopt8);
	menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(on_wvlt_type_menu), "Symlet-12", (gpointer) &menuopt9);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (wvlt_typemenu), wvlt_typemenu_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (wvlt_typemenu), 0);
	gtk_container_set_border_width (GTK_CONTAINER (wvlt_typemenu), 5);
	tt_wvlttype = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_wvlttype, wvlt_typemenu, "Choose mother wavelet", NULL);

	y++;
	label2 = MakeLabel (wvlt_window,wvlt_table,"Decomposition level","wvlt_label2",0,y);
	wvlt_decomp_adj = GTK_ADJUSTMENT(gtk_adjustment_new (2, 1, 10, 1, 1, 0));
	wvlt_decomp = MakeHScale (wvlt_decomp_adj, wvlt_window, wvlt_table,"wvlt_decomp", 1,y);
	gtk_scale_set_digits (GTK_SCALE (wvlt_decomp), 0);
	tt_wvltdecomp = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_wvltdecomp, wvlt_decomp, "Maximum dyadic decomposition levels", NULL);

	hsep1 = MakeHsep (wvlt_window, vbox1, "hsep3", -1,-1);
	hbox1 = hbox_in_window (wvlt_window, vbox1, "wvlt_hbox");

	wvlt_okbutton = Button_in_table (wvlt_window, hbox1, "OK", "wvlt_okbutton",-1,-1, NULL);
	wvlt_cancelbutton = Button_in_table (wvlt_window, hbox1, "CANCEL", "wvlt_cancelbutton",-1,-1,NULL);
	tt_wvltok = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_wvltok, wvlt_okbutton, "Execute the selected operation", NULL);
	tt_wvltcancel = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_wvltcancel, wvlt_cancelbutton, "Abort operation, close window.", NULL);

	gtk_signal_connect (GTK_OBJECT (wvlt_decomp_adj), "value_changed", GTK_SIGNAL_FUNC (wvlt_decomp_changed), &menuopt1);
	gtk_signal_connect (GTK_OBJECT (wvlt_okbutton), "clicked", GTK_SIGNAL_FUNC (on_wvlt_go_clicked), wvlt_window);
	gtk_signal_connect (GTK_OBJECT (wvlt_cancelbutton), "clicked", GTK_SIGNAL_FUNC (on_wvlt_cancel), wvlt_window);
	gtk_signal_connect (GTK_OBJECT (wvlt_window), "destroy", GTK_SIGNAL_FUNC (on_wvlt_cancel), wvlt_window);

	gtk_widget_show_all (wvlt_window); 

}



/* Top-level entry points for main menu and for macro interpreter */

void dwt_main ()
{
	create_wvlt_gui (&mainimg);
}



void dwt_macro_driver (image_type *img, WAVELET_TYPE wvlt, int idwt, int level)
{
wavefilt wfilt;	
void (*subband)(double *, unsigned long, DWTDIRECTION, wavefilt);


	subband = &general_subband;			/* We always use the general_subband function in macros */
	wfilt = wavelet_prepare (wvlt);
	
	dwt_2d_image (img, idwt, level, wfilt, *subband);
	make_minmax (img);
}



/**********************************************************/


