

#ifndef __IMAGER_WVLTCORE_H__
#define __IMAGER_WVLTCORE_H__

#include "defs.h"



/* Different mother wavelet functions implemented here */

typedef enum { W_NULL, W_HAAR, W_DAUB4, W_DAUB8, W_DAUB12, W_DAUB20, W_COIF6, W_COIF12, W_SYM4, W_SYM8, W_SYM12 } WAVELET_TYPE;

/* Forward or inverse transform, just mnemonics. */

typedef enum { W_FORWARD, W_INVERSE } DWTDIRECTION;



/* This is the structure that holds a subband filter */

typedef struct 
{
    int n_param;		/* Number of filter coefficients */
    int hk_offs;		/* Parameter offset for scale filter centering, usually -n_param/2 */
    int gk_offs;		/* Parameter offset for wavelet filter centering, usually -n_param/2 */
    double *hk;			/* Array of scale filter coefficients */
    double *gk;			/* Array of wavelet filter coefficients */
} 
wavefilt;


/* A subband filter function, needs to point at the dyadic subband decomposition function */

void (*subband)(double *, unsigned long, DWTDIRECTION, wavefilt);	/* Pointer to the selected subband filter function */


char wvlt_used[32];			/* For messaging/debugging purposes */


/********************************************************/



wavefilt wavelet_prepare (WAVELET_TYPE wlt);
void general_subband (double *data, unsigned long n, DWTDIRECTION idwt, wavefilt wfilt);
void dead_subband (double *data, unsigned long n, DWTDIRECTION idwt, wavefilt dummy);
WAVELET_TYPE set_subband (int wavelet);

void dwt_1d (double *data, unsigned long n, DWTDIRECTION idwt, wavefilt wfilt, void (*wt_subband)(double *, unsigned long, DWTDIRECTION, wavefilt));
void dwt_1dn (double *data, unsigned long n, DWTDIRECTION idwt, int levels, wavefilt wfilt, void (*wt_subband)(double *, unsigned long, DWTDIRECTION, wavefilt));
void dwt_2d (double *data, DWTDIRECTION idwt, int xmax, int ymax, int decomp_level, wavefilt wfilt, void (*wt_subband)(double *, unsigned long, DWTDIRECTION, wavefilt));
void dwt_2d_image (image_type *img, DWTDIRECTION idwt, int decomp_level, wavefilt wfilt, void (*wt_subband)(double *, unsigned long, DWTDIRECTION, wavefilt));

double element_finder (double *data, int k, int n);

double wv_thresh (double elem, double t, int hardsoft);
double estimate_sigma_1d (double *data, int len);
double estimate_sigma_2d (double *data, int xmax, int ymax);
double estimate_sigma_2d_image (image_type *img, int z);

double find_wsthreshold (double *data, int len, double sigma, int wshrinkfunc);
//double find_1d_wsthreshold (double *data, int len, double sigma, int wshrinkfunc);
void wvlt_1D_denoising  (double *data, int len, int wfunc, int wshrinkfunc, double tmult);
double wvlt_noise_estimate  (double *data, int len);
double img_noise_estimate  (image_type *img, int slice);

void dwt_main ();
void dwt_macro_driver (image_type *img, WAVELET_TYPE wvlt, int idwt, int level);


#endif
